/******************************************************************************
**
** $Header: vector4d.h,v 1.11 94/03/17 15:19:01 danix Exp $
**
** $Log:	vector4d.h,v $
 * Revision 1.11  94/03/17  15:19:01  15:19:01  danix (Dani Lischinski)
 * Changed copy constructor to copy the entire vector at once.
 * 
 * Revision 1.10  92/07/06  17:38:01  17:38:01  fxt
 * *** empty log message ***
 * 
 * Revision 1.9  92/07/02  23:39:24  23:39:24  fxt (Filippo Tampieri)
 * *** empty log message ***
 * 
 * Revision 1.8  92/04/23  15:04:25  15:04:25  fxt (Filippo Tampieri)
 * new memory management using Pool
 * 
 * Revision 1.7  92/03/11  14:23:56  14:23:56  fxt (Filippo Tampieri)
 * changed a[...] to x,y,...
 * 
 * Revision 1.6  91/11/20  15:02:39  15:02:39  danix (Dani Lischinski)
 * Added read/write.
 * 
 * Revision 1.5  91/10/16  17:13:59  17:13:59  fxt (Filippo Tampieri)
 * unary minus
 * 
 * Revision 1.4  91/08/30  10:53:26  10:53:26  fxt (Filippo Tampieri)
 * added abs()
 * 
 * Revision 1.3  91/06/06  15:17:02  15:17:02  fxt (Filippo Tampieri)
 * added constructor to homogeneous coordinates.
 * 
 * Revision 1.2  91/06/06  10:49:35  10:49:35  fxt (Filippo Tampieri)
 * changed enum Coordinate to int and
 * added operator==.
 * 
 * Revision 1.1  91/06/04  15:48:47  15:48:47  fxt (Filippo Tampieri)
 * Initial revision
 * 
**
******************************************************************************/

#ifndef VECTOR4D_H
#define VECTOR4D_H

#include <iostream.h>
#include <math.h>
#include <common.h>
#include <pool.h>
#include <geom.h>
#include <vector3d.h>

class Vector4d {
friend Vector4d operator+(const Vector4d& u, const Vector4d& v);
friend Vector4d operator-(const Vector4d& v);
friend Vector4d operator-(const Vector4d& u, const Vector4d& v);
friend Real operator|(const Vector4d& u, const Vector4d& v);
friend Vector4d operator*(const Vector4d& u, const Vector4d& v);
friend Vector4d operator*(Real c, const Vector4d& v);
friend Vector4d operator*(const Vector4d& v, Real c);
friend Vector4d operator/(const Vector4d& v, Real c);
friend Vector4d max(const Vector4d& u, const Vector4d& v);
friend Vector4d min(const Vector4d& u, const Vector4d& v);
friend Vector4d abs(const Vector4d& v);
friend int operator==(const Vector4d&, const Vector4d&);
private:
	Real x, y, z, w;
	static Mpool pool;
public:
	Vector4d()									{ ; }
	Vector4d(const Vector4d& v)					{ *this = v; }
	Vector4d(Real, Real, Real, Real);
	Vector4d(const Vector3d&, Real);
	void set(Real, Real, Real, Real);
	Real& operator[](int i)				{ return ((Real *)this)[i]; }
	const Real& operator[](int i) const	{ return ((Real *)this)[i]; }
	Real norm() const;
	Boolean normalize();
	Boolean normalize(const Vector4d&);
	Vector4d operator+=(const Vector4d&);
	Vector4d operator-=(const Vector4d&);
	Vector4d operator*=(const Vector4d&);
	Vector4d operator*=(Real);
	Vector4d operator/=(Real);
	int imax();
	int imin();
    void write(FILE *) const;
    void read(FILE *);
	static Real eps, eps2;
	static void setEps(Real _eps)		{ eps = _eps; eps2 = _eps * _eps; }
	static Vector4d* New(unsigned int n)			{ return (Vector4d*)pool.Alloc(n); }
	static void Delete(unsigned int n, Vector4d* a)	{ pool.Free(n, (void*)a); }
	static void *operator new(size_t)				{ return pool.Alloc(1); }
	static void operator delete(void *p, size_t)	{ pool.Free(1, p); }
};

extern ostream& operator<<(ostream&, const Vector4d&);
extern istream& operator>>(istream&, Vector4d&);

inline Vector4d::Vector4d(Real _x, Real _y, Real _z, Real _w)
{
	x = _x;
	y = _y;
	z = _z;
	w = _w;
}

inline Vector4d::Vector4d(const Vector3d& v, Real _w)
{
	x = v[X];
	y = v[Y];
	z = v[Z];
	w = _w;
}

inline void Vector4d::set(Real _x, Real _y, Real _z, Real _w)
{
	x = _x;
	y = _y;
	z = _z;
	w = _w;
}

inline Vector4d Vector4d::operator+=(const Vector4d& v)
{
	x += v.x;
	y += v.y;
	z += v.z;
	w += v.w;

	return *this;
}

inline Vector4d Vector4d::operator-=(const Vector4d& v)
{
	x -= v.x;
	y -= v.y;
	z -= v.z;
	w -= v.w;

	return *this;
}

inline Vector4d Vector4d::operator*=(const Vector4d& v)
{
	x *= v.x;
	y *= v.y;
	z *= v.z;
	w *= v.w;

	return *this;
}

inline Vector4d Vector4d::operator*=(Real c)
{
	x *= c;
	y *= c;
	z *= c;
	w *= c;

	return *this;
}

inline Vector4d Vector4d::operator/=(Real c)
{
	x /= c;
	y /= c;
	z /= c;
	w /= c;

	return *this;
}

inline Real Vector4d::norm() const
{
	return sqrt(x*x + y*y + z*z + w*w);
}

inline int Vector4d::imax()
{
	int i, j;

	for(i = Y, j = X; i <= W; i++)
		if((*this)[i] > (*this)[j])
			j = i;
	
	return j;
}

inline int Vector4d::imin()
{
	int i, j;

	for(i = Y, j = X; i <= W; i++)
		if((*this)[i] > (*this)[j])
			j = i;
	
	return j;
}

inline Vector4d operator+(const Vector4d& u, const Vector4d& v)
{
	return Vector4d(u.x + v.x, u.y + v.y, u.z + v.z, u.w + v.w);
}

inline Vector4d operator-(const Vector4d& v)
{
	return Vector4d(-v.x, -v.y, -v.z, -v.w);
}

inline Vector4d operator-(const Vector4d& u, const Vector4d& v)
{
	return Vector4d(u.x - v.x, u.y - v.y, u.z - v.z, u.w - v.w);
}

inline Real operator|(const Vector4d& u, const Vector4d& v)
{
	return (u.x*v.x + u.y*v.y + u.z*v.z + u.w*v.w);
}

inline Vector4d operator*(const Vector4d& u, const Vector4d& v)
{
	return Vector4d(u.x * v.x, u.y * v.y, u.z * v.z, u.w * v.w);
}

inline Vector4d operator*(Real c, const Vector4d& v)
{
	return Vector4d(c * v.x, c * v.y, c * v.z, c * v.w);
}

inline Vector4d operator*(const Vector4d& v, Real c)
{
	return Vector4d(c * v.x, c * v.y, c * v.z, c * v.w);
}

inline Vector4d operator/(const Vector4d& v, Real c)
{
	return Vector4d(v.x / c, v.y / c, v.z / c, v.w / c);
}

inline Vector4d max(const Vector4d& u, const Vector4d& v)
{
	return Vector4d(max(u.x,v.x), max(u.y,v.y), max(u.z,v.z), max(u.w,v.w));
}

inline Vector4d min(const Vector4d& u, const Vector4d& v)
{
	return Vector4d(min(u.x,v.x), min(u.y,v.y), min(u.z,v.z), min(u.w,v.w));
}

inline Vector4d abs(const Vector4d& v)
{
	return Vector4d(ABS(v.x), ABS(v.y), ABS(v.z), ABS(v.w));
}

typedef Vector4d Point4d;
typedef Vector4d Normal4d;

#endif

