/******************************************************************************
 *
 * $Source: /homes/fish/danix/latte/bsp/RCS/bsptree.C,v $
 * $Author: danix $
 * $Locker:  $
 *
 * $Log: bsptree.C,v $
 * Revision 1.4  1995/12/14  00:21:00  danix
 * *** empty log message ***
 *
 * Revision 1.3  1995/11/21  21:16:14  danix
 * Added the total extent along a direction to the score in
 * ComputeBestScore (to improve aspect ratios).
 *
 * Revision 1.2  1995/09/26  21:31:04  danix
 * Fixed the Init routine, added a makeLeaf member function.
 *
 * Revision 1.1  1995/09/25  18:20:53  danix
 * Initial revision
 *
 *****************************************************************************/

#include <bsptree.h>
#include <values.h>
#include <sort.h>

#define MAX_OBJECTS 4

class EntryExit {
public:
	EntryExit() {}
	float  value;
	int numFaces;
};

int operator<(const EntryExit& a, const EntryExit& b)
{
	return (a.value < b.value);
}

int operator<=(const EntryExit& a, const EntryExit& b)
{
	return (a.value <= b.value);
}

int operator>(const EntryExit& a, const EntryExit& b)
{
	return (a.value > b.value);
}

int operator>=(const EntryExit& a, const EntryExit& b)
{
	return (a.value >= b.value);
}

static EntryExit *minarray, *maxarray;

float ComputeScore(int behind, int inside, int infront,
				   int behind_faces, int inside_faces, int infront_faces)
{
	float score, ratio, face_ratio;
	behind += inside;
	infront += inside;
	ratio = (float)min(behind, infront) / (float)max(behind, infront);
	ratio = 1.0 / max(ratio, float(0.001));
	behind_faces += inside_faces;
	infront_faces += inside_faces;
	face_ratio = (float)min(behind_faces, infront_faces) /
		(float)max(behind_faces, infront_faces);
	face_ratio = 1.0 / max(face_ratio, float(0.001));
	score = 4.0 * pow(inside, 3) + pow(face_ratio, 2);
	return score;
}

float ComputeBestScore(ObjList& objs, int angle, float& dist)
{
	float score, min_score = MAXFLOAT;
	int k = 0, total_faces = 0, num = objs.length();

	for (ListPos p = objs.first(); !objs.isEnd(p); p = objs.next(p), k++) {
		ObjectPtr obj = objs.retrieve(p);
		minarray[k].value = obj->extent(angle).min;
		maxarray[k].value = obj->extent(angle).max;
		minarray[k].numFaces = maxarray[k].numFaces = obj->numFaces();
		total_faces += obj->numFaces();
	}		

	// sort both arrays
	QuickSort(minarray, 0, num-1);
	QuickSort(maxarray, 0, num-1);

	Assert(minarray[0].value <= maxarray[0].value);
	Assert(minarray[num-1].value <= maxarray[num-1].value);
	Assert(minarray[0].value < maxarray[num-1].value);

	// Find best splitting plane for this angle:
	// we're looking for a splitting plane that will split as few
	// extents as possible. Given two planes that split the same
	// number of extents we'll choose the one with a better balance.
	int i = -1, j = -1, inside = 0, behind = 0, infront = num;
	int inside_faces = 0, behind_faces = 0, infront_faces = total_faces;
	int lowest_count = num;
	float ratio = 0;
	float start, end;
	float extent = maxarray[num-1].value - minarray[0].value;

	while (TRUE) {

		Assert(j <= i);

		if (i < num-1 && minarray[i+1].value < maxarray[j+1].value) {
			i++, inside++, infront--;
			inside_faces += minarray[i].numFaces;
			infront_faces -= minarray[i].numFaces;
			start = minarray[i].value;
		} else if (minarray[i+1].value > maxarray[j+1].value) {
			j++, inside--, behind++;
			inside_faces -= maxarray[j].numFaces;
			behind_faces += maxarray[j].numFaces;
			start = maxarray[j].value;
		} else // minarray[i+1].value == maxarray[j+1].value
			if (inside > 0) {
				j++, inside--, behind++;
				inside_faces -= maxarray[j].numFaces;
				behind_faces += maxarray[j].numFaces;
				start = maxarray[j].value;
			} else {
				i++, inside++, infront--;
				inside_faces += minarray[i].numFaces;
				infront_faces -= minarray[i].numFaces;
				start = minarray[i].value;
			}

		if (behind >= num || maxarray[j].value >= maxarray[num-1].value)
			break;

		end = (i < num-1 && minarray[i+1].value <= maxarray[j+1].value) ?
			minarray[i+1].value : maxarray[j+1].value;

		if (end <= minarray[0].value)
			continue;

		if (end >= maxarray[num-1].value)
			end = max(minarray[num-1].value, maxarray[num-2].value);

		if (start <= minarray[0].value)
			start = end;

		// Compute score:
		score = ComputeScore(behind, inside, infront, behind_faces,
							 inside_faces, infront_faces);
		if (score < min_score) {
			min_score = score;
			lowest_count = inside;
			ratio = (float)min(behind_faces + inside_faces,
							   infront_faces + inside_faces) /
					(float)max(behind_faces + inside_faces,
							   infront_faces + inside_faces);
			dist = 0.5 * (start + end);
		}
	}
	
	cerr << "Plane is " << dist << endl;
	cerr << "Splitting " << lowest_count << " extents";
	cerr << " with balance ratio of " << ratio << endl;
	cerr << "Score is " << min_score << " + " << 50.0/extent << " = "
		<< min_score + 50.0/extent << endl;

	// The final score includes the extent of the objects along this
	// direction, in order to facilitate splits with good aspect ratio.
	return min_score + 50.0/extent;
}

void BSPtree::Init(ObjList& objs)
{
	if (objs.length() <= MAX_OBJECTS) {
		makeLeaf(objs);
		return;
	}

	cerr << "\nProcessing " << objs.length() << " objects\n";

	// ******** Create an appropriate separating plane **************
	float score, best_score = MAXFLOAT;
	int best_angle = -1;
	float dist, best_dist, radians;

	minarray = ::new EntryExit[objs.length()];
	maxarray = ::new EntryExit[objs.length()];

	for (int angle = 0; angle < Object::Angles; angle++) {
		cerr << "Angle " << angle << ":\n";
		score = ComputeBestScore(objs, angle, dist);
		if (score < best_score) {
			best_score = score;
			best_angle = angle;
			best_dist  =  dist;
		}
	}

	::delete[] minarray;
	::delete[] maxarray;

	if (best_angle < 0) {
		// Could not find a splitting plane
		makeLeaf(objs);
		return;
	}
		
	// having chosen the angle with the best score, create an
	// appropriate separating plane.
	radians = (best_angle * M_PI) / Object::Angles;
	cerr << "Best score is " << best_score << " for angle: " << radians <<
		", dist = " << best_dist << endl;
	Point3d p(best_dist * cos(radians), 0, best_dist * sin(radians));
	Normal3d n(cos(radians), 0, sin(radians));
	sep.set(n, p);
	cerr << "Separating plane is: " << sep << endl << endl;
	// **************************************************************

	// split object list into two lists, one for each side of the
	// separating plane, create children and recurse.
	ObjectPtr obj;
	ListPos lp;
	List<ObjectPtr> poslist, neglist, split;
	for (lp = objs.first(); !objs.isEnd(lp); lp = objs.next(lp)) {
		obj = objs.retrieve(lp);
		if (obj->extent(best_angle).max <= best_dist)
			// put object in negative halfspace
			neglist.append(obj);
		else if (obj->extent(best_angle).min >= best_dist)
			// put object in positive halfspace
			poslist.append(obj);
		else
			split.append(obj);
	}

	// No use in splitting unless each of the children will contain
	// fewer objects than the parent:
	if (poslist.length() + split.length() >= objs.length() ||
		neglist.length() + split.length() >= objs.length()) {
		// Make current node a leaf containing all the objects:
		makeLeaf(objs);

		// The lists are automatic variables, so they will be cleaned
		// up by their destructors upon exiting the routine
		return;
	}

	// Ok, let's actually split the objects in splitlist
	while (!split.isEmpty()) {
		obj = split.pop();

		// split object, append each of the resulting objects
		// to the appropriate list:
		Object *o1 = 0, *o2 = 0;
		obj->split(sep, o1, o2);
		Assert(o1 || o2);
		if (o1)
			neglist.append(o1);
		if (o2)
			poslist.append(o2);
	}

	Assert(poslist.length() < objs.length() && !poslist.isEmpty() &&
		   neglist.length() < objs.length() && !neglist.isEmpty());

	cerr << "Split: +" << poslist.length() << -neglist.length() << endl;

	if ((pos = new BSPtree) == NIL(BSPtree))
		Error("BSPtree::Init: new failed");
	pos->Init(poslist);

	if ((neg = new BSPtree) == NIL(BSPtree))
		Error("BSPtree::Init: new failed");
	neg->Init(neglist);
}

void BSPtree::Insert(Object& obj)
{
	if (IsLeaf()) {
		if (!objects)
			objects = new ObjList;
		if (!objects)
			Error("BSPtree::insert: new failed");
		objects->append(&obj);
	} else {
		if (obj >= sep)
			Pos().Insert(obj);
		else if (obj <= sep)
			Neg().Insert(obj);
		else {
			Object *o1 = 0, *o2 = 0;
			obj.split(sep, o1, o2);
			Assert(o1 || o2);
			if (o1) Neg().Insert(*o1);
			if (o2) Pos().Insert(*o2);
		}
	}
}

void BSPtree::makeLeaf(ObjList& objs)
{
	if ((objects = new ObjList) == NIL(ObjList))
		Error("BSPtree::makeLeaf: new failed");
	*objects = objs;
	cerr << "Leaf has " << objects->length() << " objects\n";
}

void BSPtree::print(int d) const
{
	for (int i = 0; i < d; i++)
		cerr << "    ";

	cerr << "Level " << d << ": ";
	if (!IsLeaf())
		cerr << "(" << sep << ") ";

	if (objects) {
		for (ListPos p = objects->first(); !objects->isEnd(p); p = objects->next(p))
			cerr << objects->retrieve(p)->fileName() << " ";
	}
	cerr << endl;

	if (pos)
		pos->print(d+1);
	if (neg)
		neg->print(d+1);
}

static FILE *_fp;

static void writeObject(ObjectPtr obj)
{
	obj->write(_fp);
}

void BSPtree::write(FILE *fp, Boolean root) const
{
	if (root) {
		// write out total number of objects in the tree:
		fprintf(fp, "%d\n", Object::count);
	} else
		fprintf(fp, "\n");
	if (IsLeaf()) {

		Assert(objects != NIL(ObjList) && !objects->isEmpty());

		// write out 0 0 0 0 (a bogus palne equation):
		fprintf(fp, "0 0 0 0\n");

		// write out number of objects:
		fprintf(fp, "%d\n", objects->length());

		// write out the objects themselves:
		_fp = fp;
		objects->apply(writeObject);

	} else {

		Assert(objects == NIL(ObjList) && pos && neg);

		// write out separating plane equation:
		Normal3d n(sep.normal());
		fprintf(fp, "%g %g %g %g\n", n[0], n[1], n[2], sep.d());

		// write out 0 to indicate this is an interior node
		fprintf(fp, "0\n");

		neg->write(fp);
		pos->write(fp);

	}
}
