//============================================================================
//
//		Billboard.c  :  
//
//			Jonathan Shade
//
//
//============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <GL/gl.h>
#include <GL/glu.h>
#include "Billboard.h"
#include "timing.h"


//
//	Make texture a "class" variable that all instances
//	of Billboard will share. Initialize it only once.
//
int			TextureSize		= 1280*1024*4;	// RGBA
GLubyte		*Texture 		= NULL;

int		TakeTimings = 1;

Timer	pr;		// Read Pixels
Timer	td;		// Define Texture
Timer	ct;		// Create Texture - all of the above



//----------------------------------------------------------------------------
//
//
//  constructor:
//
//
//----------------------------------------------------------------------------
Billboard::Billboard()
	{
	textureValid =  0;
	textureNameValid =  0;

	//
	//	Initialize class texture memory
	//
#ifndef COPY_TEX
	if( Texture == NULL )
		Texture = new GLubyte [TextureSize ];
#endif
	}

//----------------------------------------------------------------------------
//
//
// destructor :
//
//
//----------------------------------------------------------------------------
Billboard::~Billboard()
	{
	}

//----------------------------------------------------------------------------
//
//		getSize
//
//----------------------------------------------------------------------------
int
Billboard::getSize()
	{
	if( textureValid )	return( adj_width*adj_height );
	else				return( 0 );
	}
	
//----------------------------------------------------------------------------
//
//		getSizeUsed
//
//----------------------------------------------------------------------------
int
Billboard::getSizeUsed()
	{
	if( textureValid )	return( tex_width*tex_height );
	else				return( 0 );
	}

//----------------------------------------------------------------------------
//
//	Feasible Texture
//
//----------------------------------------------------------------------------
int
Billboard::feasibleTexture( const BoundBox & bbox, const Camera & camera,
							const Matrix4d & matrix, double extra_pixels )
	{
	int status = proj_FeasibleScreenPolygon( bbox, camera,
											 matrix, extra_pixels );
	if( status != 1 )	invalidate();

	return( status );
	}

//----------------------------------------------------------------------------
//
//	Create Texture	- Texture Object
//
//----------------------------------------------------------------------------
int
Billboard::createTextureTO( const BoundBox & bbox, double extra_pixels,
							int take_timings )
	{
	int	tmp_size,status;
	GLint	wbox[2][2];

	//
	//	Start timings
	//
	if( take_timings )
		{
		RESET_TIMER( ct ); RESET_TIMER( pr ); RESET_TIMER( td );
		}

	if( take_timings ) START_TIMER( ct );

	//
	//	Get the outline of the projected bounding box.
	//	If the outline is outside of the window, then
	//	the object is not in the field of view, ignore
	//	it.
	//
	//	If "proj_GetScreenPolygon" fails, invalidate the
	//	texture map and propagate the return value.
	// 
	status = proj_GetScreenPolygon( bbox, extra_pixels,	wbox, plane );

	//fprintf(stderr,"bill: %d %d %d %d\n",wbox[0][0],wbox[0][1],wbox[1][0],wbox[1][1]);


	if( status != 1 )
		{
		invalidate();
		STOP_TIMER( ct );
		return( status );
		}


//
//	Debugging Textures on non-RE
//
#ifdef DEBUG_TEXTURE
textureValid = 1;
newTexture = 1;
return 1;
#endif

	//
	//	Compute the larger "power of 2" texture to
	//	encase the pixels grabbed from the buffer.
	//
	//
	//	Dropped the +1 because of the ceil in the
	//	projection.
	//
	tex_width	= wbox[1][0]-wbox[0][0];
	tex_height	= wbox[1][1]-wbox[0][1];

	proj_GetPower2Size( tex_width, tex_height, adj_width, adj_height );

	//
	//	Allocate memory for the "class" texture map.
	//	This will never happen, since the texture is
	//	initialized to be the size of the screen. It is
	//	here for completeness ...
	//
	tmp_size = adj_width*adj_height*4;

	if ( tmp_size > TextureSize )
		{
		fprintf(stderr,"CT: allocating bigger texture.\n");

		TextureSize = tmp_size;

		if ( Texture )
			delete [] Texture;

		Texture = new GLubyte [ TextureSize ];
		}


	//++++++++++++++++++++++++++++++++++++++++++++++++++
	//
	//	Grab the drawn image as a texture map.
	//
	//
	//	Only read the "valid" part of the screen.
	//	This call tells GL that the texture we will
	//	define later is a bigger "power of 2" thing,
	//	and to read the pixels from the screen into
	//	memory as though we are reading a sub-rectangle
	//	of that "soon to be" texture.
	//
	//	This allows us to skip an expensive pixel
	//	copying operation from the old "small" texture
	//	to the "real" "power of 2" texture.
	//
	//++++++++++++++++++++++++++++++++++++++++++++++++++


	//
	//	Grab
	//
	//		Bad fix here, copying adj_width, adj_height
	//		when I really only need to copy something like:
	//		tex_width+5, tex_height+5.
	//
	//		Grab 10% extra - leave "outer" space garbage
	//
	if( take_timings )	START_TIMER( pr );
#ifndef COPY_TEXTURE

//	int fudge_width  = tex_width  + tex_width/10;
//	int fudge_height = tex_height + tex_height/10;

//		glPixelStorei( GL_PACK_ROW_LENGTH, adj_width );

		glReadPixels(	wbox[0][0], wbox[0][1],
						adj_width, adj_height,
						GL_RGBA, GL_UNSIGNED_BYTE,
						Texture );
#endif
	if( take_timings )	STOP_TIMER( pr );


	//
	//	Hand the "power of 2" texture off to GL
	//
	if( take_timings )	START_TIMER( td );

		textureValid = 1;

	//	GLboolean tv =  glIsTextureEXT( textureID );
	//	if( tv == GL_FALSE )
		if( textureNameValid == 0 )
			{
			textureNameValid = 1;
			glGenTexturesEXT( 1, &textureID );
			//fprintf(stderr,"%d \n", textureID );
			}
		else
			{
			//glBindTextureEXT( GL_TEXTURE_2D, textureID );
			//glDeleteTexturesEXT( 1, &textureID );
			//glGenTexturesEXT( 1, &textureID );
			}

		glBindTextureEXT( GL_TEXTURE_2D, textureID );

		glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP );
		glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP );
//		glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
//		glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
		glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
		glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );

#ifndef COPY_TEXTURE
		glTexImage2D( GL_TEXTURE_2D, 0, 4, adj_width, adj_height,
					  0, GL_RGBA, GL_UNSIGNED_BYTE, Texture );
#else
/*
		int xf = adj_width-tex_width;
		int yf = adj_height-tex_height;

		if( xf > 2 )	xf = 2;
		if( yf > 2 )	yf = 2;

		glCopyTexSubImage2DEXT( GL_TEXTURE_2D, 0, 0, 0, wbox[0][0], wbox[0][1],
							 tex_width+xf, tex_height+yf  );
*/

		glCopyTexImage2DEXT( GL_TEXTURE_2D, 0, GL_RGBA, wbox[0][0], wbox[0][1],
							 adj_width, adj_height, 0 );

#endif
	
	if( take_timings )	STOP_TIMER( td );


	//
	//	Check for GL errors
	//
#ifdef DEBUG_GL
	{
	GLenum error_code;
		while((error_code=glGetError()) != GL_NO_ERROR)
			{
			fprintf(stderr,"\t%03d Billboard::createTexture\tgl error: %s\n",
				dispListID, gluErrorString(error_code));
			}
	}
#endif

	double smax = (double) tex_width  / (double) adj_width;
	double tmax = (double) tex_height / (double) adj_height;

	//fprintf(stderr,"bill: %d %d\t %d %d\n",tex_width,tex_height,adj_width,adj_height);
	//fprintf(stderr,"bill: %f %f\n",smax,tmax );

	m_tc[0][0] = 0.0;
	m_tc[0][1] = 0.0;
	m_tc[1][0] = smax;
	m_tc[1][1] = 0.0;
	m_tc[2][0] = smax;
	m_tc[2][1] = tmax;
	m_tc[3][0] = 0.0;
	m_tc[3][1] = tmax;

	newTexture = 1;

	if( take_timings )
		{
		STOP_TIMER( ct );
		
		fprintf( stderr,"create texture:\t\t\t%d\n", calc_usec_time( ct ) );
		fprintf( stderr,"read pixels:\t\t\t%d\n", calc_usec_time( pr ) );
		fprintf( stderr,"deftex:\t\t%3dx%3d:\t%d\n", adj_width,adj_height,calc_usec_time( td ) );
		}
	
	return( adj_width*adj_height );
	}


//----------------------------------------------------------------------------
//
//		kill
//
//----------------------------------------------------------------------------
void
Billboard::kill()
	{
	if( !textureValid )	return;

	textureValid = 0;

//	GLboolean tv =  glIsTextureEXT( textureID );
//	if( tv == GL_FALSE )	fprintf(stderr,"Killing invalid texture\n");

	if( textureNameValid == 0 )	fprintf(stderr,"Killing invalid texture\n");

	glDeleteTexturesEXT( 1, &textureID );
	
	textureNameValid = 0;

	//
	//	Check for GL errors
	//
#ifdef DEBUG_GL
	GLenum error_code;
	while((error_code=glGetError()) != GL_NO_ERROR)
		{
		fprintf(stderr,"\tBillboard::invalidate %02d\tgl error: %s\n"
						,dispListID,gluErrorString(error_code));
		}
#endif
	}

//----------------------------------------------------------------------------
//
//		invalidate
//
//----------------------------------------------------------------------------
void
Billboard::invalidate()
	{
	if( !textureValid )	return;

	textureValid = 0;
//	glDeleteTexturesEXT( 1, &textureID );
//	textureNameValid = 0;

	//
	//	Check for GL errors
	//
#ifdef DEBUG_GL
	GLenum error_code;
	while((error_code=glGetError()) != GL_NO_ERROR)
		{
		fprintf(stderr,"\tBillboard::invalidate %02d\tgl error: %s\n"
						,dispListID,gluErrorString(error_code));
		}
#endif
	}

//----------------------------------------------------------------------------
//
//	Draw Texture Frame
//
//----------------------------------------------------------------------------
void
Billboard::drawTextureFrame()
	{
	//
	//	Draw an outline around the polygon
	//
	if( newTexture )
		glColor3f( 1.0, 0.0, 0.0 );
	else
		glColor3f( 1.0, 1.0, 1.0 );

	glBegin(GL_LINE_LOOP);
		glVertex3dv( &(plane[0][0]) );
		glVertex3dv( &(plane[1][0]) );
		glVertex3dv( &(plane[2][0]) );
		glVertex3dv( &(plane[3][0]) );
	glEnd();

	newTexture = 0;
	}
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void
Billboard::drawTextureFrameNocolor()
	{
	//
	//	Draw an outline around the polygon
	//
	glBegin(GL_LINE_LOOP);
		glVertex3dv( &(plane[0][0]) );
		glVertex3dv( &(plane[1][0]) );
		glVertex3dv( &(plane[2][0]) );
		glVertex3dv( &(plane[3][0]) );
	glEnd();
	}

//----------------------------------------------------------------------------
//
//		getCorner
//
//----------------------------------------------------------------------------
Point3d
Billboard::getCorner( int i )
	{
	return( Point3d( plane[i][0], plane[i][1], plane[i][2] ) );
	}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
BoundBox
Billboard::getBoundBox()
	{
	return bbox;
	}


//----------------------------------------------------------------------------
//
//		draw
//
//----------------------------------------------------------------------------
int
Billboard::draw()
	{
	if( !textureValid )
		return 0;

//
// Debugging textures on non RE
//
#ifdef DEBUG_TEXTURE
return 1;
#endif

	//
	//	Texture map the polygon
	//

#ifdef TEX_OBJ
//	if( glIsTextureEXT( textureID ) == GL_FALSE )
	if( textureNameValid == 0 )
		fprintf(stderr,"Trying to draw an invalid Texture!!!\n");

	glBindTextureEXT( GL_TEXTURE_2D, textureID );
#else
	glCallList( dispListID );
#endif

	glColor4f( 1.0, 1.0, 1.0, 1.0 );

		glBegin(GL_QUADS);

			glTexCoord2d( m_tc[0][0], m_tc[0][1] );
			glVertex3dv( &(plane[0][0]) );

			glTexCoord2d( m_tc[1][0], m_tc[1][1] );
			glVertex3dv( &(plane[1][0]) );

			glTexCoord2d( m_tc[2][0], m_tc[2][1] );
			glVertex3dv( &(plane[2][0]) );

			glTexCoord2d( m_tc[3][0], m_tc[3][1] );
			glVertex3dv( &(plane[3][0]) );
		glEnd();

	//
	//	Check for GL errors
	//
#ifdef DEBUG_GL
	{
	GLenum error_code;
		while((error_code=glGetError()) != GL_NO_ERROR)
			fprintf(stderr,"\t%03d Billboard::draw\tgl error: %s\n",
				dispListID, gluErrorString(error_code));
	}
#endif

	return 1;
	}

//----------------------------------------------------------------------------
//
//		draw
//
//----------------------------------------------------------------------------
void
Billboard::drawColored()
	{
	if( !textureValid )	return;

	//
	//	Texture map the polygon
	//
	if( textureNameValid == 0 )
		fprintf(stderr,"Trying to draw an invalid Texture!!!\n");

	glBindTextureEXT( GL_TEXTURE_2D, textureID );

	if( newTexture )
		glColor4f( 1.0, 0.0, 0.0, 1.0 );
	else
		glColor4f( 1.0, 1.0, 1.0, 1.0 );

		glBegin(GL_QUADS);

			glTexCoord2d( m_tc[0][0], m_tc[0][1] );
			glVertex3dv( &(plane[0][0]) );

			glTexCoord2d( m_tc[1][0], m_tc[1][1] );
			glVertex3dv( &(plane[1][0]) );

			glTexCoord2d( m_tc[2][0], m_tc[2][1] );
			glVertex3dv( &(plane[2][0]) );

			glTexCoord2d( m_tc[3][0], m_tc[3][1] );
			glVertex3dv( &(plane[3][0]) );
		glEnd();

	newTexture = 0;
	}


//----------------------------------------------------------------------------
//
//		Print_Timings:
//
//----------------------------------------------------------------------------
void
Billboard::printTimings( int id )
	{
	fprintf( stderr,"\tct: %ld\n", calc_usec_time( ct ) );
	fprintf( stderr,"\tpr: %ld\n", calc_usec_time( pr ) );
	fprintf( stderr,"\ttd: %ld\n", calc_usec_time( td ) );
	fprintf( stderr,"Texture: %d (%d) (%d)\n", id, dispListID, textureID );
	fprintf( stderr,"--------\n" );
	}	


//----------------------------------------------------------------------------
//
//	Create Texture	- Display List
//
//----------------------------------------------------------------------------
int
Billboard::createTextureDL( const BoundBox & bbox, double extra_pixels,
							int take_timings )
	{
	int	tmp_size,status;
	GLint	wbox[2][2];

	//
	//	Start timings
	//
	if( take_timings )
		{
		RESET_TIMER( ct ); RESET_TIMER( pr ); RESET_TIMER( td );
		}

	if( take_timings )	gettimeofday( &ct.start, &ct.zone );

	//
	//	Get the outline of the projected bounding box.
	//	If the outline is outside of the window, then
	//	the object is not in the field of view, ignore
	//	it.
	//
	//	If "proj_GetScreenPolygon" fails, invalidate the
	//	texture map and propagate the return value.
	// 
	status = proj_GetScreenPolygon( bbox, extra_pixels,	wbox, plane );

	if( status != 1 )
		{
		invalidate();
		STOP_TIMER( ct );
		return( status );
		}


//
//	Debugging Textures on non-RE
//
#ifdef DEBUG_TEXTURE
textureValid = 1;
newTexture = 1;
return 1;
#endif

	//
	//	Compute the larger "power of 2" texture to
	//	encase the pixels grabbed from the buffer.
	//
	tex_width	= wbox[1][0]-wbox[0][0];
	tex_height	= wbox[1][1]-wbox[0][1];

	proj_GetPower2Size( tex_width, tex_height, adj_width, adj_height );

	//
	//	Allocate memory for the "class" texture map.
	//	This will never happen, since the texture is
	//	initialized to be the size of the screen. It is
	//	here for completeness ...
	//
	tmp_size = adj_width*adj_height*4;

	if ( tmp_size > TextureSize )
		{
		fprintf(stderr,"CT: allocating bigger texture.\n");

		TextureSize = tmp_size;

		if ( Texture )
			delete [] Texture;

		Texture = new GLubyte [ TextureSize ];
		}


	//++++++++++++++++++++++++++++++++++++++++++++++++++
	//
	//	Grab the drawn image as a texture map.
	//
	//
	//	Only read the "valid" part of the screen.
	//	This call tells GL that the texture we will
	//	define later is a bigger "power of 2" thing,
	//	and to read the pixels from the screen into
	//	memory as though we are reading a sub-rectangle
	//	of that "soon to be" texture.
	//
	//	This allows us to skip an expensive pixel
	//	copying operation from the old "small" texture
	//	to the "real" "power of 2" texture.
	//
	//++++++++++++++++++++++++++++++++++++++++++++++++++


	//
	//	Grab
	//
	//		Bad fix here, copying adj_width, adj_height
	//		when I really only need to copy something like:
	//		tex_width+5, tex_height+5.
	//
	//		Grab 10% extra - leave "outer" space garbage
	//
	if( take_timings )	gettimeofday( &pr.start, &pr.zone );

//	int fudge_width  = tex_width  + tex_width/10;
//	int fudge_height = tex_height + tex_height/10;

//		glPixelStorei( GL_PACK_ROW_LENGTH, adj_width );

		glReadPixels(	wbox[0][0], wbox[0][1],
						adj_width, adj_height,
						GL_RGBA, GL_UNSIGNED_BYTE,
						Texture );

	if( take_timings )	gettimeofday( &pr.end, &pr.zone );


	//
	//	Hand the "power of 2" texture off to GL
	//
	if( take_timings )	gettimeofday( &td.start, &td.zone );

//	glEnable( GL_TEXTURE_2D );

		if( textureValid )
			{
			if( glIsList( dispListID ) == GL_FALSE )
				{
				fprintf(stderr,"Try to delete an invalid List!!!\n");
				return 0;
				}
//			fprintf(stderr,"Deleting List -- %d\n",dispListID);
			glDeleteLists( dispListID, 1 );
			dispListID = glGenLists( 1 );
			}
		else
			{
			textureValid = 1;
			dispListID = glGenLists( 1 );
			}

		if( glIsEnabled( GL_TEXTURE_2D ) == GL_FALSE )
			{
			fprintf(stderr,"Texturing not Enabled!!!\n");
			return 0;
			}

//		fprintf(stderr,"Creating List -- %d\n",dispListID);

		glNewList( dispListID, GL_COMPILE );

		glTexImage2D( GL_TEXTURE_2D, 0, 4, adj_width, adj_height,
					  0, GL_RGBA, GL_UNSIGNED_BYTE, Texture );

		glEndList();

//	glDisable( GL_TEXTURE_2D );

	if( take_timings )	gettimeofday( &td.end, &td.zone );

	//
	//	Check for GL errors
	//
#ifdef DEBUG_GL
	{
	GLenum error_code;
		while((error_code=glGetError()) != GL_NO_ERROR)
			{
			fprintf(stderr,"\t%03d Billboard::createTexture\tgl error: %s\n",
				dispListID, gluErrorString(error_code));
			}
	}
#endif

	double smax,tmax;

	smax = 0.996*((double) tex_width  / (double) adj_width );
	tmax = 0.996*((double) tex_height / (double) adj_height );
	fprintf(stderr,"%f %f\n", smax,tmax );

	m_tc[0][0] = 0.0;
	m_tc[0][1] = 0.0;
	m_tc[1][0] = smax;
	m_tc[1][1] = 0.0;
	m_tc[2][0] = smax;
	m_tc[2][1] = tmax;
	m_tc[3][0] = 0.0;
	m_tc[3][1] = tmax;

	newTexture = 1;

	if( take_timings )	gettimeofday( &ct.end, &ct.zone );

	return( adj_width*adj_height );
	}
