//============================================================================
//
//		BoundBox.c
//
//
//			Jonathan Shade
//
//============================================================================

#include <math.h>
#include <stdio.h>
#include "Geometry.h"
#include "BoundBox.h"

#define BBOX_LINE_WIDTH 2.0

#define BBEPS (1000.0)

//----------------------------------------------------------------------------
//
//	Constructor
//
//----------------------------------------------------------------------------
BoundBox::BoundBox()
	{
	//
	//	This is redundant since Point3d defaults to zero.
	//
	min.setValue( 0.0, 0.0, 0.0 );
	max.setValue( 0.0, 0.0, 0.0 );
	xeps = yeps = zeps = 0.0;
	}

BoundBox::BoundBox ( const Point3d & p1, const Point3d & p2 )
	{
	min = p1;
	max = p2;
	xeps = fabs(max[0] - min[0]) / BBEPS;
	yeps = fabs(max[1] - min[1]) / BBEPS;
	zeps = fabs(max[2] - min[2]) / BBEPS;
	}

BoundBox::BoundBox ( const Point3f & p1, const Point3f & p2 )
	{
	min.setValue( (double) p1[0], (double) p1[1], (double) p1[2] );
	max.setValue( (double) p2[0], (double) p2[1], (double) p2[2] );
	xeps = fabs(max[0] - min[0]) / BBEPS;
	yeps = fabs(max[1] - min[1]) / BBEPS;
	zeps = fabs(max[2] - min[2]) / BBEPS;
	}

//----------------------------------------------------------------------------
//
//	Destructor
//
//----------------------------------------------------------------------------
BoundBox::~BoundBox()
	{
	}

//----------------------------------------------------------------------------
//
//		setValue
//
//----------------------------------------------------------------------------
void	BoundBox::setValue ( const Point3d & p1, const Point3d & p2 )
	{
	min = p1;
	max = p2;
	xeps = fabs(max[0] - min[0]) / BBEPS;
	yeps = fabs(max[1] - min[1]) / BBEPS;
	zeps = fabs(max[2] - min[2]) / BBEPS;
	}

void	BoundBox::setValue ( const Point3f & p1, const Point3f & p2 )
	{
	min.setValue( (double) p1[0], (double) p1[1], (double) p1[2] );
	max.setValue( (double) p2[0], (double) p2[1], (double) p2[2] );
	xeps = fabs(max[0] - min[0]) / BBEPS;
	yeps = fabs(max[1] - min[1]) / BBEPS;
	zeps = fabs(max[2] - min[2]) / BBEPS;
	}

void	BoundBox::getValue ( Point3d &p1, Point3d &p2 )
	{
	p1 = min;
	p2 = max;
	}

//----------------------------------------------------------------------------
//
//		pointInBox
//
//----------------------------------------------------------------------------
int		BoundBox::pointInBox ( const Point3d & p )
	{
	if(    ( p[0] >= min[0] && p[0] <= max[0] )
		&& ( p[1] >= min[1] && p[1] <= max[1] )
		&& ( p[2] >= min[2] && p[2] <= max[2] ) )
		{
		return 1;
		}
	else
		return 0;
	}


//----------------------------------------------------------------------------
//
//		intersect
//
//       o     -- ray origin
//       d     -- ray direction
//       _tmin -- minimum t parameter of ray at intersection
//			 	  (input/output parameter)
//       _tmax -- maximum t parameter of ray at intersection
//				  (input/output parameter)
//
//    NOTE: set _tmin to 0, _tmax to HUGE to find entire region where ray
//			is inside box.
//
//    Returns 1 if ray intersects box; 0 if not.
//
//----------------------------------------------------------------------------
int
BoundBox::intersect	( const Point3d &o, const Point3d &d,
						double & _tmin, double & _tmax )
	{
    double tlo,thi;
    double tmin,tmax;

    tmin = _tmin;
    tmax = _tmax;

    for (int i = 0; i < 3; i++)
		{
		if (d[i] > 0)
			{
            tlo = (min[i] - o[i])/d[i];
            thi = (max[i] - o[i])/d[i];
			}
		else if (d[i] < 0)
			{
            thi = (min[i] - o[i])/d[i];
            tlo = (max[i] - o[i])/d[i];
			}
		else
			{
            if (o[i] < min[i] || o[i] > max[i]) return 0;
            continue;
			}

        if (tlo > tmin) tmin = tlo;
        if (thi < tmax) tmax = thi;
        if (tmax < tmin) return 0;
		}

    _tmin = tmin;
    _tmax = tmax;

    return 1;
	}

//----------------------------------------------------------------------------
//
//		intersect
//
//----------------------------------------------------------------------------
int
BoundBox::intersect( const BoundBox & bbox )
	{
	if( bbox.min[0] > max[0] )	return 0;
	if( bbox.max[0] < min[0] )	return 0;

	if( bbox.min[1] > max[1] )	return 0;
	if( bbox.max[1] < min[1] )	return 0;

	if( bbox.min[2] > max[2] )	return 0;
	if( bbox.max[2] < min[2] )	return 0;

	if( bbox.min[0] >= min[0] &&
		bbox.max[0] <= max[0] &&
		bbox.min[1] >= min[1] &&
		bbox.max[1] <= max[1] &&
		bbox.min[2] >= min[2] &&
		bbox.max[2] <= max[2] )		return 2;

	return 1;
	}

//----------------------------------------------------------------------------
//
//		intersection
//
//----------------------------------------------------------------------------
int
BoundBox::intersection( const BoundBox & bbox, BoundBox & int_bbox )
	{
	Point3d	new_min,new_max;

	//
	//	X
	//
	if( bbox.min[0] > max[0] )	return 0;
	new_min[0] = ( bbox.min[0] > min[0] ) ? bbox.min[0] : min[0];

	if( bbox.max[0] < min[0] )	return 0;
	new_max[0] = ( bbox.max[0] < max[0] ) ? bbox.max[0] : max[0];
	

	//
	//	Y
	//
	if( bbox.min[1] > max[1] )	return 0;
	new_min[1] = ( bbox.min[1] > min[1] ) ? bbox.min[1] : min[1];

	if( bbox.max[1] < min[1] )	return 0;
	new_max[1] = ( bbox.max[1] < max[1] ) ? bbox.max[1] : max[1];

	//
	//	Z
	//
	if( bbox.min[2] > max[2] )	return 0;
	new_min[2] = ( bbox.min[2] > min[2] ) ? bbox.min[2] : min[2];

	if( bbox.max[2] < min[2] )	return 0;
	new_max[2] = ( bbox.max[2] < max[2] ) ? bbox.max[2] : max[2];


	int_bbox.setValue( new_min, new_max );

	return 1;
	}


//----------------------------------------------------------------------------
//
//		transform
//
//----------------------------------------------------------------------------
BoundBox	BoundBox::transform( const Matrix4d & matrix )
	{
	BoundBox	new_box = *this;
	Vector4d	p1,p2;

	p1.setValue( new_box.min, 1.0 );
	p2.setValue( new_box.max, 1.0 );

	p1 = matrix * p1;
	p2 = matrix * p2;

	new_box.min.setValue( p1[0], p1[1], p1[2] );
	new_box.max.setValue( p2[0], p2[1], p2[2] );

	return( new_box );
	}


//----------------------------------------------------------------------------
//
//		getRejectionPoints
//
//----------------------------------------------------------------------------
void
BoundBox::getRejectionPoints( const Normal3d & norm, Point3d &p1, Point3d &p2 ) const
	{
	for( int i=0 ; i<3 ; i++ )
		if( norm[i] >= 0.0)
			{
			p1[i] = max[i];
			p2[i] = min[i];
			}
		else
			{
			p1[i] = min[i];
			p2[i] = max[i];
			}

	return;
	}

//----------------------------------------------------------------------------
//
//		getMin
//
//----------------------------------------------------------------------------
Point3d	BoundBox::getMin() const
	{
	return min;
	}

//----------------------------------------------------------------------------
//
//		getMax
//
//----------------------------------------------------------------------------
Point3d	BoundBox::getMax() const
	{
	return max;
	}

//----------------------------------------------------------------------------
//
//		getCentroid
//
//----------------------------------------------------------------------------
Point3d	BoundBox::getCentroid () const
	{
	return( ((max - min)/2.0f + min) );
	}


//----------------------------------------------------------------------------
//
//	getRadius
//
//----------------------------------------------------------------------------
double	BoundBox::getRadius() const
	{
	return ( 0.5f * (max - min).length() );
	}


//----------------------------------------------------------------------------
//
//		getPoint
//
//----------------------------------------------------------------------------
Point3d	BoundBox::getPoint( int i ) const
	{
	Point3d	p;

	switch( i )
		{
		case 0:
			p.setValue( min[0], min[1], min[2] );
			break;

		case 1:
			p.setValue( max[0], min[1], min[2] );
			break;

		case 2:
			p.setValue( max[0], max[1], min[2] );
			break;

		case 3:
			p.setValue( min[0], max[1], min[2] );
			break;

		case 4:
			p.setValue( min[0], min[1], max[2] );
			break;

		case 5:
			p.setValue( max[0], min[1], max[2] );
			break;

		case 6:
			p.setValue( max[0], max[1], max[2] );
			break;

		case 7:
			p.setValue( min[0], max[1], max[2] );
			break;

		default:
			//
			//	Zero Point
			//
			break;
		};

	return p;
	}

//----------------------------------------------------------------------------
//
//		operator []
//
//----------------------------------------------------------------------------
double
BoundBox::operator[]( int i )
	{
	if( i<3 )	return min[i];
	else		return max[i-3];
	}
//----------------------------------------------------------------------------
//
//		superBox
//
//----------------------------------------------------------------------------
BoundBox
BoundBox::superBox( const BoundBox &b )
	{
	return BoundBox( min.lesserComponents(b.getMin()),
					 max.greaterComponents(b.getMax()) );
	}

//----------------------------------------------------------------------------
//
//		BoundBox = Matrix4d * BoundBox
//
//----------------------------------------------------------------------------
BoundBox operator * ( const Matrix4d &m, const BoundBox &bbox )
	{
	Point4d min4( bbox.min, 1 );
	Point4d max4( bbox.max, 1 );

	Point3d p,q;
	BoundBox b;

	min4 = m * min4;
	max4 = m * max4;

	p.setValue(min4[0],min4[1],min4[2]);
	q.setValue(max4[0],max4[1],max4[2]);
	b.setValue(p,q);
	return b;
	}


//----------------------------------------------------------------------------
//
//	Draw
//
//----------------------------------------------------------------------------
void	BoundBox::Draw_solid( GLdouble color[3] )
	{
	glEnable( GL_LIGHTING );

	glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
	glEnable(GL_COLOR_MATERIAL);

	glColor3dv( color );

	glBegin( GL_QUADS );

		//
		//	Square in the XZ plane, Y min
		//
		glNormal3d( 0, -1, 0 );
		glVertex3d( min[0], min[1], min[2] );
		glVertex3d( max[0], min[1], min[2] );
		glVertex3d( max[0], min[1], max[2] );
		glVertex3d( min[0], min[1], max[2] );

		//
		//	Square in the XZ plane, Y max
		//
		glNormal3d( 0, 1, 0 );
		glVertex3d( min[0], max[1], max[2] );
		glVertex3d( max[0], max[1], max[2] );
		glVertex3d( max[0], max[1], min[2] );
		glVertex3d( min[0], max[1], min[2] );


		//
		//	Square in the YZ plane, X min
		//
		glNormal3d( -1, 0, 0 );
		glVertex3d( min[0], min[1], min[2] );
		glVertex3d( min[0], min[1], max[2] );
		glVertex3d( min[0], max[1], max[2] );
		glVertex3d( min[0], max[1], min[2] );


		//
		//	Square in the YZ plane, X max
		//
		glNormal3d( 1, 0, 0 );
		glVertex3d( max[0], max[1], min[2] );
		glVertex3d( max[0], max[1], max[2] );
		glVertex3d( max[0], min[1], max[2] );
		glVertex3d( max[0], min[1], min[2] );


		//
		//	Square in the XY plane, Z min
		//
		glNormal3d( 0, 0, -1 );
		glVertex3d( min[0], max[1], min[2] );
		glVertex3d( max[0], max[1], min[2] );
		glVertex3d( max[0], min[1], min[2] );
		glVertex3d( min[0], min[1], min[2] );

		//
		//	Square in the XY plane, Z max
		//
		glNormal3d( 0, 0, 1 );
		glVertex3d( min[0], min[1], max[2] );
		glVertex3d( max[0], min[1], max[2] );
		glVertex3d( max[0], max[1], max[2] );
		glVertex3d( min[0], max[1], max[2] );

	glEnd();

	glDisable( GL_LIGHTING );

	glLineWidth( 2.0 );
	glColor3d( 0.0, 0.0, 0.0 );
	glLineWidth( 1.0 );
	Draw();	
	}

//----------------------------------------------------------------------------
//
//	Draw
//
//----------------------------------------------------------------------------
void	BoundBox::Draw( GLdouble color[3] )
	{
	glColor3dv( color );

//	glLineWidth( BBOX_LINE_WIDTH );
	glBegin( GL_LINES );

		//
		//	Square in the XZ plane, Y min
		//
		glVertex3d( min[0], min[1], min[2] );
		glVertex3d( max[0], min[1], min[2] );

		glVertex3d( max[0], min[1], min[2] );
		glVertex3d( max[0], min[1], max[2] );

		glVertex3d( max[0], min[1], max[2] );
		glVertex3d( min[0], min[1], max[2] );

		glVertex3d( min[0], min[1], max[2] );
		glVertex3d( min[0], min[1], min[2] );

		//
		//	Square in the XZ plane, Y max
		//
		glVertex3d( min[0], max[1], min[2] );
		glVertex3d( max[0], max[1], min[2] );

		glVertex3d( max[0], max[1], min[2] );
		glVertex3d( max[0], max[1], max[2] );

		glVertex3d( max[0], max[1], max[2] );
		glVertex3d( min[0], max[1], max[2] );

		glVertex3d( min[0], max[1], max[2] );
		glVertex3d( min[0], max[1], min[2] );


		//	
		//	Lines connecting the vertices of the squares
		//
		glVertex3d( min[0], min[1], min[2] );
		glVertex3d( min[0], max[1], min[2] );
		
		glVertex3d( max[0], min[1], min[2] );
		glVertex3d( max[0], max[1], min[2] );

		glVertex3d( max[0], min[1], max[2] );
		glVertex3d( max[0], max[1], max[2] );

		glVertex3d( min[0], min[1], max[2] );
		glVertex3d( min[0], max[1], max[2] );

	glEnd();
//	glLineWidth( 1.0 );
	}

//----------------------------------------------------------------------------
//
//	Draw
//
//----------------------------------------------------------------------------
void	BoundBox::Draw()
	{
//	glLineWidth( BBOX_LINE_WIDTH );
	glBegin( GL_LINES );

		//
		//	X axis lines
		//
//		glColor3d( 0.9, 0.4, 0.0 );

		glVertex3d( min[0], min[1], min[2] );
		glVertex3d( max[0], min[1], min[2] );

		glVertex3d( min[0], min[1], max[2] );
		glVertex3d( max[0], min[1], max[2] );

		glVertex3d( min[0], max[1], min[2] );
		glVertex3d( max[0], max[1], min[2] );

		glVertex3d( min[0], max[1], max[2] );
		glVertex3d( max[0], max[1], max[2] );


		//
		//	Z axis lines
		//
//		glColor3d( 0.7, 0.0, 0.4 );

		glVertex3d( min[0], min[1], min[2] );
		glVertex3d( min[0], min[1], max[2] );

		glVertex3d( max[0], min[1], min[2] );
		glVertex3d( max[0], min[1], max[2] );

		glVertex3d( min[0], max[1], min[2] );
		glVertex3d( min[0], max[1], max[2] );

		glVertex3d( max[0], max[1], min[2] );
		glVertex3d( max[0], max[1], max[2] );


		//	
		//	Y axis lines
		//
//		glColor3d( 0.2, 0.3, 0.8 );

		glVertex3d( min[0], min[1], min[2] );
		glVertex3d( min[0], max[1], min[2] );
		
		glVertex3d( max[0], min[1], min[2] );
		glVertex3d( max[0], max[1], min[2] );

		glVertex3d( max[0], min[1], max[2] );
		glVertex3d( max[0], max[1], max[2] );

		glVertex3d( min[0], min[1], max[2] );
		glVertex3d( min[0], max[1], max[2] );

	glEnd();
//	glLineWidth( 1.0 );
	}


//----------------------------------------------------------------------------
//
//	DrawRadiusFootprint
//
//----------------------------------------------------------------------------
void	BoundBox::DrawRadiusFootprint()
	{
	int	i;
	double x,y,z;

	Point3d min2d  = min;
	Point3d max2d  = max;

	// factor out the y
	min2d[1] = 0.0;
	max2d[1] = 0.0;

	Point3d center = ( max2d + min2d ) / 2.0f;

	double step = 2*M_PI/20.0;
	double radius = ( center - min2d ).length();
	
	x = y = z = 0.0;

//	glLineWidth( BBOX_LINE_WIDTH );
	glColor3d( 0.0, 0.0, 0.0 );

	glBegin( GL_LINE_LOOP );

	for( i=0 ; i<20 ; i++ )
		{
		x = radius*cos( i*step );
		z = radius*sin( i*step );

		glVertex3d( center[0]+x, center[1]+y, center[2]+z );
		}
	
	glEnd();
	
//	glLineWidth( 1.0 );

	return;
	}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
void
BoundBox::setGLClipPlanes()
{
	double C[4];

	C[0] =  1; C[1] =  0; C[2] =  0; C[3] = -min[0];	// Left
	glClipPlane( GL_CLIP_PLANE0, C );
	glEnable( GL_CLIP_PLANE0 );

	C[0] = -1; C[1] =  0; C[2] =  0; C[3] =  max[0];	// Right
	glClipPlane( GL_CLIP_PLANE1, C );
	glEnable( GL_CLIP_PLANE1 );

	C[0] =  0; C[1] =  1; C[2] =  0; C[3] = -min[1];	// Bottom
	glClipPlane( GL_CLIP_PLANE2, C );
	glEnable( GL_CLIP_PLANE2 );


	C[0] =  0; C[1] = -1; C[2] =  0; C[3] =  max[1];	// Top
	glClipPlane( GL_CLIP_PLANE3, C );
	glEnable( GL_CLIP_PLANE3 );

	C[0] =  0; C[1] =  0; C[2] =  1; C[3] = -min[2];	// Back
	glClipPlane( GL_CLIP_PLANE4, C );
	glEnable( GL_CLIP_PLANE4 );

	C[0] =  0; C[1] =  0; C[2] = -1; C[3] =  max[2];	// Front
	glClipPlane( GL_CLIP_PLANE5, C );
	glEnable( GL_CLIP_PLANE5 );

/*
	fprintf(stderr," %f %f %f %f %f %f\n",
		-min[0], max[0],
		-min[1], max[1],
		-min[2], max[2] );
		*/
}
void
BoundBox::setGLClipPlanesEps()
{
	double C[4];

	C[0] =  1; C[1] =  0; C[2] =  0; C[3] = -(min[0]-xeps);	// Left
	glClipPlane( GL_CLIP_PLANE0, C );
	glEnable( GL_CLIP_PLANE0 );

	C[0] = -1; C[1] =  0; C[2] =  0; C[3] =  (max[0]+xeps);	// Right
	glClipPlane( GL_CLIP_PLANE1, C );
	glEnable( GL_CLIP_PLANE1 );

	C[0] =  0; C[1] =  1; C[2] =  0; C[3] = -(min[1]-yeps);	// Bottom
	glClipPlane( GL_CLIP_PLANE2, C );
	glEnable( GL_CLIP_PLANE2 );


	C[0] =  0; C[1] = -1; C[2] =  0; C[3] =  (max[1]+yeps);	// Top
	glClipPlane( GL_CLIP_PLANE3, C );
	glEnable( GL_CLIP_PLANE3 );

	C[0] =  0; C[1] =  0; C[2] =  1; C[3] = -(min[2]-zeps);	// Back
	glClipPlane( GL_CLIP_PLANE4, C );
	glEnable( GL_CLIP_PLANE4 );

	C[0] =  0; C[1] =  0; C[2] = -1; C[3] =  (max[2]+zeps);	// Front
	glClipPlane( GL_CLIP_PLANE5, C );
	glEnable( GL_CLIP_PLANE5 );

	/*
	fprintf(stderr," %f %f %f %f %f %f\n",
		-(min[0]-xeps), (max[0]+xeps),
		-(min[1]-yeps), (max[1]+yeps),
		-(min[2]-zeps), (max[2]+zeps) );
	*/
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
void
BoundBox::unsetGLClipPlanes()
{
	glDisable( GL_CLIP_PLANE0 );
	glDisable( GL_CLIP_PLANE1 );
	glDisable( GL_CLIP_PLANE2 );
	glDisable( GL_CLIP_PLANE3 );
	glDisable( GL_CLIP_PLANE4 );
	glDisable( GL_CLIP_PLANE5 );
}

//----------------------------------------------------------------------------
//
//	Print
//
//----------------------------------------------------------------------------
void	BoundBox::Print() const
	{
	fprintf(stderr,"bbox: "); min.Print();
	fprintf(stderr,"\t");	max.Print();
	fprintf(stderr,"\n");
	}

