//
// Bspline.c
//

#include <stdio.h>
#include <iostream.h>
#include "Bspline.h"


static inline float 
affineCombination(float a, float b, float param)
{
  return (param * b) + (1.0 - param) * a;
}

inline float
evalBezier(float *cp, float param)
{
  // divide to 2 segs
  cp[0] = affineCombination(cp[0], cp[1], param);
  cp[1] = affineCombination(cp[1], cp[2], param);
  cp[2] = affineCombination(cp[2], cp[3], param);

  // divide to 1 seg
  //
  cp[0] = affineCombination(cp[0], cp[1], param);
  cp[1] = affineCombination(cp[1], cp[2], param);

  // divide to point
  return affineCombination(cp[0], cp[1], param);
}

inline float
evalBezier_tan(float *cp, float param)
{
  // divide to 2 segs
  cp[0] = affineCombination(cp[0], cp[1], param);
  cp[1] = affineCombination(cp[1], cp[2], param);
  cp[2] = affineCombination(cp[2], cp[3], param);

  // divide to 1 seg
  //
  //	cp[1] - cp[0] is the tangent vector to
  //	the bspline curve
  //
  cp[0] = affineCombination(cp[0], cp[1], param);
  cp[1] = affineCombination(cp[1], cp[2], param);

  return ( cp[1] - cp[0] );
}


Bspline::Bspline( char *filename )
{
  FILE *file = fopen( filename, "r" );

  if (!file)
    {
      _dims = 0;
      _ncp  = 0;
      _cp   = NULL;

      cerr << "warning: cannot open file for read: " << filename << endl;
      return;
    }

  fscanf(file, "%d", &_dims);
  fscanf(file, "%d", &_ncp);

  _len  = _ncp * _dims;
  _nseg = _ncp - 3;

  if (_len <= 0)
    {
      cerr << "warning: bad contents of file " << filename << endl;
      return;
    }

  _cp = new float[_len];

  for (int i = 0; i < _ncp; i++)
    for (int d = 0; d < _dims; d++)
      {
	fscanf(file, "%f", &_cp[d*_ncp+i]);
      }
}


Bspline::Bspline(float *cp, int ncp, int dims )
{
  _ncp  = ncp;
  _dims = dims;

  _len = _ncp * _dims;
  _nseg= _ncp - 3;

  _cp = new float[_len];

  int index = 0;
  for (int d = 0; d < _dims; d++)
    for (int i = 0; i < _ncp; i++)
      {
	_cp[d * ncp + i] = cp[index];
	index++;
      }
}


Bspline::~Bspline( void )
{
  if (_cp) delete _cp;
}

int
Bspline::get_dim ()
{
	return( _dims );
}

int
Bspline::get_numControlPoints()
{
	return( _ncp );
}

float
Bspline::get_cp	( int i, int dim )
{
	return( _cp[ dim*_ncp + i ] );
}

float 
Bspline::eval( float param, int dim )
{
  if (param < 0.0 ||
      param > 1.0)
    {
      return 0.0;
    }

  int   seg    = (int) (param * _nseg);
  float bparam = param * _nseg - seg;

  if (param == 1.0)		// end point
    {
      seg    = _nseg - 1;
      bparam = 1.0;
    }

  float *cp = &_cp[dim * _ncp + seg];

  static float thirds[4];
  static float bezier[4];

  const float t1   = 1.0 / 3.0;
  const float t2   = 2.0 / 3.0;
  const float half = 0.5;

  thirds[0] = affineCombination(cp[0], cp[1], t2);
  thirds[1] = affineCombination(cp[1], cp[2], t1);
  thirds[2] = affineCombination(cp[1], cp[2], t2);
  thirds[3] = affineCombination(cp[2], cp[3], t1);

  bezier[0] = affineCombination(thirds[0], thirds[1], half);
  bezier[1] = thirds[1];
  bezier[2] = thirds[2];
  bezier[3] = affineCombination(thirds[2], thirds[3], half);

  return evalBezier(bezier, bparam);
}

float 
Bspline::eval_tan( float param, int dim )
{
  if (param < 0.0 ||
      param > 1.0)
    {
      return 0.0;
    }

  int   seg    = (int) (param * _nseg);
  float bparam = param * _nseg - seg;

  if (param == 1.0)		// end point
    {
      seg    = _nseg - 1;
      bparam = 1.0;
    }

  float *cp = &_cp[dim * _ncp + seg];

  static float thirds[4];
  static float bezier[4];

  const float t1   = 1.0 / 3.0;
  const float t2   = 2.0 / 3.0;
  const float half = 0.5;

  thirds[0] = affineCombination(cp[0], cp[1], t2);
  thirds[1] = affineCombination(cp[1], cp[2], t1);
  thirds[2] = affineCombination(cp[1], cp[2], t2);
  thirds[3] = affineCombination(cp[2], cp[3], t1);

  bezier[0] = affineCombination(thirds[0], thirds[1], half);
  bezier[1] = thirds[1];
  bezier[2] = thirds[2];
  bezier[3] = affineCombination(thirds[2], thirds[3], half);

  return evalBezier_tan(bezier, bparam);
}


void  
Bspline::eval( float param, float *results)
{
  for (int d = 0; d < _dims; d++)
    {
      results[d] = eval( param, d );
    }
}

void  
Bspline::eval_tan( float param, float *results)
{
  for (int d = 0; d < _dims; d++)
    {
      results[d] = eval_tan( param, d );
    }
}


void  
Bspline::dump( void )
{
  for (int i = 0; i < _ncp; i++)
    {
      cout << i;
      for (int d = 0; d < _dims; d++)
	{
	  cout << " " << _cp[d * _ncp + i];
	}
      cout << endl;
    }
}

void  
Bspline::write( char *filename )
{
  FILE *file = fopen( filename, "w" );

  if (!file)
    {
      cerr << "warning: cannot open file for write: " << filename << endl;
      return;
    }

  fprintf(file, "%d\n", _dims);
  fprintf(file, "%d\n", _ncp);
  
  for (int d = 0; d < _dims; d++)
    for (int i = 0; i < _ncp; i++)
      {
	fprintf(file, "%f\n", _cp[d * _ncp + i]);
      }
}



//#define TESTING
#ifdef  TESTING

void
main()
{
  Bspline *spline = new Bspline( "trajectory.dat" );

  spline->dump();

  for (int i = 0; i <= 10; i ++)
    {
      float u = i * 0.1;
      cerr << i 
	   << " " << spline->eval(u,0) 
	   << " " << spline->eval(u,1) 
	   << " " << spline->eval(u,2) << endl;
    }

  delete spline;
}

#endif
