//============================================================================
//
//		Camera.c
//
//
//			Jonathan Shade
//
//============================================================================

#include <stdio.h>
#include <math.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "Camera.h"
#include "jitter.h"
#include "accpersp.h"

//----------------------------------------------------------------------------
//
//		Constructor
//
//----------------------------------------------------------------------------
Camera::Camera()
	{
	pos.setValue  ( 0.0,  0.0,  0.0 );
	look.setValue ( 0.0,  0.0, -1.0 );
	up.setValue   ( 0.0,  1.0,  0.0 );

	Reset();
	}


//----------------------------------------------------------------------------
//
//		Destructor
//
//----------------------------------------------------------------------------
Camera::~Camera()
	{
	}


//----------------------------------------------------------------------------
//
//		Reset Frustum Points
//
//----------------------------------------------------------------------------
void
Camera::resetFrustPoints()
	{
	//
	//  Near Plane points
	//
	frustPoints[0].setValue( frustum[0], frustum[3], -frustum[4] );//left,top
	frustPoints[1].setValue( frustum[0], frustum[2], -frustum[4] );//left,bot
	frustPoints[2].setValue( frustum[1], frustum[2], -frustum[4] );//right,bot
	frustPoints[3].setValue( frustum[1], frustum[3], -frustum[4] );//right,top

	//
	//  Far Plane points
	//
	frustPoints[4] = frustPoints[0] * frustum[5]/frustum[4]; // left,top
	frustPoints[5] = frustPoints[1] * frustum[5]/frustum[4]; // left,bot
	frustPoints[6] = frustPoints[2] * frustum[5]/frustum[4]; // right,bot
	frustPoints[7] = frustPoints[3] * frustum[5]/frustum[4]; // right,top
	}

//----------------------------------------------------------------------------
//
//		create_norm
//
//		Private function to create a normal from three points.
//
//		Specify in CCW order for right hand rule.
//
//----------------------------------------------------------------------------
Normal3d
create_norm( Point3d p, Point3d q, Point3d r)
	{
	Vector3d u = p - q;
	Vector3d v = r - q;
	
	return( u.cross(v) );
	}

//----------------------------------------------------------------------------
//
//		Reset Frustum Normals
//
//		Front, Left, Bottom, Right, Top, Far
//
//----------------------------------------------------------------------------
void
Camera::resetFrustNorms( Normal3d norms[6], Point3d points[8] )
	{
	//  Near Plane normal
	norms[0] = create_norm(points[1],points[2],points[3]);

	//	Left Plane normal
	norms[1] = create_norm(points[1],points[0],points[4]);

	//	Bottom Plane normal
	norms[2] = create_norm(points[2],points[1],points[5]);

	//	Right Plane normal
	norms[3] = create_norm(points[7],points[3],points[2]);

	//	Top Plane normal
	norms[4] = create_norm(points[4],points[0],points[3]);

	//	Far Plane normal
	norms[5] = create_norm(points[7],points[6],points[5]);

	for( int i=0 ; i<6 ; i++)	norms[i].normalize();
	}
   

//----------------------------------------------------------------------------
//
//		initFrustum
//
//		frustum[i] = 
//		
//		0     1      2       3    4     5
//      ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//		left, right, bottom, top, near, far
//
//----------------------------------------------------------------------------
void
Camera::initFrustum( double Angle, double Aspect,
					 double Near, double Far )
	{
	angle	= Angle;
	aspect	= Aspect;
	near	= Near;
	far		= Far;

	// top
	frustum[3] =  near*tan((angle/2.0)*M_PI/180.0);

	// bottom
	frustum[2] = -frustum[3];

	// left
	frustum[0] = -frustum[3]*aspect;

	// right
	frustum[1] = -frustum[0];

	// near
	frustum[4] = near;

	// far
	frustum[5] = far;

	resetFrustPoints();
	}

//----------------------------------------------------------------------------
//
//		getFrustumParams
//
//----------------------------------------------------------------------------
void
Camera::getFrustumParams( double &Angle, double &Aspect,
						  double &Near,  double &Far ) const
	{
	Angle	= angle;
	Aspect	= aspect;
	Near	= near;
	Far		= far;
	}

//----------------------------------------------------------------------------
//
//		setViewport
//
//----------------------------------------------------------------------------
void
Camera::setViewport( int x, int y, int width, int height )
	{
	vx		= x;
	vy		= y;
	vwidth	= width;
	vheight	= height;
	}

//----------------------------------------------------------------------------
//
//		getViewport
//
//----------------------------------------------------------------------------
void
Camera::getViewport( int &x, int &y, int &width, int &height ) const
	{
	x		= vx;
	y		= vy;
	width	= vwidth;
	height	= vheight;
	}

//----------------------------------------------------------------------------
//
//		Reset
//
//----------------------------------------------------------------------------
void
Camera::Reset()
	{
	camTrans.Identity();

	curr_p = pos;
	curr_l = look;
	curr_u = up;

	for( int i=0 ; i<8 ; i++ )
		curr_fp[i] = frustPoints[i];	

	resetFrustNorms( frustNorms, curr_fp );

	modelPos.setValue ( 0.0, 0.0, 0.0 );
	modelTrans.Identity();

	quat[0] = 0.0;
	quat[1] = 0.0;
	quat[2] = 0.0;
	quat[3] = 1.0;
	}

//----------------------------------------------------------------------------
//
//		setupGLProjectionMoveModel
//
//----------------------------------------------------------------------------
void
Camera::setupGLProjectionMoveModel() const
	{
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	glFrustum(  frustum[0], frustum[1], frustum[2],
				frustum[3], frustum[4], frustum[5]  );
	}

//----------------------------------------------------------------------------
//
//		setupGLProjectionCamera
//
//----------------------------------------------------------------------------
void
Camera::setupGLProjectionMoveCamera() const
	{
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( angle, aspect, near, far );
	}

//----------------------------------------------------------------------------
//
//		setupGLProjectionCamera
//
//		Meant to be used when rendering something temporarily
//		at higher Z resolution. 
//
//		!!!! The caller is responsible for the glPopMatrix() !!!!
//
//----------------------------------------------------------------------------
void
Camera::setupGLProjectionMoveCamera( double hither, double yon ) const
	{
	glMatrixMode( GL_PROJECTION );
	glPushMatrix();
		glLoadIdentity();
		gluPerspective( angle, aspect, hither, yon );
	}

//----------------------------------------------------------------------------
//
//		setupGLProjectionCameraAntiAlias
//
//----------------------------------------------------------------------------
void
Camera::setupGLProjectionMoveCameraAntiAlias( int jitter_index ) const
	{
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();

	accPerspective( angle, aspect, near, far,
					j8[jitter_index].x, j8[jitter_index].y,
					0, 0, 1 );
	}

//----------------------------------------------------------------------------
//
//		setupGLModelViewMoveModel
//
//----------------------------------------------------------------------------
void
Camera::setupGLModelViewMoveModel() const
	{
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();

	//
	//	For walkthrus, this is correct
	//
	glMultMatrixd( modelTrans.getValue() );
	glTranslated( modelPos[0], modelPos[1], modelPos[2] );

	//
	//	For trackball, this is correct
	//
	//glTranslated( modelPos[0], modelPos[1], modelPos[2] );
	//glMultMatrixd( modelTrans.getValue() );
	}

//----------------------------------------------------------------------------
//
//		setupGLModelViewMoveCamera
//
//----------------------------------------------------------------------------
void
Camera::setupGLModelViewMoveCamera() const
	{
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();

	gluLookAt(	curr_p[0],  curr_p[1], curr_p[2],
				curr_p[0] + curr_l[0],
				curr_p[1] + curr_l[1],
				curr_p[2] + curr_l[2],
				curr_u[0],  curr_u[1], curr_u[2]	);

	}


//----------------------------------------------------------------------------
//
//		Place the Look Vector
//
//----------------------------------------------------------------------------
void
Camera::Look( double x, double y, double z)
	{
	curr_l[0] = x;
	curr_l[1] = y;
	curr_l[2] = z;

	curr_l.normalize();
	}

//----------------------------------------------------------------------------
//
//		Move by deltas
//
//----------------------------------------------------------------------------
void
Camera::Move( double x, double y, double z)
	{
	curr_p[0] += x;
	curr_p[1] += y;
	curr_p[2] += z;

	modelPos[0] -= x;
	modelPos[1] -= y;
	modelPos[2] -= z;
	}

//----------------------------------------------------------------------------
//
//		Place the Camera
//
//----------------------------------------------------------------------------
void
Camera::Place( double x, double y, double z)
	{
	curr_p[0] = x;
	curr_p[1] = y;
	curr_p[2] = z;

	modelPos[0] = -x;
	modelPos[1] = -y;
	modelPos[2] = -z;
	}

//----------------------------------------------------------------------------
//
//		Dolly by delta
//
//----------------------------------------------------------------------------
void
Camera::Dolly( double t )
	{
	curr_p	 += t*(curr_l);
	modelPos -= t*(curr_l);
	}


//----------------------------------------------------------------------------
//
//		Rotate view vector around the Y-axis by delta
//
//----------------------------------------------------------------------------
void
Camera::Rotate( double angle )
	{
//	modelTrans	= modelTrans.preConcatRotY( -angle );
//	camTrans	= camTrans.postConcatRotY( angle );
	modelTrans	= modelTrans.postConcatRotY( -angle );
	camTrans	= camTrans.preConcatRotY( angle );
	}

//----------------------------------------------------------------------------
//
//		Update
//
//----------------------------------------------------------------------------
void
Camera::Update()
	{
	curr_l = camTrans*look;
	curr_u = camTrans*up;
	
	for( int i=0 ; i<8 ; i++ )
		curr_fp[i] = (camTrans * frustPoints[i]) + curr_p; 

	resetFrustNorms( curr_fn, curr_fp );
	}

//----------------------------------------------------------------------------
//
//		inside_test8
//
//----------------------------------------------------------------------------
int
Camera::inside_test8( const Point3d & pos,   const Normal3d & norm, 
					  const BoundBox & bbox, const Matrix4d & matrix )
	{
	int	inside[8],inFlag;

	Vector3d p,q;
	Vector4d v;

	for( int i=0 ; i<8 ; i++ )
		{
		v.setValue( bbox.getPoint(i), 1.0 );
		v = matrix * v;
		p.setValue( v[0], v[1], v[2] );

		q = p - pos;
		q.normalize();
			
		if( q.dot( norm ) < 0.0 )	inside[i] = 0;
		else						inside[i] = 1;
		}

	inFlag = 1;
	for( i=0 ; i<8 ; i++ )
		if( inside[i] ) inFlag = 0;

	if( inFlag )
		return 0;
	else
		return 1;
	}

//----------------------------------------------------------------------------
//
//		inFrustumMoveModel
//
//----------------------------------------------------------------------------
int
Camera::inFrustumMoveModel( const BoundBox & bbox, const Matrix4d & m )
	{
	if( !inside_test8( frustPoints[2], frustNorms[0], bbox, m ) )	return 0;
	if( !inside_test8( frustPoints[0], frustNorms[1], bbox, m ) )	return 0;
	if( !inside_test8( frustPoints[1], frustNorms[2], bbox, m ) )	return 0;
	if( !inside_test8( frustPoints[3], frustNorms[3], bbox, m ) )	return 0;
	if( !inside_test8( frustPoints[0], frustNorms[4], bbox, m ) )	return 0;
	if( !inside_test8( frustPoints[6], frustNorms[5], bbox, m ) )	return 0;

	return 1;
	}

//----------------------------------------------------------------------------
//
//		inside_test1
//
//	   -1 - point is behind plane
//		0 - point straddles plane
//		1 - point is in front of plane
//
//----------------------------------------------------------------------------
int
Camera::inside_test1( const Point3d & pos, const Normal3d & norm, const BoundBox & bbox ) const
	{
	Vector3d p1,p2,q1,q2;

	bbox.getRejectionPoints( norm, p1, p2 );

	q1 = p1 - pos;
	q1.normalize();

	q2 = p2 - pos;
	q2.normalize();

	if( q1.dot( norm ) < 0.0 )		return -1;
	else if( q2.dot( norm ) < 0.0 )	return 0;
	else							return 1;
	}

//----------------------------------------------------------------------------
//
//		point_in_frontof_plane
//
//		0 - no
//		1 - yes
//
//----------------------------------------------------------------------------
int
point_infrontof_plane( const Point3d & pos, const Normal3d & norm, const Point3d & p )
	{
	Vector3d q;

	q = p - pos;
	q.normalize();

	if( q.dot( norm ) < 0.0 )	return 0;
	else						return 1;
	}

//----------------------------------------------------------------------------
//
//		inFrustumMoveCamera
//
//		p[0-6] are the points in the bbox "nearest" each
//		of the planes of the frustum.
//
//		This is not quite robust, there is still a problem with
//		boxes behind the viewpoint.
//
//----------------------------------------------------------------------------
int
Camera::inFrustumMoveCamera( const BoundBox & bbox, int r[6], int f[5] ) const
	{
	Point3d p[6];
	
	if( (r[0]=inside_test1( curr_fp[2], curr_fn[0], bbox )) == -1) return 0; // front
	if( (r[1]=inside_test1( curr_fp[0], curr_fn[1], bbox )) == -1) return 0; // left
	if( (r[2]=inside_test1( curr_fp[3], curr_fn[3], bbox )) == -1) return 0; // right
	if( (r[3]=inside_test1( curr_fp[1], curr_fn[2], bbox )) == -1) return 0; // bot
	if( (r[4]=inside_test1( curr_fp[0], curr_fn[4], bbox )) == -1) return 0; // top
//	if( (r[5]=inside_test1( curr_fp[6], curr_fn[5], bbox )) == -1) return 0; // far

 	//
	//	In front of the front clipping plane, must be ok !
	//
	if( r[0] == 1 )	return 1;

return 1;

	//
	//	Test Straddle Left  -> In front of Right
	//				  Right ->             Left
	//
	if( r[1] == 0 )
		{
		bbox.getRejectionPoints( curr_fn[1], p[0], p[1] );

		if( f[1]=point_infrontof_plane(curr_fp[3], curr_fn[3], p[0]) )
			return 1;
		}
	if( r[2] == 0 )
		{
		bbox.getRejectionPoints( curr_fn[3], p[0], p[1] );

		if( f[0]=point_infrontof_plane(curr_fp[0], curr_fn[1], p[0]) )
			return 1;
		}

//	if( (r[1]==0) || (r[2]==0) )	return 0;

	//
	//	Test Straddle Bot -> In front of Top
	//				  Top ->             Bot
	//
	if ( r[3] == 0 )
		{
		bbox.getRejectionPoints( curr_fn[2], p[0], p[1] );
		
		if( f[2]=point_infrontof_plane(curr_fp[0], curr_fn[4], p[0]) )
			return 1;
		} 
	if( r[4] == 0 )
		{
		bbox.getRejectionPoints( curr_fn[4], p[0], p[1] );

		if( f[3]=point_infrontof_plane(curr_fp[1], curr_fn[2], p[0]) )
			return 1;
		}

	return 0;

	}

//----------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------
int
Camera::point_in_frustum( const Point3d & p, int r[6] )
	{
	for( int i=0 ; i<6 ; i++)	r[i] = -1;

	if( !(r[0]=point_infrontof_plane( curr_fp[2], curr_fn[0], p )) )	return 0; // front
	if( !(r[1]=point_infrontof_plane( curr_fp[1], curr_fn[2], p )) )	return 0; // left
	if( !(r[2]=point_infrontof_plane( curr_fp[3], curr_fn[3], p )) )	return 0; // right
	if( !(r[3]=point_infrontof_plane( curr_fp[0], curr_fn[1], p )) )	return 0; // bot
	if( !(r[4]=point_infrontof_plane( curr_fp[0], curr_fn[4], p )) )	return 0; // top
//	if( !(r[5]=point_infrontof_plane( curr_fp[6], curr_fn[5], p )) )	return 0; // far

	return 1;
	}

//
//	Dont use this method currently
//
int
Camera::inFrustumMoveCamera( const Point3d & beg, const Point3d & end )
	{
	int	r1[6],r2[6];

	if( point_in_frustum( beg, r1 ) ) return 1;
	if( point_in_frustum( end, r2 ) ) return 1;

	//	
	//	Loose check on straddling the frustum -- DOESNT WORK
	//
	if( (r1[1]==1)&&( (r1[0]==1)||(r2[0]==1) ) )
		{
		if( point_infrontof_plane( curr_fp[3], curr_fn[3], end ) )
		return 1;
		}

	if( (r2[1]==1)&&( (r1[0]==1)||(r2[0]==1) ) )
		{
		if( point_infrontof_plane( curr_fp[3], curr_fn[3], beg ) )
		return 1;
		}
 
	return 0;
	}



//----------------------------------------------------------------------------
//
//		getSummary
//
//----------------------------------------------------------------------------
void
Camera::getSummary( Point3d &p, Vector3d &l, Vector3d &u)
	{
	p = pos;
	l = look;
	u = up;
	}

//----------------------------------------------------------------------------
//
//		getFrustum
//
//----------------------------------------------------------------------------
double*
Camera::getFrustum() const
	{
	return (double *)frustum;
	}

//----------------------------------------------------------------------------
//
//		(i,j)
//
//		i = 0		near plane
//		i = 1		far  plane
//
//		j = 0		left,top
//		j = 1		left,bot
//		j = 2		right,bot
//		j = 3		right,top
//
//----------------------------------------------------------------------------
Point3d
Camera::operator () ( int i, int j )
	{
	return (curr_fp[i*4+j]);
	}

Point3d
Camera::operator [] ( int i )
	{
	return ( curr_fp[i] );
	}

//----------------------------------------------------------------------------
//
//		getQuat
//
//----------------------------------------------------------------------------
Vector4d
Camera::getQuat()
	{
	return quat;
	}

//----------------------------------------------------------------------------
//
//		setQuat
//
//----------------------------------------------------------------------------
void
Camera::setQuat( const Vector4d q )
	{
	quat = q;
	}

//----------------------------------------------------------------------------
//
//		getCameraTransform
//
//----------------------------------------------------------------------------
const double*
Camera::getCameraTransform()
	{
	return camTrans.getValue();
	}

//----------------------------------------------------------------------------
//
//		setCameraTransform
//
//----------------------------------------------------------------------------
void
Camera::setCameraTransform( Matrix4d matrix )
	{
	camTrans = matrix;
	}

//----------------------------------------------------------------------------
//
//		getModelTransform
//
//----------------------------------------------------------------------------
const double*
Camera::getModelTransform()
	{
	return modelTrans.getValue();
	}

//----------------------------------------------------------------------------
//
//		setModelTransform
//
//----------------------------------------------------------------------------
void
Camera::setModelTransform( Matrix4d matrix )
	{
	modelTrans = matrix;
	}

//----------------------------------------------------------------------------
//
//		getModelPosition
//
//----------------------------------------------------------------------------
Point3d
Camera::getModelPosition() const
	{
	return( modelPos );
	}

//----------------------------------------------------------------------------
//
//		getPosition
//
//----------------------------------------------------------------------------
Point3d
Camera::getPosition() const
	{
	return( curr_p );
	}


//----------------------------------------------------------------------------
//
//		getLook
//
//----------------------------------------------------------------------------
Vector3d
Camera::getLook() const
	{
	return( curr_l );
	}


//----------------------------------------------------------------------------
//
//		getUp
//  
//----------------------------------------------------------------------------
Vector3d
Camera::getUp() const
	{
	return( curr_u );
	}	


//----------------------------------------------------------------------------
//
//		Print
//
//----------------------------------------------------------------------------
void
Camera::Print() const
	{
	fprintf( stderr, "\n" );

	fprintf( stderr, "Position    : " ); curr_p.Print(); 
	fprintf( stderr, "\n" );
	fprintf( stderr, "Look vec    : " ); curr_l.Print();
	fprintf( stderr, "\n" );
	fprintf( stderr, "Up vec      : " ); curr_u.Print();
	fprintf( stderr, "\n" );

	fprintf( stderr, "Quat        : %f\t%f\t%f\t%f\t%f\n",
		quat[0],quat[1],quat[2],quat[3],quat.length() );

	fprintf( stderr, "Camera Trans: %f\t%f\t%f\t%f\n", 
		camTrans[0],camTrans[1],camTrans[2],camTrans[3] );

	fprintf( stderr, "            : %f\t%f\t%f\t%f\n", 
		camTrans[4],camTrans[5],camTrans[6],camTrans[7] );

	fprintf( stderr, "            : %f\t%f\t%f\t%f\n", 
		camTrans[8],camTrans[9],camTrans[10],camTrans[11] );

	fprintf( stderr, "            : %f\t%f\t%f\t%f\n", 
		camTrans[12],camTrans[13],camTrans[14],camTrans[15] );

	fprintf( stderr, "\nPosition    : "); modelPos.Print(); 
	fprintf( stderr, "\n" );

	fprintf( stderr, "Model Trans: %f\t%f\t%f\t%f\n", 
		modelTrans[0],modelTrans[1],modelTrans[2],modelTrans[3] );

	fprintf( stderr, "           : %f\t%f\t%f\t%f\n", 
		modelTrans[4],modelTrans[5],modelTrans[6],modelTrans[7] );

	fprintf( stderr, "           : %f\t%f\t%f\t%f\n", 
		modelTrans[8],modelTrans[9],modelTrans[10],modelTrans[11] );

	fprintf( stderr, "           : %f\t%f\t%f\t%f\n", 
		modelTrans[12],modelTrans[13],modelTrans[14],modelTrans[15] );

	fprintf( stderr, "\n" );
	}



//----------------------------------------------------------------------------
//
//		Draw Norms
//
//----------------------------------------------------------------------------
void
Camera::DrawNorms()
	{
	double scale = 20.0;

	glColor3d( 0.0, 0.0, 0.0 );

	glBegin( GL_LINES );

	glVertex3dv( curr_fp[2].getValue() );
	glVertex3dv( (curr_fn[0]*scale + curr_fp[2]).getValue() );

	glVertex3dv( curr_fp[0].getValue() );
	glVertex3dv( (curr_fn[1]*scale + curr_fp[0]).getValue() );

//	glVertex3dv( curr_fp[1].getValue() );
//	glVertex3dv( (curr_fn[2]*scale + curr_fp[1]).getValue() );

//	glVertex3dv( curr_fp[3].getValue() );
//	glVertex3dv( (curr_fn[3]*scale + curr_fp[3]).getValue() );

//	glVertex3dv( curr_fp[0].getValue() );
//	glVertex3dv( (curr_fn[4]*scale + curr_fp[0]).getValue() );

//	glVertex3dv( curr_fp[6].getValue() );
//	glVertex3dv( (curr_fn[5]*scale + curr_fp[6]).getValue() );

	glEnd();
	
	}
