//============================================================================
//
//		Matrix4f
//
//
//		OpenGL transform vertices by premultiplication:
//
//		Mnew = Mold x Tnew ( the new matrix is premultiplied by the old matrix ).
//		Vn   = M x Vold	   ( the vertex is premultiplied by a matrix ).
//
//		Post multiplication goes the other way around.
//
//		In this file, preMethodName implies transformation by premultiplication.
//
//
//		Internal Representation:
//
//			Array of Floats:	0  4  8  12
//								1  5  9  13
//								2  6 10  14
//								3  7 11  15
//
//============================================================================

#include <stdio.h>
#include <math.h>
#include "Matrix4f.h"



//----------------------------------------------------------------------------
//
//		Constructors
//
//----------------------------------------------------------------------------
Matrix4f::Matrix4f()
	{
	Identity();
	}

Matrix4f::Matrix4f( const float v[16] )
	{
	setValue( v );
	}


//----------------------------------------------------------------------------
//
//		Reset the matrix to the Identity matrix
//
//----------------------------------------------------------------------------
void
Matrix4f::Identity()
	{
	m[0]  = 1.0;
	m[1]  = 0.0;
	m[2]  = 0.0;
	m[3]  = 0.0;

	m[4]  = 0.0;
	m[5]  = 1.0;
	m[6]  = 0.0;
	m[7]  = 0.0;

	m[8]  = 0.0;
	m[9]  = 0.0;
	m[10] = 1.0;
	m[11] = 0.0;

	m[12] = 0.0;
	m[13] = 0.0;
	m[14] = 0.0;
	m[15] = 1.0;
	}
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
Matrix4f
Matrix4f::Transpose()
	{
	Matrix4f	tmat;

	for( int i=0 ; i<4 ; i++ )
		for( int j=0 ; j<4 ; j++ )
			tmat( i, j ) = (*this)( j, i );

	return( tmat );
	}

//----------------------------------------------------------------------------
//
//		result = pre x this
//
//----------------------------------------------------------------------------
Matrix4f
Matrix4f::postConcat( const Matrix4f &post )
	{
	Matrix4f mat;

	mat(0,0) = post(0,0)*m[0]  + post(0,1)*m[1]  + post(0,2)*m[2]  + post(0,3)*m[3];
	mat(0,1) = post(0,0)*m[4]  + post(0,1)*m[5]  + post(0,2)*m[6]  + post(0,3)*m[7];
	mat(0,2) = post(0,0)*m[8]  + post(0,1)*m[9]  + post(0,2)*m[10] + post(0,3)*m[11];
	mat(0,3) = post(0,0)*m[12] + post(0,1)*m[13] + post(0,2)*m[14] + post(0,3)*m[15];

	mat(1,0) = post(1,0)*m[0]  + post(1,1)*m[1]  + post(1,2)*m[2]  + post(1,3)*m[3];
	mat(1,1) = post(1,0)*m[4]  + post(1,1)*m[5]  + post(1,2)*m[6]  + post(1,3)*m[7];
	mat(1,2) = post(1,0)*m[8]  + post(1,1)*m[9]  + post(1,2)*m[10] + post(1,3)*m[11];
	mat(1,3) = post(1,0)*m[12] + post(1,1)*m[13] + post(1,2)*m[14] + post(1,3)*m[15];

	mat(2,0) = post(2,0)*m[0]  + post(2,1)*m[1]  + post(2,2)*m[2]  + post(2,3)*m[3];
	mat(2,1) = post(2,0)*m[4]  + post(2,1)*m[5]  + post(2,2)*m[6]  + post(2,3)*m[7];
	mat(2,2) = post(2,0)*m[8]  + post(2,1)*m[9]  + post(2,2)*m[10] + post(2,3)*m[11];
	mat(2,3) = post(2,0)*m[12] + post(2,1)*m[13] + post(2,2)*m[14] + post(2,3)*m[15];

	mat(3,0) = post(3,0)*m[0]  + post(3,1)*m[1]  + post(3,2)*m[2]  + post(3,3)*m[3];
	mat(3,1) = post(3,0)*m[4]  + post(3,1)*m[5]  + post(3,2)*m[6]  + post(3,3)*m[7];
	mat(3,2) = post(3,0)*m[8]  + post(3,1)*m[9]  + post(3,2)*m[10] + post(3,3)*m[11];
	mat(3,3) = post(3,0)*m[12] + post(3,1)*m[13] + post(3,2)*m[14] + post(3,3)*m[15];

	return( mat );
	}

//----------------------------------------------------------------------------
//
//		result = this x pre
//
//----------------------------------------------------------------------------
Matrix4f
Matrix4f::preConcat( const Matrix4f &pre ) 
	{
	Matrix4f mat;

	mat(0,0) = m[0]*pre(0,0) + m[4]*pre(1,0) + m[8]*pre(2,0) + m[12]*pre(3,0);
	mat(0,1) = m[0]*pre(0,1) + m[4]*pre(1,1) + m[8]*pre(2,1) + m[12]*pre(3,1);
	mat(0,2) = m[0]*pre(0,2) + m[4]*pre(1,2) + m[8]*pre(2,2) + m[12]*pre(3,2);
	mat(0,3) = m[0]*pre(0,3) + m[4]*pre(1,3) + m[8]*pre(2,3) + m[12]*pre(3,3);

	mat(1,0) = m[1]*pre(0,0) + m[5]*pre(1,0) + m[9]*pre(2,0) + m[13]*pre(3,0);
	mat(1,1) = m[1]*pre(0,1) + m[5]*pre(1,1) + m[9]*pre(2,1) + m[13]*pre(3,1);
	mat(1,2) = m[1]*pre(0,2) + m[5]*pre(1,2) + m[9]*pre(2,2) + m[13]*pre(3,2);
	mat(1,3) = m[1]*pre(0,3) + m[5]*pre(1,3) + m[9]*pre(2,3) + m[13]*pre(3,3);

	mat(2,0) = m[2]*pre(0,0) + m[6]*pre(1,0) + m[10]*pre(2,0) + m[14]*pre(3,0);
	mat(2,1) = m[2]*pre(0,1) + m[6]*pre(1,1) + m[10]*pre(2,1) + m[14]*pre(3,1);
	mat(2,2) = m[2]*pre(0,2) + m[6]*pre(1,2) + m[10]*pre(2,2) + m[14]*pre(3,2);
	mat(2,3) = m[2]*pre(0,3) + m[6]*pre(1,3) + m[10]*pre(2,3) + m[14]*pre(3,3);

	mat(3,0) = m[3]*pre(0,0) + m[7]*pre(1,0) + m[11]*pre(2,0) + m[15]*pre(3,0);
	mat(3,1) = m[3]*pre(0,1) + m[7]*pre(1,1) + m[11]*pre(2,1) + m[15]*pre(3,1);
	mat(3,2) = m[3]*pre(0,2) + m[7]*pre(1,2) + m[11]*pre(2,2) + m[15]*pre(3,2);
	mat(3,3) = m[3]*pre(0,3) + m[7]*pre(1,3) + m[11]*pre(2,3) + m[15]*pre(3,3);

	return( mat );
	}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
Matrix4f
Matrix4f::postConcatRotY( float angle )
	{
	Matrix4f	&mat = *this;
	Matrix4f	new_mat;
	
	float c = cos( angle );
	float s = sin( angle );

	for( int i=0 ; i<4 ; i++ )
		{
		new_mat(0,i) = mat(0,i)*c + mat(2,i)*s;
		new_mat(1,i) = mat(1,i);
		new_mat(2,i) = mat(0,i)*(-s) + mat(2,i)*c;
		new_mat(3,i) = mat(3,i);
		}
	
	return( new_mat );
	}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
Matrix4f
Matrix4f::preConcatRotY( float angle )
	{
	Matrix4f	&mat = *this;
	Matrix4f	new_mat;
	
	float c = cos( angle );
	float s = sin( angle );

	for( int i=0 ; i<4 ; i++ )
		{
		new_mat(i,0) = mat(i,0)*c - mat(i,2)*s;
		new_mat(i,1) = mat(i,1);
		new_mat(i,2) = mat(i,0)*s + mat(i,2)*c;
		new_mat(i,3) = mat(i,3);
		}
	
	return( new_mat );
	}
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
Matrix4f
Matrix4f::postConcatTrans( float tx, float ty, float tz )
	{
	Matrix4f	&mat = *this;
	Matrix4f	new_mat;
	
	for( int i=0 ; i<4 ; i++ )
		{
		new_mat(0,i) = mat(0,i) + tx*mat(3,i);
		new_mat(1,i) = mat(1,i) + ty*mat(3,i);
		new_mat(2,i) = mat(2,i) + tz*mat(3,i);
		new_mat(3,i) = mat(3,i);
		}

	return( new_mat );	
	}
//----------------------------------------------------------------------------
//
//		result = this x trans
//
//----------------------------------------------------------------------------
Matrix4f
Matrix4f::preConcatTrans( float tx, float ty, float tz )
	{
	Matrix4f	&mat = *this;
	Matrix4f	new_mat;
	
	for( int i=0 ; i<4 ; i++ )
		{
		new_mat(i,0) = mat(i,0);
		new_mat(i,1) = mat(i,1);
		new_mat(i,2) = mat(i,2);

		new_mat(i,3) = tx*mat(i,0) + ty*mat(i,1) + tz*mat(i,2) + mat(i,3);
		}

	return( new_mat );	
	}
//----------------------------------------------------------------------------
//
//		result = scale x this
//
//----------------------------------------------------------------------------
Matrix4f
Matrix4f::postConcatScale( float sx, float sy, float sz )
	{
	Matrix4f	&mat = *this;
	Matrix4f	new_mat;
	
	for( int i=0 ; i<4 ; i++ )
		{
		new_mat(0,i) = sx*mat(0,i);
		new_mat(1,i) = sy*mat(1,i);
		new_mat(2,i) = sz*mat(2,i);
		new_mat(3,i) = mat(3,i);
		}

	return( new_mat );	
	}
//----------------------------------------------------------------------------
//
//		result = this x scale
//
//----------------------------------------------------------------------------
Matrix4f
Matrix4f::preConcatScale( float sx, float sy, float sz )
	{
	Matrix4f	&mat = *this;
	Matrix4f	new_mat;
	
	for( int i=0 ; i<4 ; i++ )
		{
		new_mat(i,0) = sx*mat(i,0);
		new_mat(i,1) = sy*mat(i,1);
		new_mat(i,2) = sz*mat(i,2);
		new_mat(i,3) = mat(i,3);
		}

	return( new_mat );	
	}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
Matrix4f &
Matrix4f::setValue( const float v[16] )
	{
	for( int i=0 ; i<16 ; i++ )
		m[i] = v[i];
	
	return *this;
	}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
const float*
Matrix4f::getValue() const
	{
	return m;
	}

float*
Matrix4f::getValue()
	{
	return m;
	}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
float &
Matrix4f::operator[](int i)
	{
	return( m[i] );
	}

const float &
Matrix4f::operator[](int i) const
	{
	return( m[i] );
	}

//----------------------------------------------------------------------------
//
//			( row, col )
//
//----------------------------------------------------------------------------
float &
Matrix4f::operator() (int i, int j) const
	{
	return( m[i + j*4] );
	}


//----------------------------------------------------------------------------
//
//		Print
//
//----------------------------------------------------------------------------
void
Matrix4f::Print()
	{
	for( int i=0 ; i<4 ; i++ )
		fprintf( stderr, "\t%6.2f %6.2f %6.2f %6.2f\n",
					m[i], m[i+4], m[i+8], m[i+12] );
	}



//----------------------------------------------------------------------------
//
//		read the matrix from a file.
//
//----------------------------------------------------------------------------
void
Matrix4f::Read( FILE * dbfile )
    {
    fscanf( dbfile, "%f %f %f %f", &m[0], &m[4], &m[8],  &m[12] );
    fscanf( dbfile, "%f %f %f %f", &m[1], &m[5], &m[9],  &m[13] );
    fscanf( dbfile, "%f %f %f %f", &m[2], &m[6], &m[10], &m[14] );
    fscanf( dbfile, "%f %f %f %f", &m[3], &m[7], &m[11], &m[15] );
    }

//----------------------------------------------------------------------------
//
//		Write the matrix to a file.
//
//----------------------------------------------------------------------------
void
Matrix4f::Write( FILE * dbfile )
    {
    fprintf( dbfile, "%f %f %f %f\n", m[0], m[4], m[8],  m[12] );
    fprintf( dbfile, "%f %f %f %f\n", m[1], m[5], m[9],  m[13] );
    fprintf( dbfile, "%f %f %f %f\n", m[2], m[6], m[10], m[14] );
    fprintf( dbfile, "%f %f %f %f\n", m[3], m[7], m[11], m[15] );
    }

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
Vector3f operator * ( const Matrix4f &m, const Vector3f &v )
	{
	Vector3f result;

	result[0] = m[0] * v[0] + m[4] * v[1] + m[8]  * v[2];
	result[1] = m[1] * v[0] + m[5] * v[1] + m[9]  * v[2];
	result[2] = m[2] * v[0] + m[6] * v[1] + m[10] * v[2];
	
	return( result );
	}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
Vector4f operator * ( const Matrix4f &m, const Vector4f &v )
	{
	Vector4f result;

	result[0] =  m[0]*v[0]  +  m[4]*v[1]  +  m[8] *v[2]  +  m[12]*v[3];
	result[1] =  m[1]*v[0]  +  m[5]*v[1]  +  m[9] *v[2]  +  m[13]*v[3];
	result[2] =  m[2]*v[0]  +  m[6]*v[1]  +  m[10]*v[2]  +  m[14]*v[3];
	result[3] =  m[3]*v[0]  +  m[7]*v[1]  +  m[11]*v[2]  +  m[15]*v[3];

	return( result );
	}

