//============================================================================
//
//	Metric.c
//
//
//============================================================================

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include "Metric.h"
#include "projection.h"

//#include "cost_re2.h"
//#include "cost_high_impact.h"
#include "cost_max_impact.h"

#define OVERSAMPLE 8

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//
//
//  Base (abstract) Metric Class
//
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Metric::Metric(void)
	{
	}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//
//  AngleDistance Metric
//
//
//		This metric calculates the angle between the new camera position,
//	the centroid of the bounding sphere, and the old camera position.
//	It also checks the ratio of the distance from the centroid to the
//	new camera position to that to the old camera position.
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
AngleDistanceMetric::AngleDistanceMetric(double at, double dr)
	{
	m_angleThresh = cos(at);
	m_distRatioThresh = dr;
	}

//----------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------
int AngleDistanceMetric::isA(MetricType t)
	{
	if ( t == AngleDistance )	return 1;
	else						return Metric::isA(t);
	}

//----------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------
int
AngleDistanceMetric::computeError(	const Point3d new_cam_pos,
									const BoundBox &bbox_world )
{
	Point4d		centroid;
	Point3d		cent3;
	Vector3d	v1,v2;

	double cosAngle, distError, oldLen, newLen, radius;

	numEvals++;

	//
	//		Move the model style
	//
	//	assume camera is at 0,0,0
//	v1 = model_pos;
//	v2 = old_pos;

	//
	//		Move the camera style
	//
	cent3 = bbox_world.getCentroid();
	v1 = new_cam_pos - cent3;
	v2 = m_old_cam_pos - cent3;

	// 
	// Calculate the angle between the two vectors and return false if
	// it is above the threshold.
	//
	//	Since the camparison is actually between cos(angle),cos(threshold),
	//	the test is reversed in polarity( cos is larger for smaller angles ).
	//	So the if the  cos(angle) is less than cos(thresh)
	//	then (angle > thresh).
	//
	newLen = v1.length();
	oldLen = v2.length();
  
	v1.normalize();
	v2.normalize();

	cosAngle = v1.dot( v2 );

//fprintf(stderr,"angle err: %f\t%f\n", cosAngle, angleThresh );
	if ( fabs(cosAngle) < m_angleThresh ) 
		{
		return 0;
		}

	//
	// Find the radius of the bounding sphere, and subtract it from
	// the magnitudes of the vectors above.  Check the ratio between
	// the new magnitudes and return false if it is above the threshold.
	//
	radius = bbox_world.getRadius();

	distError = fabs( (newLen-oldLen)/(oldLen) );

//fprintf(stderr,"dist  err: %f\t%f\t%f\n\n", distError, oldLen, newLen );

//	if ( distError > distRatioThresh) 
	
	if( (distError * radius/newLen) > .05 )
		{
		return 0;
		}
	else
		{
		cacheHits++;
		return 1;
		}   
	}

//----------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------
void
AngleDistanceMetric::cache( const Point3d new_cam_pos )
	{
	m_old_cam_pos = new_cam_pos;
	}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//
//	Corners Metric
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
CornersMetric::CornersMetric( double pixel_delta )
	{
	m_pixel_delta = pixel_delta;
	}

//----------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------
int CornersMetric::isA(MetricType t)
	{
	if ( t == Corners )	return 1;
	else				return Metric::isA(t);
	}

//----------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------
int
CornersMetric::computeError( BoundBox bbox, const Matrix4d & matrix )
	{
	double actual_sc[8][2];
	double approx_sc[8][2];

	proj_GetActualScreenCorners( bbox, matrix, actual_sc );
	proj_GetApproxScreenCorners( m_wc, matrix, approx_sc );

	int error_ok = 1;
	for( int i=0 ; i<8 ; i++ )
		{
		if( fabs(approx_sc[i][0]-actual_sc[i][0]) > m_pixel_delta ||
			fabs(approx_sc[i][1]-actual_sc[i][1]) > m_pixel_delta )
			{
			error_ok = 0;
			break;
			} 
		}

	return error_ok;
	}

//----------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------
void
CornersMetric::cache( BoundBox bbox, const Matrix4d & matrix )
	{
	double actual_sc[8][2];

	proj_GetActualScreenCorners( bbox, matrix, actual_sc );
	proj_GetActualWorldCorners ( bbox, matrix, actual_sc, m_wc );
	}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//
//	Brick Metric
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#define COS45			.707

BrickMetric::BrickMetric( double pixel_eps, const Camera &camera )
	{
	m_num_tris			= 0;
	m_num_texels		= 0;
	m_cost_draw_geom    = -1;
	m_cost_draw_image   = -1;
	m_cost_create_image = -1;
	m_cost_ammortized	= -1;

	m_valid_cache	= 0;
	m_frames_valid	= 0;

	m_valid_box.setValue( Point3d(0,0,0), Point3d(0,0,0) );

	m_pixel_eps = pixel_eps;

	int		vx,vy,vwidth,vheight;
	camera.getViewport( vx,vy,vwidth,vheight );
	double *frustum = camera.getFrustum();
	double delta = fabs(m_pixel_eps * (frustum[1]-frustum[0]) / vwidth);
	double h = frustum[4];

	m_sineps = delta/( sqrt( delta*delta + h*h) );
	}

//----------------------------------------------------------------------------
//
//		isA
//
//----------------------------------------------------------------------------
int
BrickMetric::isA(MetricType t)
	{
	if ( t == Brick )	return 1;
	else				return Metric::isA(t);
	}


//----------------------------------------------------------------------------
//
//		setGeomSize
//
//----------------------------------------------------------------------------
void
BrickMetric::setGeomSize( int num_tris )
	{
	m_num_tris = num_tris;

	if( m_num_tris <= 100000 )
		m_cost_draw_geom = cost_draw_geom[ m_num_tris/1000 ][1];
	else if( m_num_tris <= 1000000 )
		m_cost_draw_geom = cost_draw_geom[ (m_num_tris-100000)/9000 + 99 ][1];
	else
		m_cost_draw_geom = cost_draw_geom[ 199 ][1];

//fprintf(stderr,"%d %d\n",m_num_tris,m_cost_draw_geom);
	}

//----------------------------------------------------------------------------
//
//		setImageSize
//
//----------------------------------------------------------------------------
void
BrickMetric::setImageSize( int num_texels )
	{
	m_num_texels = num_texels;

	for( int i=0 ; i<NUM_IMAGE_SIZES ; i++ )
		if( cost_draw_image[i][0] == m_num_texels )
			break;

	//	bound to largest known image size
	if( i >= NUM_IMAGE_SIZES ) i=NUM_IMAGE_SIZES-1;

	m_cost_draw_image   = cost_draw_image[i][1];
	m_cost_create_image = cost_create_image[i][1];
	}


//----------------------------------------------------------------------------
//
//		invalidate
//
//----------------------------------------------------------------------------
void
BrickMetric::invalidate()
	{
	m_valid_cache = 0;	
	}

//----------------------------------------------------------------------------
//
//		computeError
//
//----------------------------------------------------------------------------
int
BrickMetric::computeError( const Point3d & v )
	{
	return( m_valid_box.pointInBox( v ) );
	}

//----------------------------------------------------------------------------
//
//		cache
//
//----------------------------------------------------------------------------
int
BrickMetric::cache( const BoundBox & bbox,
					const Camera & camera, const double velocity,
					const BrickMetric * m1, const BrickMetric * m2,
					const Point3d & ray, int anti_alias )
	{
	int		sp[4][2];
	double	wc[8][3];

	if( proj_GetWorldCornersAndScreenPolygon( bbox, wc, sp ) == 0 )
		{
		//
		//	Node was too close to camera,
		//	texture would not fit on screen.
		//
		m_valid_cache = 0;
		return( -1 );
		}


	//
	//	Assumes an aspect ratio of 1.0
	//
	int		vx,vy,vwidth,vheight;
	camera.getViewport( vx,vy,vwidth,vheight );
	double *frustum = camera.getFrustum();
	double eps = fabs(m_pixel_eps * (frustum[1]-frustum[0]) / vwidth);
	double nptd;
	double dist,tmp_dist,d1,d2;

	dist = 0;
	Point3d	p1,p2;
	Point3d cam_pos = camera.getPosition();
	
	//
	//	Iterate over 8 corners of bounding box,
	//	determine how much the camera could move
	//	for each one. Take the minimum.
	//
	for( int i=0 ; i<8 ; i++ )
		{
		//
		//	Determine if the point or its image is closer
		//
		p1 = bbox.getPoint( i );
		p2.setValue( wc[i][0], wc[i][1], wc[i][2] );

		d1 = (p1-cam_pos).length();
		d2 = (p2-cam_pos).length();

/*
		//
		//	Old Method
		//
		double point_sep;
		double npld = frustum[4];

		if( d1>d2 )	nptd = d2-npld;
		else		nptd = d1-npld;

		point_sep = (p1-p2).length();
		
		tmp_dist =	( eps * (point_sep + nptd + npld) * (nptd + npld) )
							/ ( point_sep*npld );

*/
//fprintf(stderr,"\told dist:\t%f\n", tmp_dist );

		//
		//	New Method
		//
		if( d1>d2 )	nptd = d2;
		else		nptd = d1;

		double r = (p1-p2).length() / (2.0*m_sineps);
		
		tmp_dist = sqrt( nptd*nptd + r*r + 2.0*nptd*r*m_sineps ) - r;

//fprintf(stderr,"\tnew dist:\t%f\n", tmp_dist );
//fprintf(stderr,"\tsin eps:\t%f\n", m_sineps );

		if( i==0 )					dist = tmp_dist;
		else if( tmp_dist < dist )	dist = tmp_dist;
/*
fprintf(stderr,"\teps:\t%f\n", eps );
fprintf(stderr,"\tps:\t%f\n", point_sep );
fprintf(stderr,"\tpd:\t%f\n", near_point_dist );
fprintf(stderr,"\tnpd:\t%f\n", near_plane_dist );
fprintf(stderr,"\tdist:\t%f\n", dist );
*/
		}

	//
	//	Construct a "valid box" with the camera at the
	//	center and a radius of "dist".
	//
	double	delta = dist*COS45;
//	double	delta = dist;

//fprintf(stderr,"%f %f\n", dist, delta );

	p1.setValue( cam_pos[0]-delta, cam_pos[1]-delta, cam_pos[2]-delta );
	p2.setValue( cam_pos[0]+delta, cam_pos[1]+delta, cam_pos[2]+delta );

	m_valid_box.setValue( p1, p2 );

	//
	//	Intersect the Valid Box with valid boxes of children.
	//
	int status;
	int	isa_leaf;

	if( m1 == NULL && m2 == NULL )	isa_leaf = 1;
	else							isa_leaf = 0;

	if( !isa_leaf )
		{
		if( m1->m_valid_cache )
			{
			//
			//	intersection does this safely !!!
			//
			status = m_valid_box.intersection( m1->m_valid_box, m_valid_box );

			if( status == 0 )
				{
				m_valid_cache = 0;
				return -2;
				}
			}

		if( m2->m_valid_cache )
			{
			status = m_valid_box.intersection( m2->m_valid_box, m_valid_box );

			if( status == 0 )
				{
				m_valid_cache = 0;
				return -2;
				}
			}
		}

	//
	//	Given the final valid_box, figure out how many frames
	//	it could be valid and compute the ammortized cost of
	//	rendering a texture map at this node.
	//
/*
	p1 = cam_pos - (Point3d) m_valid_box.getMin();
	dist = ( p1[0] > p1[2] ) ? p1[0] : p1[2];
*/
	double tmin = 0;		// we know we are in the box
	double tmax = HUGE;
	m_valid_box.intersect( cam_pos, ray, tmin, tmax );
	dist = tmax;

	m_frames_valid = (int) (dist / velocity) * OVERSAMPLE;

	if( m_frames_valid == 0 )
		{
		m_valid_cache = 0;
		return -3;
		}

	//
	//	Set values for:		m_cost_create_image
	//						m_cost_draw_image
	//
	int width,height;
	int ow	= sp[1][0]-sp[0][0];
	int oh	= sp[1][1]-sp[0][1];

	proj_GetPower2Size( ow, oh, width, height );

	setImageSize( width * height );

	if( !isa_leaf )
		{
		m_cost_ammortized =
			(m1->getCostDraw() + m2->getCostDraw() + m_cost_create_image) /
			m_frames_valid +
			m_cost_draw_image;
		}
	else
		{
		int oversample = 1;
		if( anti_alias )	oversample = OVERSAMPLE;

		m_cost_ammortized = 
			(m_cost_draw_geom*oversample + m_cost_create_image) /
			m_frames_valid +
			m_cost_draw_image;

		//fprintf(stderr,"%d:%d\t%d\t%d\t%d\n",
		//		m_cost_ammortized,m_cost_draw_geom,
		//		m_cost_create_image,m_frames_valid,m_cost_draw_image  );
		}

	//fprintf(stderr,"Frames Valid = %d\n", m_frames_valid );

	m_valid_cache = 1;

	return m_cost_ammortized;
	}

//----------------------------------------------------------------------------
//
//		getCostDraw
//
//----------------------------------------------------------------------------
int
BrickMetric::getCostDraw() const
	{
	if( m_valid_cache == 1 )	return m_cost_draw_image;
	else						return m_cost_draw_geom;
	}

//----------------------------------------------------------------------------
//
//		getCostAmmortized
//
//----------------------------------------------------------------------------
int
BrickMetric::getCostAmmortized() const
	{
	if( m_valid_cache == 1 )	return m_cost_ammortized;
	else						return m_cost_draw_geom;
	}

//----------------------------------------------------------------------------
//
//		getFramesValid
//
//----------------------------------------------------------------------------
int
BrickMetric::getFramesValid() const
	{
	return m_frames_valid;
	}

//----------------------------------------------------------------------------
//
//		getPixelEpsilon
//
//----------------------------------------------------------------------------
double
BrickMetric::getPixelEpsilon() const
	{
	return m_pixel_eps;
	}	

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//
//  SphereMetric
//
//
//		Computes an equilateral triangle centered at the bounding sphere 
//	of the object and orients it so that it lies in a plane parallel to
//	the current view plane.  A ray is then fired from the current camera
//	position through each of the vertices of the triangle and its xsections
//	with the sphere are calculated.  The angle from one intersection point
//	to the old camera position to the other intersection point for the
//	ray is compared to a threshold value.
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
