//============================================================================
//
//	Metric.h
//
//
//============================================================================

#ifndef METRIC_H
#define METRIC_H

#include <stdio.h>
#include <math.h>
#include "Geometry.h"
#include "BoundBox.h"
#include "Camera.h"

#define M1_ANGLE_EPS        (M_PI_2 / 9.0)	// 10 degrees
#define M1_DIST_EPS         0.05

enum MetricType { Base, AngleDistance, Corners, Brick };


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//
//
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class Metric
	{
protected:

	long numEvals;
	long cacheHits;

public:

						Metric();
	virtual int			isA (MetricType t)	{ return ( t == Base ); }
	virtual void		printName()			{ fprintf(stderr,"Base Metric\n"); }; 
	virtual MetricType	Type()				{ return Base; };
	};

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//
//
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class AngleDistanceMetric : public Metric
	{
private:

	double	m_angleThresh;
	double	m_distRatioThresh;
	Point3d	m_old_cam_pos;

public:

					AngleDistanceMetric( double at = M1_ANGLE_EPS,
										 double dr = M1_DIST_EPS );

			void	setAngleThresh (double a) { m_angleThresh = a; }
			double	getAngleThresh (void)	  { return m_angleThresh; }
  
			void	setDistRatioThresh (double d) { m_distRatioThresh = d; }
			double	getDistRatioThresh (void)     { return m_distRatioThresh; }

			int		computeError( const Point3d, const BoundBox& );
			void	cache( const Point3d );

	virtual int 	isA         ( MetricType t );
	virtual void	printName() { fprintf(stderr,"AngleDistance Metric\n"); }; 
	virtual MetricType	Type()	{ return AngleDistance; };
};



//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//
//
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class CornersMetric : public Metric
	{
private:

	double	m_pixel_delta;
	double	m_wc[8][3];

public:

					CornersMetric( double pixel_delta = 2.0 );
  
			void	setPixelDelta (double pd) { m_pixel_delta = pd; }
			int		getPixelDelta (void)	 { return m_pixel_delta; }
  			void	cache		  ( BoundBox bbox, const Matrix4d & matrix );
  			int		computeError  ( BoundBox bbox, const Matrix4d & matrix );

	virtual int 	isA         (MetricType t);
	virtual void	printName() { fprintf(stderr,"Corners Metric\n"); }; 
	virtual MetricType	Type()	{ return Corners; };
	};



//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//
//
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class BrickMetric : public Metric
	{
private:
	
	double		m_pixel_eps;

	int			m_num_tris;
	int			m_num_texels;
	int			m_cost_draw_geom;
	int			m_cost_draw_image;
	int			m_cost_create_image;
	int			m_cost_ammortized;

	int			m_valid_cache;
	int			m_frames_valid;
	BoundBox	m_valid_box;

	double		m_sineps;

public:

				 BrickMetric( double pixel_eps, const Camera & );

			void setGeomSize ( int num_tris );
			void setImageSize( int num_texels );
			void invalidate();
  
  			 int cache ( const BoundBox &,
						 const Camera &, const double velocity, 
						 const BrickMetric *, const BrickMetric *,
						 const Point3d & ray,
						 int anti_alias );

  			 int computeError  ( const Point3d & v );
			 int getCostDraw() const;
			 int getCostAmmortized() const;
			 int getFramesValid() const;
		  double getPixelEpsilon() const;
	
	virtual int 	isA         (MetricType t);
	virtual void	printName() { fprintf(stderr,"Brick Metric\n"); }; 
	virtual MetricType	Type()	{ return Brick; };
	};


#endif   // #ifndef METRIC_H

