//============================================================================
//
//	Node.c  :	 
//  
//
//		Jonathan Shade 
//
//
//============================================================================

#include <GL/gl.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <iostream.h>
#include "TriangleList.h"
#include "Geometry.h"
#include "Metric.h"
#include "Camera.h"
#include "Node.h"
#include "timing.h"
#include "state.h"

#include "glutstroke.h"
extern void* glutStrokeRoman;

extern GLfloat light_pos0[];
extern GLfloat light_pos1[];

#define	NODE_CULLED			0
#define	NODE_VISIBLE		1
#define NODE_INTERNAL		2
#define NODE_TEXTURE		3
#define NODE_GEOMETRY		4

#define NODE_FRONT	50
#define	NODE_BACK	51

#define PIXEL_PAD	4.0

#define MAX_TEX_W	512
#define MAX_TEX_H	512

char	*annotation_str[5] =
{
	"CULLED",
	"VISIBLE",
	"INTERNAL",
	"TEXTURE",
	"GEOMETRY"
};

GLdouble g_color[5][3] =
{
	0.2, 0.0, 0.8,
	0.5, 0.5, 0.5,
	0.0, 0.0, 0.0,
	0.8, 0.0, 0.0,
	0.8, 0.8, 0.0,
};

GLdouble	g_color_black[3]	= { 0.0, 0.0, 0.0 };
GLdouble	g_color_geom[3]		= { 0.8, 0.8, 0.0 };
GLdouble	g_color_image[3]	= { 0.8, 0.0, 0.0 };
GLdouble	g_color_culled[3]	= { 0.2, 0.0, 0.8 };
GLdouble	g_color_visible[3]	= { 0.0, 0.0, 0.0 };

int			Node::m_multisample = 0;
MetricType	Node::m_metric_type = Corners;

int			nextID = 100;
int			g_inflate = 0;
BoundBox	g_per_bbox;
BoundBox	g_bbox;
Node		*g_root;
double		g_radius;

#define	MAX_NEIGHBORS		2000
#define	MAX_NEIGHBOR_TRIS	50000
#define MAX_NEIGHBOR_LEAFS	1000

VisualObject *g_scratch_vobj_list[MAX_NEIGHBORS];
VisualObject *g_scratch_whole_list[MAX_NEIGHBORS];
int			 g_scratch_tlist[MAX_NEIGHBOR_TRIS];
Node*		 g_scratch_leaf_list[MAX_NEIGHBOR_LEAFS];

//----------------------------------------------------------
//
//	Print Annotation
//
//----------------------------------------------------------
void
print_annotation( int label )
	{
	fprintf(stderr,"%s",annotation_str[label]);
	}

//----------------------------------------------------------
//
//	constructor:
//
//----------------------------------------------------------
Node::Node()
	{
	m_height	 = 0;
	m_id		 = nextID++;
	m_annotation = NODE_CULLED;

	m_metric	 = NULL;
	m_force_update_image = 0;

	m_back		 = NULL;
	m_front		 = NULL;
	m_parent	 = NULL;

	m_num_vobjs	 	= 0;
	m_num_nb_vobjs	= 0;
	m_num_nb_whole	= 0;
	m_num_nb_leafs	= 0;

	m_plane.setValue( 0,0,0,0 );

	m_num_tris_local  = 0;
	m_num_tris_total  = 0;
	m_num_tris_stored = 0;
	}


//----------------------------------------------------------
//
// destructor :
//
//
//----------------------------------------------------------
Node::~Node()
	{
	}


//----------------------------------------------------------
//
//		numBytes
//
//----------------------------------------------------------
void
Node::traversal_numBytes( int &bytes, int &extra )
	{
	if( m_num_vobjs == 0 )
		{
		m_back->traversal_numBytes( bytes, extra );
		m_front->traversal_numBytes( bytes ,extra );
		return;
		}
	else
		{
		for( int i=0 ; i<m_num_vobjs ; i++ )
			{
			if( m_vobj_inst_tag[i] == 0 )
				bytes += ((TriangleList *)m_vobj_list[i])->getBytes();
			}

		for( int j=0 ; j<m_num_nb_vobjs ; j++ )
			extra += 4*m_nb_num_tris[j];
		}

	return;
	}


//----------------------------------------------------------
//
//		finalPassStats
//
//----------------------------------------------------------
void
Node::traversal_finalPassStats( FrameStat &stat )
	{
	if( m_annotation == NODE_TEXTURE )
		{
		if( m_billboard.textureAvailable() )
			{
			stat.num_tmap += 1;
			stat.sum_tmap_size += m_billboard.getSize();
			stat.sum_tmap_size_used += m_billboard.getSizeUsed();
			}
		}
	
	if( m_num_vobjs == 0 )
		{
		m_back->traversal_finalPassStats( stat );
		m_front->traversal_finalPassStats( stat );
		}

	return;
	}

//----------------------------------------------------------
//
//		getHeight
//
//----------------------------------------------------------
int
Node::getHeight()
	{
	return( m_height );
	}

//----------------------------------------------------------
//
//		getSceneExtents
//
//----------------------------------------------------------
void
Node::getSceneExtents( Point3d &min, Point3d &max )
	{
	min = m_bbox.getMin();
	max = m_bbox.getMax();
	}

//----------------------------------------------------------
//
//		getBoundBox
//
//----------------------------------------------------------
BoundBox &
Node::getBoundBox()
	{
	return( m_bbox );
	}


//----------------------------------------------------------
//
//		setMultisample
//
//----------------------------------------------------------
void
Node::setMultisample( int flag )
	{
	m_multisample = flag;
	}

//----------------------------------------------------------
//
//		setMetricType
//
//		Should be called before Read()
//
//----------------------------------------------------------
void
Node::setMetricType( MetricType metric_type )
	{
	m_metric_type = metric_type;
	}


//----------------------------------------------------------
//
//		traversal_printAnnotations()
//
//----------------------------------------------------------
void
Node::traversal_printAnnotations()
	{
	if( m_id == 100 )	fprintf( stderr,"\n" );

	fprintf(stderr,"%4d: %s\n", 
			m_id, annotation_str[m_annotation] );

	if( m_num_vobjs == 0 )
		{
		m_back->traversal_printAnnotations();
		m_front->traversal_printAnnotations();
		}	
	}



//----------------------------------------------------------
//
//
//
//----------------------------------------------------------
void
Node::findNeighbors( Node &target )
	{
	if( target.m_id == m_id )	return;

	if( m_num_vobjs == 0 )
		{
		if( target.m_bloated_bbox.intersect( m_back->m_bbox ) != 0 )
			m_back->findNeighbors ( target );

		if( target.m_bloated_bbox.intersect( m_front->m_bbox ) != 0 )
			m_front->findNeighbors( target );
		}
	else
		{
		//
		//	Find Neighboring Objects that intersect
		//	the bloated bounding box.
		//
		//fprintf(stderr,"\t%d\n",m_id);

		int res;
		int found = 0;
		for( int i=0 ; i<m_num_vobjs ; i++ )
			{
			res=target.m_bloated_bbox.intersect(m_vobj_list[i]->getBoundBox());

			if( res == 1 )
				{
				found = 1;
				g_scratch_vobj_list[target.m_num_nb_vobjs] = m_vobj_list[i];

				target.m_num_nb_vobjs++;

				if( target.m_num_nb_vobjs == MAX_NEIGHBORS )
					{
					fprintf( stderr, "Neighbor list (partial) hit MAX!!\n");
					exit(1);
					}
				}
			else if( res == 2 )
				{
				found = 1;

				g_scratch_whole_list[target.m_num_nb_whole] = m_vobj_list[i];

				target.m_num_nb_whole++;

				if( target.m_num_nb_whole == MAX_NEIGHBORS )
					{
					fprintf( stderr, "Neighbor list (whole) hit MAX!!\n");
					exit(1);
					}
				}

			}

		if( found )
			{
			g_scratch_leaf_list[target.m_num_nb_leafs] = this;
			target.m_num_nb_leafs++;

			if( target.m_num_nb_leafs == MAX_NEIGHBOR_LEAFS )
				{
				fprintf( stderr, "Neighbor LEAF list hit MAX!!\n");
				exit(1);
				}
			}
		}
	}

//----------------------------------------------------------
//
//		traversal_findNeighbors()
//
//----------------------------------------------------------
void
Node::traversal_findNeighbors( double inflation )
	{
	//
	//	Internal Nodes
	//
	if( m_num_vobjs == 0 )
		{
		m_back->traversal_findNeighbors( inflation );
		m_front->traversal_findNeighbors( inflation );		

		m_num_tris_total =	m_back->m_num_tris_total +
							m_front->m_num_tris_total;

		m_num_tris_stored = m_back->m_num_tris_stored +
							m_front->m_num_tris_stored;

		((BrickMetric *)m_metric)->setGeomSize( m_num_tris_total );

		m_bloated_bbox = m_back->m_bloated_bbox;
		m_bloated_bbox = m_bloated_bbox.superBox( m_front->m_bloated_bbox );
		return;
		}
	
	//
	//	Leaf Nodes
	//
	Matrix4d xform;
	Point3d	c = m_bbox.getCentroid();

	xform = xform.preConcatTrans( c[0], c[1], c[2] );
	xform = xform.preConcatScale( inflation, inflation, inflation );
	xform = xform.preConcatTrans( -c[0], -c[1], -c[2] );
	m_bloated_bbox = xform * m_bbox;

	g_root->findNeighbors( *this );

		//
		//	Allocate buffer space in target
		//
		m_nb_vobj_list	= new VisualObject* [m_num_nb_vobjs];
		m_nb_whole_list	= new VisualObject* [m_num_nb_whole];

		m_nb_tlist		= new int*  [m_num_nb_vobjs];

		m_nb_num_tris	= new int   [m_num_nb_vobjs];

		m_nb_leaf_list	= new Node* [m_num_nb_leafs];

		//
		//	Copy the neighbors found from the scratch
		//	buffer into a private buffer for the node.
		//
		for( int i=0 ; i<m_num_nb_vobjs ; i++ )
			{
			m_nb_vobj_list[i] = g_scratch_vobj_list[i];
			}

		for( i=0 ; i<m_num_nb_whole ; i++ )
			{
			m_nb_whole_list[i] = g_scratch_whole_list[i];
			}

		for( i=0 ; i<m_num_nb_leafs ; i++ )
			{
			m_nb_leaf_list[i] = g_scratch_leaf_list[i];
			}
		
		//
		//	Now that we have the set of neighboring objects
		//	that are juxtaposed to target, we must clip the
		//	triangles of the neighbors to bloated bounding
		//	box of the target.
		//
		//	This is not actually a clipping operation. It is
		//	an inclusion test. If any of the vertices of a
		//	triangle intersect the bounding box, the entire
		//	triangle is "copied" as is (it is not clipped to
		//	the bloated bounding box).
		//
		//	Copying is actually performed by just storing the
		//	index into the list of triangles of the neighbor.
		//
		Point3d p1,p2,p3;
		VisualObject *v_tmp;
		int num_tris;
		int count;

		for( i=0 ; i<m_num_nb_vobjs ; i++ )
			{
			v_tmp = m_nb_vobj_list[i];

			count = 0;
			num_tris = ((TriangleList *)v_tmp)->getNumTriangles();
			for( int j=0; j<num_tris ; j++ )
				{
				((TriangleList *)v_tmp)->getTriangle(j,p1,p2,p3);

				if( m_bloated_bbox.pointInBox(p1) ||
					m_bloated_bbox.pointInBox(p2) ||
					m_bloated_bbox.pointInBox(p3) )
					{
					g_scratch_tlist[count] = j;

					count++;

					if( count == MAX_NEIGHBOR_TRIS )
					  {
					  fprintf(stderr,"Node::findNeighbors ---batch -image too many tris\n");
					  exit(1);
					  }
					}

				}
			//
			//	Copy the included triangle indices to a buffer
			//	in target.
			//
			m_nb_num_tris[i] = count;
			m_nb_tlist[i] = new int[count];

			for( j=0 ; j<count ; j++ )
				{
				m_nb_tlist[i][j] = g_scratch_tlist[j];
				}
			}

		//
		//	Print Results
		//
		int sum_partial = 0;
		int sum_whole   = 0;

	for( i=0 ; i<m_num_nb_vobjs ; i++ )
		sum_partial += m_nb_num_tris[i];

	for( i=0 ; i<m_num_nb_whole ; i++ )
		sum_whole += ((TriangleList*) m_nb_whole_list[i])->getNumTriangles();
/*
	fprintf(stderr,"%6d: has %6d neighbor tris partial\n",m_id,sum_partial );
	if( sum_whole != 0 )
		fprintf(stderr,"%6d: has %6d neighbor tris whole\n",m_id,sum_whole );
*/

	m_num_tris_total  = m_num_tris_local + sum_partial + sum_whole;
	m_num_tris_stored = m_num_tris_local + sum_partial;

	((BrickMetric *)m_metric)->setGeomSize( m_num_tris_total );
	}

//----------------------------------------------------------
//
//		traversal_initBrickMetricWOInflation()
//
//----------------------------------------------------------
void
Node::traversal_initBrickMetricWOInflation()
	{
	//
	//	Internal Nodes
	//
	if( m_num_vobjs == 0 )
		{
		m_back->traversal_initBrickMetricWOInflation();
		m_front->traversal_initBrickMetricWOInflation();		

		m_num_tris_total =	m_back->m_num_tris_total +
							m_front->m_num_tris_total;

		m_num_tris_stored = m_back->m_num_tris_stored +
							m_front->m_num_tris_stored;

		((BrickMetric *)m_metric)->setGeomSize( m_num_tris_total );

		m_bloated_bbox = m_back->m_bloated_bbox;
		m_bloated_bbox = m_bloated_bbox.superBox( m_front->m_bloated_bbox );
		return;
		}
	
	//
	//	Leaf Nodes
	//
	m_bloated_bbox = m_bbox;

	//
	//	Results
	//
	m_num_tris_total  = m_num_tris_local;
	m_num_tris_stored = m_num_tris_local;

	((BrickMetric *)m_metric)->setGeomSize( m_num_tris_total );
	}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//
//		Traversal Routines for Bricks
//
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


//----------------------------------------------------------
//
//		traversal_printAnnotations_Brick()
//
//----------------------------------------------------------
void
Node::traversal_printAnnotations_Brick()
	{
	if( m_id == 100 )	fprintf( stderr,"\n" );

	fprintf(stderr,"%4d: %s\t%d\t%d\n", 
			m_id, annotation_str[m_annotation],
			m_num_tris_total, ((BrickMetric *)m_metric)->getCostDraw() );

	if( m_num_vobjs == 0 )
		{
		m_back->traversal_printAnnotations_Brick();
		m_front->traversal_printAnnotations_Brick();
		}	
	}

//----------------------------------------------------------
//
//		traversal_cullAndDelete_Brick
//
//----------------------------------------------------------
void
Node::traversal_cullAndDelete_Brick( FrameStat &stat )
	{
	//fprintf(stderr,"*%3d:%2d inv - cad\n",m_id,m_annotation);

	if( m_annotation == NODE_CULLED )	return;

	if( m_annotation == NODE_TEXTURE )
		{
		if( m_billboard.textureAvailable() )
			{
			stat.num_tmap_deleted += 1;
			stat.sum_tmap_size_deleted += m_billboard.getSize();
			}

		m_billboard.kill();
		((BrickMetric *)m_metric)->invalidate();
		}

	m_annotation = NODE_CULLED;

	if( m_num_vobjs == 0 )
		{
		m_back->traversal_cullAndDelete_Brick( stat );
		m_front->traversal_cullAndDelete_Brick( stat );
		}
	}


//----------------------------------------------------------
//
//		traversal_cullAndAnnotate_Brick
//
//		Should be independent of order in which children
//		are culled.
//
//----------------------------------------------------------
void
Node::traversal_cullAndAnnotate_Brick( const Camera &camera, FrameStat &stat )
	{
	int	r[6] = { -2,-2,-2,-2,-2,-2 };
	int	f[5] = { -2,-2,-2,-2,-2 };

	int infrst = camera.inFrustumMoveCamera( m_bbox,r,f );
	int	inside = m_bbox.pointInBox( camera.getPosition() );

//	if( m_id == 100 )	fprintf(stderr,"\n");

	if( !inside )
		{
		//
		//	NOT in the viewing frustum
		//
		if( !infrst )
			{
			//	Invalidate any culled textures
			//
			//fprintf(stderr,"*%3d:%2d inv - caa - culled\n",m_id,m_num_vobjs);
			if( m_annotation == NODE_TEXTURE )
				{
				if( m_billboard.textureAvailable() )
					{
					stat.num_tmap_deleted += 1;
					stat.sum_tmap_size_deleted += m_billboard.getSize();
					}

				m_billboard.kill();
				((BrickMetric *)m_metric)->invalidate();
				}

			//
			//	Recursively cull (invalidate) children
			//
			if( m_num_vobjs == 0 )
				{
				m_back->traversal_cullAndDelete_Brick( stat );
				m_front->traversal_cullAndDelete_Brick( stat );
				}

			m_annotation = NODE_CULLED;
			return;
			}
		//
		//	INSIDE the viewing frustum
		//
		//		- Preserve the TEXTURE annotation of nodes
		//		  from frame to frame.
		//
		//		- Newly visible nodes are VISIBLE.
		//
		else
			{
			if( m_annotation != NODE_TEXTURE )
				m_annotation = NODE_VISIBLE;
			}
		}
	//
	//	Camera inside the bounding box of this node
	//
	else
		{
		//fprintf(stderr,"%d : inside\n", m_id );

		//	This case probably never happens
		if( m_annotation == NODE_TEXTURE )
			{
			//fprintf(stderr,"*%3d inv - caa - camera in box\n",m_id);

			if( m_billboard.textureAvailable() )
				{
				stat.sum_tmap_size_deleted += m_billboard.getSize();
				stat.num_tmap_deleted += 1;
				}

			m_billboard.kill();		
			((BrickMetric *)m_metric)->invalidate();
			}

		if( m_num_vobjs == 0 )
			m_annotation = NODE_INTERNAL;
		else
			m_annotation = NODE_GEOMETRY;
		}

	//
	//	Recurse to children.
	//
	if( m_num_vobjs == 0 )
		{
		m_back->traversal_cullAndAnnotate_Brick( camera, stat );
		m_front->traversal_cullAndAnnotate_Brick( camera, stat );
		}

	//fprintf(stderr,"%3d: ",m_id);
	//print_annotation( m_annotation  );
	//fprintf(stderr,"\n");
	return;
	}


//----------------------------------------------------------
//
//		traversal_UIC_Brick
//
//----------------------------------------------------------

char *err_str[5] = 
{
	"Too Close",
	"Child vbox did not intersect Parent's vbox",
	"Frames Valid == 0",
	"Cost too high",
	"Child not visible or too close"
};

int
Node::traversal_UIC_Brick( const Camera &camera,
							double velocity,
							const Point3d & ray,
							int anti_alias,
							int print_timings,
							FrameStat &stat )
	{
	//cerr << m_id << ": "; print_annotation( m_annotation ); cerr <<"\n";

		 if  ( m_annotation == NODE_CULLED   )
		{
		return( -1 );
		}

	else if ( m_annotation == NODE_INTERNAL )
		{
		m_back->traversal_UIC_Brick(camera, velocity, ray,
									anti_alias, print_timings, stat);
		m_front->traversal_UIC_Brick(camera, velocity, ray,
									anti_alias, print_timings, stat);
		return( -1 );
		}

	else if ( m_annotation == NODE_GEOMETRY )
		{
		return( ((BrickMetric *)m_metric)->getCostDraw() );
		}

	else if ( m_annotation == NODE_TEXTURE  )
		{		
		if( ((BrickMetric *)m_metric)->computeError( camera.getPosition() ) )
			{
			return( ((BrickMetric *)m_metric)->getCostAmmortized() );
			}
		else
			{
			//cerr << "*" << m_id << " inv - uic - brick out of date\n";

			if( m_billboard.textureAvailable() )
				{
				stat.sum_tmap_size_deleted += m_billboard.getSize();
				stat.num_tmap_deleted += 1;
				}

			((BrickMetric *)m_metric)->invalidate();
			m_billboard.invalidate();			
			}
		}

	else if  ( m_annotation == NODE_VISIBLE  )
		{
		//cerr << "Newly Visible\n";
		}

	else
		{
		cerr << "UNKOWN ANNOTATION !!!\n";
		return( -1 );
		}


	//
	//	At this point either the current texture is
	//	invalid, or this is a newly visible node.
	//
	//	Recurse to children:
	//
	int		Cfa,Cba,Cfd,Cbd,C;
	BrickMetric *m1,*m2;

	if( m_num_vobjs == 0 )
		{
		Cba = m_back->traversal_UIC_Brick ( camera, velocity, ray,
											anti_alias, print_timings, stat );
		Cfa = m_front->traversal_UIC_Brick( camera, velocity, ray,
											anti_alias, print_timings, stat );

		m1 = (BrickMetric *)m_back->m_metric;
		m2 = (BrickMetric *)m_front->m_metric;

		Cbd = m1->getCostDraw();
		Cfd = m2->getCostDraw();
		}
	else
		{
		Cba = 0;
		Cfa = 0;
		m1 = NULL;
		m2 = NULL;
		Cbd = ((BrickMetric *)m_metric)->getCostDraw();
		Cfd = 0;
		}

	if( Cfa != -1 && Cba != -1 )
		{
//Timer	brick_timer;
//RESET_TIMER(brick_timer); START_TIMER(brick_timer);

		C = ((BrickMetric *)m_metric)->cache( m_bbox, camera, velocity,
											  m1, m2, ray, anti_alias );
//STOP_TIMER(brick_timer);
//cerr << "\tBrick:\t" << calc_usec_time( brick_timer ) << "\n";

		if( (C < Cbd + Cfd) && (C > -1) )
			{
			m_annotation = NODE_TEXTURE;

//Timer	create_timer;
//RESET_TIMER( create_timer ); START_TIMER( create_timer );

			if( m_billboard.textureAvailable() )
				{
				stat.sum_tmap_size_deleted += m_billboard.getSize();
				stat.num_tmap_deleted += 1;
				}

			int status;
			double pixel_eps = ((BrickMetric *)m_metric)->getPixelEpsilon();

			//setupRotatedLookat( camera, m_funky, 1 );
			status = setupFrustum( m_bbox, camera, pixel_eps );

			if( status != 0 )
				{
				//
				//	Internal Node
				//
				if( m_num_vobjs == 0 )
					{
					this->drawChildrenSpecial( camera );
#ifdef TEX_OBJ
					status = m_billboard.createTextureTO(
								m_bbox, pixel_eps+PIXEL_PAD, 0 );
#else
					status = m_billboard.createTextureDL(
								m_bbox, pixel_eps+PIXEL_PAD, 0 );
#endif
					//m_billboard.printTimings( m_id );
					}
				//
				//	Leaf Node
				//	
				else
					{
					if(anti_alias)	this->drawSelfSpecialAA( camera, 1 );
					else			this->drawSelfSpecial( camera, 1 );
#ifdef TEX_OBJ
					status = m_billboard.createTextureTO( m_bbox, pixel_eps+PIXEL_PAD, 0 );
#else
					status = m_billboard.createTextureDL( m_bbox, pixel_eps+PIXEL_PAD, 0 );
#endif
					//m_billboard.printTimings( m_id );
					}

				glMatrixMode( GL_PROJECTION );
					glPopMatrix();
				glMatrixMode( GL_MODELVIEW );
				}
			else
				{
				//fprintf(stderr,"%d: not feasible\n",m_id);
				}

//STOP_TIMER( create_timer );
//cerr << "\tCreate:\t" << calc_usec_time( create_timer ) << "\n";

			//
			//	Error creating texture
			//
			if( status <= 0 )
				{
				//fprintf(stderr,"%d: billboard creation failed\n",m_id);
				if( m_billboard.textureAvailable() )
					{
					stat.sum_tmap_size_deleted += m_billboard.getSize();
					stat.num_tmap_deleted += 1;
					}

				m_billboard.kill();
				((BrickMetric *)m_metric)->invalidate();

				if( m_num_vobjs == 0 )	m_annotation = NODE_VISIBLE;
				else					m_annotation = NODE_GEOMETRY;
				}

			//
			//	We created a new texture !
			//
			else
				{
				stat.sum_tmap_size_created += m_billboard.getSize();
				stat.num_tmap_created += 1;

				if( stat.max_composite_depth == -1 )
					stat.max_composite_depth = m_height;
				else if( m_height > stat.max_composite_depth )
					stat.max_composite_depth = m_height;

				stat.sum_composite_depth += m_height;
				stat.num_composites += 1;

				int fv = ((BrickMetric *)m_metric)->getFramesValid(); 

				if( stat.max_frames_valid == -1 )
					stat.max_frames_valid = fv;
				else if ( fv > stat.max_frames_valid )
					stat.max_frames_valid = fv;

				if( stat.min_frames_valid == -1 )
					stat.min_frames_valid = fv;
				else if ( fv < stat.min_frames_valid )
					stat.min_frames_valid = fv;
	
				stat.sum_frames_valid += fv;
				}
			}
		//
		//	Cost was too HIGH !!!
		//
		else
			{
			/*
			cerr << "*" << m_id << " inv - uic - cost too high - "
					<< C << " " << Cbd << " " << Cfd << "\n";
			*/
			if( C > -1 )
				{
				C = -4;
				}

			if( m_billboard.textureAvailable() )
				{
				stat.sum_tmap_size_deleted += m_billboard.getSize();
				stat.num_tmap_deleted += 1;
				}

			m_billboard.kill();
			((BrickMetric *)m_metric)->invalidate();

			if( m_num_vobjs == 0 )	m_annotation = NODE_VISIBLE;
			else					m_annotation = NODE_GEOMETRY;
			}

		}
	//
	//	One of the children was not visible or was
	//	 too close to the viewer to cache an image.
	//
	else
		{
		if( m_annotation == NODE_TEXTURE )
			{
			//cerr << "*" << m_id << " inv - uic - too close or no child\n";
			if( m_billboard.textureAvailable() )
				{
				stat.sum_tmap_size_deleted += m_billboard.getSize();
				stat.num_tmap_deleted += 1;
				}

			m_billboard.kill();
			((BrickMetric *)m_metric)->invalidate();
			}
		C = -5;

		if( m_num_vobjs == 0 )	m_annotation = NODE_VISIBLE;
		else					m_annotation = NODE_GEOMETRY;
		}
/*
cerr << m_id << ": ";
print_annotation( m_annotation );

	if( C<0 ) cerr << "\t" << err_str[ -C-1 ] << "\n";
	else	  cerr << "\t" << ((BrickMetric *)m_metric)->getFramesValid() << ":" << C <<
                              Cbd << Cfd << "\n";
*/

	if( C<0 ) C = -1;

	return( C );
	}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


//----------------------------------------------------------
//
//		traversal_cullAndDelete
//
//----------------------------------------------------------
void
Node::traversal_cullAndDelete( FrameStat &stat )
	{
	if( m_annotation == NODE_CULLED )	return;

	if( m_annotation == NODE_TEXTURE )
		{
		if( m_billboard.textureAvailable() )
			{
			stat.sum_tmap_size_deleted += m_billboard.getSize();
			stat.num_tmap_deleted += 1;
			}

		m_billboard.invalidate();
		}
	m_annotation = NODE_CULLED;

	if( m_num_vobjs == 0 )
		{
		m_back->traversal_cullAndDelete( stat );
		m_front->traversal_cullAndDelete( stat );
		}
	}

//----------------------------------------------------------
//
//		traversal_cullAndAnnotate
//
//		Should be independent of order in which children
//		are culled.
//
//----------------------------------------------------------
void
Node::traversal_cullAndAnnotate( const Camera &camera, int level, FrameStat &stat )
	{
	int	r[6] = { -2,-2,-2,-2,-2,-2 };
	int	f[5] = { -2,-2,-2,-2,-2 };
	int infrst = camera.inFrustumMoveCamera( m_bbox,r,f );
	int	inside = m_bbox.pointInBox( camera.getPosition() );

//	if( m_id == 100 )	fprintf(stderr,"\n");


	if( !inside )
		{
		//
		//	NOT in the viewing frustum
		//
		if( !infrst )
			{
			if( m_annotation == NODE_TEXTURE )
				{
//fprintf( stderr,"%4d: Culled\n", m_id );
			if( m_billboard.textureAvailable() )
				{
				stat.sum_tmap_size_deleted += m_billboard.getSize();
				stat.num_tmap_deleted += 1;
				}

				m_billboard.invalidate();
				}

			m_annotation = NODE_CULLED;
			if( m_num_vobjs == 0 )
				{
				m_back->traversal_cullAndDelete( stat );
				m_front->traversal_cullAndDelete( stat );
				}
			return;
			}
		//
		//	INSIDE the viewing frustum
		//
		else
			{
			if( m_num_vobjs != 0 || m_height == 1)
				{
				if( m_annotation != NODE_TEXTURE )
					{
//fprintf( stderr,"%4d: No Longer Culled: %d : %d\n",m_id,m_annotation,
//				m_billboard.textureAvailable());

					m_force_update_image = 1;

					//
					//	March up the tree invalidating parent textures.
					//
					Node *np = m_parent;
					while( np != NULL )
						{
						if( np->m_annotation == NODE_TEXTURE )
							{
							np->m_force_update_image = 1;
							np = np->m_parent;
							}
						else
							np = NULL; 
						}
					}
				m_annotation = NODE_TEXTURE;
				}
			else
				{
				m_annotation = NODE_INTERNAL;
				}
			}
		}
	else
		{
//		fprintf(stderr,"%d : inside\n", m_id );

		if( m_annotation == NODE_TEXTURE )
			{
			if( m_billboard.textureAvailable() )
				{
				stat.sum_tmap_size_deleted += m_billboard.getSize();
				stat.num_tmap_deleted += 1;
				}

			m_billboard.invalidate();		
			}

		if( m_num_vobjs == 0 )
			m_annotation = NODE_INTERNAL;
		else
			m_annotation = NODE_GEOMETRY;
		}

	if( m_num_vobjs == 0 )
		{
		m_back->traversal_cullAndAnnotate( camera, level + 1, stat );
		m_front->traversal_cullAndAnnotate( camera, level + 1, stat );
		}

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	//	Clean up the annotations given child information
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	if( m_annotation == NODE_TEXTURE )
		{
		if( m_num_vobjs == 0 )
			{
			//
			//		Only Create an image cache if all children are visible.
			//
			if( (m_back->m_annotation == NODE_CULLED)   ||
				(m_back->m_annotation == NODE_INTERNAL) ||	
				(m_front->m_annotation == NODE_CULLED)  ||	
				(m_front->m_annotation == NODE_INTERNAL) )	
				{
				//fprintf(stderr,"%4d: Cant see some child\n", m_id );
				m_annotation = NODE_INTERNAL;
				return;
				}
			}

		//
		//		Check to see if object is too close(large) to
		//		create an image cache for it.
		//
		Matrix4d m_funky;
		setupRotatedLookat( camera, m_funky, 1 );
		double pixel_eps = ((BrickMetric *)m_metric)->getPixelEpsilon();
		int		  status = m_billboard.feasibleTexture( m_bbox,  camera,
														m_funky, pixel_eps );

		if( status != 1 )
			{
			if( m_billboard.textureAvailable() )
				{
				stat.sum_tmap_size_deleted += m_billboard.getSize();
				stat.num_tmap_deleted += 1;
				}

			m_billboard.invalidate();
			if( m_num_vobjs != 0 )	m_annotation = NODE_GEOMETRY;	// Leaf
			else					m_annotation = NODE_INTERNAL;	// Internal
			}
		}	

	return;
	}

//----------------------------------------------------------
//
//		traversal_UIC
//
//----------------------------------------------------------
void
Node::traversal_UIC( const Camera &camera, int print_timings,
					 int draw_frames, int anti_alias, FrameStat &stat )
	{
	if( m_annotation == NODE_CULLED )	return;

	else if ( m_annotation == NODE_INTERNAL )
		{
		if( m_plane.isInFront( camera.getPosition() ) )
			{
			m_back->traversal_UIC ( camera, print_timings, draw_frames,
									anti_alias, stat );
			m_front->traversal_UIC( camera, print_timings, draw_frames,
									anti_alias, stat );
			}
		else
			{
			m_front->traversal_UIC( camera, print_timings, draw_frames,
									anti_alias, stat );
			m_back->traversal_UIC ( camera, print_timings, draw_frames,
									anti_alias, stat );
			}
		}
	else if ( m_annotation == NODE_TEXTURE )
		{
		Point3d new_cam_pos	= camera.getPosition();
		int error_ok;

		Matrix4d m_funky;

		setupRotatedLookat( camera, m_funky, 1 );

		switch( m_metric->Type() )
			{
			case AngleDistance:
				error_ok	= ((AngleDistanceMetric *)m_metric)->
							  computeError(new_cam_pos, m_bbox);
				break;

			case Corners:
				error_ok	= ((CornersMetric *)m_metric)->
							  computeError(m_bbox,m_funky);
				break;
			}

		if( !error_ok ||
			!(m_billboard.textureAvailable()) ||
			m_force_update_image )
			{
			m_force_update_image = 0;

			switch( m_metric->Type() )
				{
				case AngleDistance:
					((AngleDistanceMetric *)m_metric)->cache(new_cam_pos);
					break;

				case Corners:
					((CornersMetric *)m_metric)->cache(m_bbox,m_funky);
					break;
				}

			if( m_billboard.textureAvailable() )
				{
				stat.sum_tmap_size_deleted += m_billboard.getSize();
				stat.num_tmap_deleted += 1;
				}

			setupRotatedLookat( camera, m_funky, 1 );

			double pixel_eps = ((BrickMetric *)m_metric)->getPixelEpsilon();

			//
			//	Internal Node
			//
			if( m_num_vobjs == 0 )
				{
				if( m_plane.isInFront( camera.getPosition() ) )
					{
					m_back->traversal_UIC( camera, print_timings, draw_frames,
											anti_alias, stat);

					m_front->traversal_UIC(camera, print_timings, draw_frames,
											anti_alias, stat);
					}
				else
					{
					m_front->traversal_UIC(camera, print_timings, draw_frames,
										   anti_alias, stat);
					m_back->traversal_UIC( camera, print_timings, draw_frames,
										   anti_alias, stat);
					}

				this->drawChildrenSpecial( camera );
#ifdef TEX_OBJ
				m_billboard.createTextureTO( m_bbox, pixel_eps, 0 );
#else
				m_billboard.createTextureDL( m_bbox, pixel_eps, 0 );
#endif
//				m_billboard.printTimings( m_id );
				}
			//
			//	Leaf Node
			//	
			else
				{
				if(anti_alias)	this->drawSelfSpecialAA( camera, 1 );
				else			this->drawSelfSpecial( camera, 1 );
#ifdef TEX_OBJ
				m_billboard.createTextureTO( m_bbox, pixel_eps, 0 );
#else
				m_billboard.createTextureDL( m_bbox, pixel_eps, 0 );
#endif
//				m_billboard.printTimings( m_id );
				}

			//
			//	Created a new texture !!!
			//
			stat.sum_tmap_size_created += m_billboard.getSize();
			stat.num_tmap_created += 1;

			} // ERROR THRESH EXCEEDED
		} // TEXTURE

	return;
	}

//----------------------------------------------------------
//
//		traversal_init
//
//----------------------------------------------------------
#define EPSILON	0.000001
int
intersect( const Point3d & Q, const Normal3d & w, const Point3d & P,
			const Normal3d & n, Point3d &q )
	{
	double denom = w.dot(n);

	if( fabs(denom) < EPSILON )	return 0;

	double t = ( (P - Q).dot(n) )/( denom );

	q = Q + t*w;

	return 1;
	}


//----------------------------------------------------------
//
//	Search Up the hierarchy, looking for
//	intersections with grandparents.
//
//----------------------------------------------------------
int
Node::intersect_with_parents( const Point3d & p, const Vector3d & v,
								Point3d &q, Node *seed )
	{
	int			status;
	Point3d		P;
	Vector3d	N;
	Node		*gp = seed;

	status = 0;
	while( gp != NULL )
		{
		P = gp->m_plane.getPoint3d();
		N = gp->m_plane.getNormal3d();

		status = intersect( p, v, P, N, q );

		//
		//	Dont allow to double back
		//
		if( (m_tag == NODE_FRONT) && (m_parent->m_plane.isBehind( q )) )
			status = 0;
		else if( (m_tag == NODE_BACK) && (m_parent->m_plane.isInFront( q )) )
			status = 0;

		if( status == 1)	break;
		else				gp = gp->m_parent;
		}

	return status;
	}


//----------------------------------------------------------
//
//		intersect_with_boundary
//
//----------------------------------------------------------
int
Node::intersect_with_boundary( const Point3d & p, const Vector3d & v,
								Point3d &q )
	{
	int		status;
	Point3d	min,max;

	g_bbox.getValue( min, max );

	if( v[0] < 0.0 )
		status = intersect(p, v, Point3d(min[0],0,0), Normal3d( -1,0,0 ), q);
	else
		status = intersect(p, v, Point3d(max[0],0,0), Normal3d(  1,0,0 ), q);

	if( (status==0) && v[2]>=0.0 )
		status = intersect(p, v, Point3d(0,0,max[2]), Normal3d(  0,0,1 ), q);

	else if( (status==0) && v[2]<0.0 )
		status = intersect(p, v, Point3d(0,0,min[2]), Normal3d(  0,0,-1 ), q);

	else if( q[2]>max[2] )
		status = intersect(p, v, Point3d(0,0,max[2]), Normal3d(  0,0,1 ), q);

	else if( q[2]<min[2] )
		status = intersect(p, v, Point3d(0,0,min[2]), Normal3d(  0,0,-1 ), q);

	return status;
	}


//----------------------------------------------------------
//
//		traversal_init2DBoundary
//
//----------------------------------------------------------
void
Node::traversal_init2DBoundary()
	{
	int		status;
	Vector3d	bi,n;
	Point3d		p;
	Point3d		p1,p2;

	//
	//	Debugging Print Statements
	//
	/*
	if		( m_parent == NULL )	fprintf( stderr,"ROOT\n" );
	else if	( m_tag == NODE_FRONT )	fprintf( stderr,"FRONT\n" );
	else if ( m_tag == NODE_BACK )	fprintf( stderr,"BACK\n" );
	*/

	if( m_num_vobjs != 0 )
		{
		//fprintf( stderr,"  Object\n\n\n" );
		return;
		}


	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	//	1st Intersection Point:
	//	----------------------	
	//	Try to intersect with parents
	//
	//	Special case immediate parent
	//
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	p  = m_plane.getPoint3d();
	n  = m_plane.getNormal3d();
	bi = n.cross( Vector3d( 0,1,0 ) );
	bi.normalize();

	status = 0;
	if( m_parent != NULL )
		{
		Point3d  P = m_parent->m_plane.getPoint3d();
		Vector3d N = m_parent->m_plane.getNormal3d();

		status = intersect( p, bi, P, N, p1 );

		if( status == 0 )
			status = intersect_with_parents( p, bi, p1, m_parent->m_parent );	
		}

	if( status == 0 )
		status = intersect_with_boundary( p, bi, p1 );

	//fprintf(stderr,"P1: %d: ", status ); p1.Print();
	
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	//
	//	2nd Intersection Point
	//
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	//===========================================================
	//	redefine the binormal
	//
	//	We are now going to find the intersection of a ray
	//	"shot" from p in the direction of the splitting plane.
	//
	//	We first see if this ray hits any of the parents of the plane,
	//	then we see where it hits the boundary of the scene.
	//
	//	Perturb the origin of the ray so that it does not
	//	coincide with the first intersection point.
	//
	//	Then find the ray along the splitting plane as
	//	defined by:
	//				ray = P1 - P
	//===========================================================

	if( p == p1 )	p += bi;
	bi = p - p1;
	bi.normalize();

	//
	//	Make sure the ray is pointing in the correct direction
	//	by finding on which side of the parent plane the point p lies. 
	//
	if( (m_tag == NODE_FRONT) && (m_parent->m_plane.isBehind(p)) )
		bi.negate();
	else if ( (m_tag == NODE_BACK) && (m_parent->m_plane.isInFront(p)) )
		bi.negate();

	//===========================================================
	//	Again, try to intersect with grand parents first.
	//
	//	The first point P1 was the intersection with the 
	//	immediate parent ( if any ), so we want to skip over that
	//	plane.
	//
	//	Should intersect_with_parents return the node that was the
	//	first intersection ??? It may not be the immediate parent !
	//===========================================================
	status = 0;
	if( m_parent != NULL )
		status = intersect_with_parents( p1, bi, p2, m_parent->m_parent );

	if( status == 0 )
		status = intersect_with_boundary( p1, bi, p2 );

	//
	//	Done, recurse
	//
	m_line_beg = p1;
	m_line_end = p2;

	m_back->traversal_init2DBoundary();
	m_front->traversal_init2DBoundary();
	
	return;
	}

//----------------------------------------------------------
//
//		traversal_draw2DBoundary
//
//		Back/Front order not dependent on camera position.
//
//----------------------------------------------------------
void
Node::traversal_draw2DBoundary()
	{
	if( m_num_vobjs != 0 )	return;

	Point3d	b   = m_line_beg;
	Point3d e   = m_line_end;
	Vector3d v	= (e-b);
	double len	= v.normalize();
	
	b += v*(0.05*len);
	e -= v*(0.05*len);

	glColor3d( 1.0, 0.0, 1.0 );

	glBegin( GL_LINES );

	glVertex3d( b[0], b[1], b[2] );
	glVertex3d( e[0], e[1], e[2] );

	glEnd();

	m_back->traversal_draw2DBoundary();
	m_front->traversal_draw2DBoundary();
	
	return;	
	}

//----------------------------------------------------------
//
//		traversal_drawImageCache
//
//----------------------------------------------------------
void
Node::traversal_drawImageCache( const Camera &camera,
								int draw_frames, 
								int draw_leaf_textures,
								int draw_psuedo_color,
								FrameStat &stat )
	{
/*
	GLboolean wm;
	glGetBooleanv( GL_DEPTH_WRITEMASK, &wm );
	if( wm != GL_FALSE )
		cerr << "dic: Depth Mask not false\n";

	if( glIsEnabled( GL_DEPTH_TEST ) != GL_TRUE )
		cerr << "dic: Depth Test not enabled\n";
	if( glIsEnabled( GL_TEXTURE_2D ) != GL_TRUE )
		cerr << "dic: Texturing not enabled\n";
	if( glIsEnabled( GL_BLEND ) != GL_TRUE )
		cerr << "dic: Blending not enabled\n";
*/
	int texture_recurse = 0;

	switch( m_annotation )
		{
		case NODE_TEXTURE:
			if( (draw_leaf_textures && m_num_vobjs != 0)
				|| !draw_leaf_textures )
				{
				if( draw_frames )
					{
					glDisable( GL_TEXTURE_2D );
					glDisable( GL_BLEND );
					glColor3f( 1.0, 1.0, 1.0 );
					m_billboard.drawTextureFrameNocolor();
					glEnable( GL_TEXTURE_2D );
					glEnable( GL_BLEND );
					}

				if( draw_psuedo_color )
					{
					m_billboard.drawColored();
					}
				else
					{
					if ( m_billboard.draw() == 0)
						fprintf(stderr,
							"%3d -- Error -- Invalid Texture Drawn\n",m_id);
					}
				}

			if( draw_leaf_textures && m_num_vobjs == 0 )
				texture_recurse = 1;

			stat.num_tmap_drawn      += 1;
			stat.sum_tmap_size_drawn += m_billboard.getSize();
			break;
		
		case NODE_GEOMETRY:
			//
			//	This call may be redundant.
			//	It is here because drawChildrenSpecial
			//	calls this function.
			//
			glDisable( GL_TEXTURE_2D );
			glDisable( GL_BLEND );
			glDepthMask( GL_TRUE );
	
			if( m_multisample )	glEnable(GL_MULTISAMPLE_SGIS);

			m_bbox.setGLClipPlanesEps();

			for( int i=0 ; i<m_num_vobjs ; i++ )
				if( draw_psuedo_color )
					((TriangleList *)m_vobj_list[i])->draw_psuedoColor();
				else
					m_vobj_list[i]->draw();
				
			m_bbox.unsetGLClipPlanes();

			int found=0;
			for( i=0 ; i<m_num_nb_leafs ; i++ )
				if( m_nb_leaf_list[i]->m_annotation == NODE_TEXTURE )
					{found = 1; break;}
		
			found =0;
			if( found )
				{
				m_bloated_bbox.setGLClipPlanesEps();
				for( i=0 ; i<m_num_nb_vobjs ; i++ )
					((TriangleList *)m_nb_vobj_list[i])->
						draw_list(m_nb_tlist[i],m_nb_num_tris[i]);

				for( i=0 ; i<m_num_nb_whole ; i++ )
					((TriangleList *)m_nb_whole_list[i])->draw();

				m_bloated_bbox.unsetGLClipPlanes();
				}


			if( m_multisample )	glDisable(GL_MULTISAMPLE_SGIS);

			glDepthMask( GL_FALSE );
			glEnable( GL_BLEND );
			glEnable(GL_TEXTURE_2D );

			stat.num_tri_geom += m_num_tris_local;
			break;
		};

	if( m_annotation == NODE_INTERNAL ||
		m_annotation == NODE_VISIBLE  ||
		texture_recurse )
		{
		if( m_back != NULL && m_front != NULL )
			{
			if( m_plane.isInFront( camera.getPosition() ) )
				{
				m_back->traversal_drawImageCache
					(camera, draw_frames, draw_leaf_textures,
					 draw_psuedo_color, stat);
				m_front->traversal_drawImageCache
					(camera, draw_frames, draw_leaf_textures,
					 draw_psuedo_color, stat);
				}
			else
				{
				m_front->traversal_drawImageCache
					(camera, draw_frames, draw_leaf_textures,
					 draw_psuedo_color, stat);
				m_back->traversal_drawImageCache
					(camera, draw_frames, draw_leaf_textures,
					 draw_psuedo_color, stat);
				}
			}
		}
	}

//----------------------------------------------------------
//
//		traversal_drawImageCache
//
//----------------------------------------------------------
void
Node::traversal_drawImageCachePriv( const Camera &camera )
	{
/*
	GLboolean wm;
	glGetBooleanv( GL_DEPTH_WRITEMASK, &wm );
	if( wm != GL_FALSE )
		cerr << "dic: Depth Mask not false\n";

	if( glIsEnabled( GL_DEPTH_TEST ) != GL_TRUE )
		cerr << "dic: Depth Test not enabled\n";
	if( glIsEnabled( GL_TEXTURE_2D ) != GL_TRUE )
		cerr << "dic: Texturing not enabled\n";
	if( glIsEnabled( GL_BLEND ) != GL_TRUE )
		cerr << "dic: Blending not enabled\n";
*/
	switch( m_annotation )
		{
		case NODE_TEXTURE:
			if ( m_billboard.draw() == 0)
				fprintf(stderr,"%3d -- Error -- Invalid Texture Drawn\n",m_id);
			break;
		

		case NODE_GEOMETRY:
			//
			//	This call may be redundant.
			//	It is here because drawChildrenSpecial
			//	calls this function.
			//
			glDisable( GL_TEXTURE_2D );
			glDisable( GL_BLEND );
			glDepthMask( GL_TRUE );
	
			if( m_multisample )	glEnable(GL_MULTISAMPLE_SGIS);

			m_bbox.setGLClipPlanesEps();

			for( int i=0 ; i<m_num_vobjs ; i++ )
				m_vobj_list[i]->draw();

			m_bbox.unsetGLClipPlanes();

			int found=0;
			for( i=0 ; i<m_num_nb_leafs ; i++ )
				if( m_nb_leaf_list[i]->m_annotation == NODE_TEXTURE )
					{found = 1; break;}
		
			found =0;
			if( found )
				{
				m_bloated_bbox.setGLClipPlanesEps();
				for( i=0 ; i<m_num_nb_vobjs ; i++ )
					((TriangleList *)m_nb_vobj_list[i])->
						draw_list(m_nb_tlist[i],m_nb_num_tris[i]);

				for( i=0 ; i<m_num_nb_whole ; i++ )
					((TriangleList *)m_nb_whole_list[i])->draw();

				m_bloated_bbox.unsetGLClipPlanes();
				}


			if( m_multisample )	glDisable(GL_MULTISAMPLE_SGIS);

			glDepthMask( GL_FALSE );
			glEnable( GL_BLEND );
			glEnable(GL_TEXTURE_2D );
			break;
		};

	if( m_annotation == NODE_INTERNAL ||
		m_annotation == NODE_VISIBLE  )
		{
		if( m_back != NULL && m_front != NULL )
			{
			if( m_plane.isInFront( camera.getPosition() ) )
				{
				m_back ->traversal_drawImageCachePriv( camera );
				m_front->traversal_drawImageCachePriv( camera );
				}
			else
				{
				m_front->traversal_drawImageCachePriv( camera );
				m_back ->traversal_drawImageCachePriv( camera );
				}
			}
		}
	}

//----------------------------------------------------------
//
//		traversal_drawGeometry
//
//		Traverse down to leaves and draw the geometry.
//
//		Assumes no NULL children except at leaves !!!!
//
//----------------------------------------------------------
void
Node::traversal_drawGeometry( const Camera &camera, FrameStat &stat,
								int ignore_culling )
	{
//	fprintf(stderr,"%2d,%d: drawing\n",m_id,m_num_vobjs); 

	if( !ignore_culling )
		if( m_annotation == NODE_CULLED )	return;

	if( m_num_vobjs == 0 )
		{
		if( m_plane.isInFront( camera.getPosition() ) )
			{
			m_back->traversal_drawGeometry( camera, stat, ignore_culling );
			m_front->traversal_drawGeometry( camera, stat, ignore_culling );
			}
		else
			{
			m_front->traversal_drawGeometry( camera, stat, ignore_culling );
			m_back->traversal_drawGeometry( camera, stat, ignore_culling );
			}
		}
	else
		{
		m_bbox.setGLClipPlanesEps();
		for( int i=0 ; i< m_num_vobjs ; i++ )
			{
			m_vobj_list[i]->draw();
			}
		m_bbox.unsetGLClipPlanes();

		stat.num_tri_geom += m_num_tris_local;
		}
	return;
	}

//----------------------------------------------------------
//
//		traversal_drawBoundingBoxes
//
//----------------------------------------------------------
void
Node::traversal_drawBoundingBoxes( int draw_frames, int draw_id )
	{
	switch( m_annotation )
		{
		case NODE_CULLED:
				//if( m_num_vobjs != 0 )
					{
					glColor3dv( g_color_culled );
					m_bbox.Draw();
		
					if( draw_id ) {glColor3dv( g_color_black );drawID();}
					}
				return;
				break;

		case NODE_VISIBLE:
				if( m_num_vobjs != 0 )
					{
					glColor3dv( g_color_visible );
					m_bbox.Draw();
		
					if( draw_id ) {glColor3dv( g_color_black );drawID();}
					}
				break;

		case NODE_TEXTURE:
				if( draw_frames )
					{
					glColor3dv( g_color_black );
					m_billboard.drawTextureFrameNocolor();		
					}

				//if( m_num_vobjs != 0 )
					{
					glColor3dv( g_color_image  );
					m_bbox.Draw();
					//glColor3f( 0,0,1 );
					//m_bloated_bbox.Draw();

					//for( int i=0 ; i< m_num_vobjs ; i++ )
					//	(m_vobj_list[i]->getBoundBox()).Draw();
					}

				if( draw_id ) {glColor3dv( g_color_black );drawID();}
				return;
				break;

		case NODE_GEOMETRY:
			if( m_multisample )
				glEnable(GL_MULTISAMPLE_SGIS); 

				glColor3dv( g_color_geom   );
				m_bbox.Draw();

				/*
				for( int j=0 ; j< m_num_vobjs ; j++ )
					{
					glColor3dv( g_color_geom   );
					(m_vobj_list[j]->getBoundBox()).Draw();
					}
				*/
				if( draw_id ) {glColor3dv( g_color_black );drawID();}

			if( m_multisample )
				glDisable(GL_MULTISAMPLE_SGIS); 
				break;

		case NODE_INTERNAL:
//				m_back->traversal_drawBoundingBoxes ( draw_frames, draw_id );
//				m_front->traversal_drawBoundingBoxes( draw_frames, draw_id );
				break;
		};

		if( m_num_vobjs == 0 )
			{
			m_back->traversal_drawBoundingBoxes ( draw_frames, draw_id );
			m_front->traversal_drawBoundingBoxes( draw_frames, draw_id );
			}
	return;
	}

//----------------------------------------------------------
//
//		traversal_drawBoundingBoxesSolid
//
//----------------------------------------------------------
void
Node::traversal_drawBoundingBoxesSolid()
	{
	if( m_num_vobjs == 0 )
		{
		m_back->traversal_drawBoundingBoxesSolid();
		m_front->traversal_drawBoundingBoxesSolid();
		return;
		}


	if( m_multisample )
		glEnable(GL_MULTISAMPLE_SGIS); 

	for( int j=0 ; j< m_num_vobjs ; j++ )
		{
		(m_vobj_list[j]->getBoundBox()).Draw_solid( g_color[m_annotation] );
		}

	if( m_multisample )
		glDisable(GL_MULTISAMPLE_SGIS); 

	return;
	}

//----------------------------------------------------------
//
//		traversal_drawHierarchy
//
//----------------------------------------------------------
void
Node::traversal_drawHierarchy( double x, double y )
	{
	switch( m_annotation )
		{
		case NODE_CULLED:
				glColor3d( 0, 0, 1 );
				break;

		case NODE_TEXTURE:
				glColor3d( 1, 1, 0 );
				break;

		case NODE_GEOMETRY:
				glColor3d( 1, 0, 0 );
				break;

		case NODE_INTERNAL:
				glColor3d( 0, 0, 0 );
				break;

		default:
				glColor3d( 0, 1, 0 );
		};

	glRectd( x-4,y-4,x+4,y+4 );

	if( m_num_vobjs == 0 )
		{
		double width = pow(2,m_height);
		m_back->traversal_drawHierarchy ( x-(2.5*width), y-10 );
		m_front->traversal_drawHierarchy( x+(2.5*width), y-10 );
		}

	return;
	}

 
//----------------------------------------------------------
//
//		drawID
//
//----------------------------------------------------------
void
Node::drawID()
	{
	char	msg[80];
	float	factor = ( m_bbox.getRadius() * 0.18 ) / 66.0 ;

	glColor3f( 0.0, 0.0, 0.0 );

	sprintf( msg, "%03d", m_id );

	Point3d c = m_bbox.getCentroid();

	glPushMatrix();
		glTranslatef( c[0], c[1], c[2] );
		glTranslatef( -m_bbox.getRadius() * 0.4, 0.0, 0.0 );
		glRotatef( -90.0, 1.0, 0.0, 0.0 ); 
		glScalef( factor, factor, factor );
		for( int i=0 ; i<strlen(msg) ; i++ )
			glutStrokeCharacter(&glutStrokeRoman, msg[i]);
	glPopMatrix();
	}

 
//----------------------------------------------------------
//
//		drawPeriphery
//
//----------------------------------------------------------
void
Node::drawPeriphery()
	{
	Point3d b,e;

	g_per_bbox.getValue( b, e );

	glBegin( GL_LINE_LOOP );
		glVertex3f( b[0],0,b[2] );
		glVertex3f( b[0],0,e[2] );
		glVertex3f( e[0],0,e[2] );
		glVertex3f( e[0],0,b[2] );
	glEnd();
	}

//----------------------------------------------------------
//
//		drawSelfSpecial
//
//		OpenGL Assumptions:
//
//			Depth test is on
//			Depth mask is false
//			Blending   is on
//			Texturing  is on
//
//----------------------------------------------------------
void
Node::drawSelfSpecial( const Camera &, int draw_neighbors )
	{
/*
	GLboolean wm;
	glGetBooleanv( GL_DEPTH_WRITEMASK, &wm );
	if( wm != GL_FALSE )
		cerr << "dss: Depth Mask not false\n";

	if( glIsEnabled( GL_DEPTH_TEST ) != GL_TRUE )
		cerr << "dss: Depth Test not enabled\n";
	if( glIsEnabled( GL_TEXTURE_2D ) != GL_TRUE )
		cerr << "dss: Texturing not enabled\n";
	if( glIsEnabled( GL_BLEND ) != GL_TRUE )
		cerr << "dss: Blending not enabled\n";
*/

	glDepthMask( GL_TRUE );
	glDisable( GL_BLEND );
	glDisable( GL_TEXTURE_2D );

	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	if( m_multisample )	glEnable(GL_MULTISAMPLE_SGIS); 

	m_bloated_bbox.setGLClipPlanesEps();

	for( int i=0 ; i<m_num_vobjs ; i++ )
		m_vobj_list[i]->draw();

	if( draw_neighbors && g_inflate )
		{
		for( i=0 ; i<m_num_nb_vobjs ; i++ )
			((TriangleList *)m_nb_vobj_list[i])->
				draw_list(m_nb_tlist[i],m_nb_num_tris[i]);

		for( i=0 ; i<m_num_nb_whole ; i++ )
			((TriangleList *)m_nb_whole_list[i])->draw();
		}

	m_bloated_bbox.unsetGLClipPlanes();

	glEnable( GL_TEXTURE_2D );
	glEnable( GL_BLEND );
	glDepthMask( GL_FALSE );

	if( m_multisample )	glDisable(GL_MULTISAMPLE_SGIS);
	}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//
//		Anti aliasing version
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void
Node::drawSelfSpecialAA( const Camera & camera, int draw_neighbors )
	{
	glClear(GL_ACCUM_BUFFER_BIT);
	glDepthMask( GL_TRUE );
	glDisable( GL_BLEND );
	glDisable( GL_TEXTURE_2D );

	for( int j=0 ; j<8 ; j++ )
		{
		camera.setupGLProjectionMoveCameraAntiAlias(j);

		glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
		
		if( m_multisample )	glEnable(GL_MULTISAMPLE_SGIS); 

		m_bloated_bbox.setGLClipPlanesEps();

		for( int i=0 ; i<m_num_vobjs ; i++ )
			m_vobj_list[i]->draw();

		if( draw_neighbors && g_inflate )
			{
			for( i=0 ; i<m_num_nb_vobjs ; i++ )
				((TriangleList *)m_nb_vobj_list[i])->
					draw_list(m_nb_tlist[i],m_nb_num_tris[i]);

			for( i=0 ; i<m_num_nb_whole ; i++ )
				((TriangleList *)m_nb_whole_list[i])->draw();
			}

		m_bloated_bbox.unsetGLClipPlanes();

		if( m_multisample )	glDisable(GL_MULTISAMPLE_SGIS);

		glAccum(GL_ACCUM, 1.0/8.0 );
		}
	glAccum(GL_RETURN, 1.0);

	glEnable( GL_TEXTURE_2D );
	glEnable( GL_BLEND );
	glDepthMask( GL_FALSE );

	camera.setupGLProjectionMoveCamera();
	}

//----------------------------------------------------------
//
//		drawChildrenSpecial
//
//		OpenGL Assumptions:
//
//			Depth test is on
//			Depth mask is flase
//			Blending   is on
//			Texturing  is on
//
//----------------------------------------------------------
void
Node::drawChildrenSpecial( const Camera & camera )
	{
/*
	GLboolean wm;
	glGetBooleanv( GL_DEPTH_WRITEMASK, &wm );
	if( wm != GL_FALSE )
		cerr << "dcs: Depth Mask not false\n";

	if( glIsEnabled( GL_DEPTH_TEST ) != GL_TRUE )
		cerr << "dcs: Depth Test not enabled\n";
	if( glIsEnabled( GL_TEXTURE_2D ) != GL_TRUE )
		cerr << "dcs: Texturing not enabled\n";
	if( glIsEnabled( GL_BLEND ) != GL_TRUE )
		cerr << "dcs: Blending not enabled\n";
*/
	glDepthMask( GL_TRUE );
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	glDepthMask( GL_FALSE );

	if( m_plane.isInFront( camera.getPosition() ) )
		{
		m_back ->traversal_drawImageCachePriv( camera );
		m_front->traversal_drawImageCachePriv( camera );
		}
	else
		{
		m_front->traversal_drawImageCachePriv( camera );
		m_back ->traversal_drawImageCachePriv( camera );
		}
	}

//----------------------------------------------------------
//
//		setupFrustum
//
//----------------------------------------------------------
int
Node::setupFrustum( const BoundBox & bbox, const Camera & camera, double pixel_eps )
	{
	int max[2]; max[0] = MAX_TEX_W; max[1] = MAX_TEX_H;
	int vp[4];
	int vp_delta[2];

	glGetIntegerv( GL_VIEWPORT, vp );

	int status = proj_FeasibleViewport( bbox, pixel_eps+PIXEL_PAD, vp_delta, max );
	//int status = proj_FeasibleViewport( bbox, 0, vp_delta, max );

	if(status != 1)	return 0;

	double *frustum = camera.getFrustum();
	double xfac = (frustum[1]-frustum[0])/(double)vp[2];
	double yfac = (frustum[3]-frustum[2])/(double)vp[3];

	glMatrixMode( GL_PROJECTION );
		glPushMatrix();
		glLoadIdentity();
		glFrustum(  frustum[0] + vp_delta[0]*xfac,
					frustum[1] + vp_delta[0]*xfac,	
					frustum[2] + vp_delta[1]*yfac,
					frustum[3] + vp_delta[1]*yfac,
					frustum[4], frustum[5] );

	glMatrixMode( GL_MODELVIEW );

	return 1;
	}

//----------------------------------------------------------
//
//		setupRotatedLookat
//
//----------------------------------------------------------
void
Node::setupRotatedLookat( const Camera &camera, Matrix4d & matrix,
						  int check_feasibility ) const
	{
	double	mview_matrix[16];
	Point3d centroid	= m_bbox.getCentroid();
	Point3d cam_pos     = camera.getPosition();
	Point3d cam_look    = centroid - cam_pos;

	// assure up vector is colinear with Y axis
	//cam_look[1] = 0.0;
	cam_look.normalize();

	//
	//	Find Near/Far points of object, set hither/yon planes
	//
	double dist = (centroid-cam_pos).length();
	double radius = m_bbox.getRadius();

	camera.setupGLProjectionMoveCamera( dist-radius, dist+radius );

	glGetDoublev( GL_MODELVIEW_MATRIX, mview_matrix );

	if( check_feasibility )
		{
		matrix.setValue( mview_matrix );
		int feasible = proj_FeasibleScreenPolygon( m_bbox, camera, matrix,
				       ((BrickMetric*)m_metric)->getPixelEpsilon()+PIXEL_PAD);

		if( feasible == 1 )
			{
			glMatrixMode( GL_PROJECTION );
				glPopMatrix();
			return;
			}
		else if( feasible == 2 )
			{
			//
			//	Rotate the object until it is just in view.
			//
			}
		}

	//
	//	The UP vector is not robust here.
	//
	glMatrixMode( GL_MODELVIEW );
		glPushMatrix();
			glLoadIdentity();
			gluLookAt( cam_pos[0], cam_pos[1], cam_pos[2],
					   cam_pos[0]+cam_look[0],
					   cam_pos[1]+cam_look[1],
					   cam_pos[2]+cam_look[2],
					   0.0, 1.0, 0.0 );
			glGetDoublev( GL_MODELVIEW_MATRIX, mview_matrix );
		glPopMatrix();
	
	glMatrixMode( GL_PROJECTION );
		glPopMatrix();

	matrix.setValue( mview_matrix );
	return;	
	}

//==========================================================
//
//
//		File I/O Routines
//
//
//==========================================================
typedef struct _instance_tag {
	TriangleList *tobj;
	BoundBox bbox;
	char name[256];
	int num_tris;
	struct _instance_tag *next;
	} Instance;

Instance *inst_island = NULL;
Instance *inst_list = NULL;
Instance *inst_tail = NULL;
int	g_num_instances = 0;
int g_object_count = 0;

Matrix4f	g_imat;

int
search_inst_list( char *filename, Instance **inst )
	{
	Instance *ip = inst_list;
	for( int j=0 ; j<g_num_instances ; j++ )
		{
		if( strcmp( ip->name, filename) == 0 )
			{
			*inst = ip;
			return 1;
			}
		else
			ip = ip->next;
		}

	inst = NULL;
	return 0;
	}

int
add_instance( char *filename, Matrix4f matrix, TriangleList * & tlist )
	{
	//
	//	Add a node to the Instance List
	//
	if( inst_list == NULL )
		{
		inst_list = new Instance();
		inst_tail = inst_list;
		}
	else
		{
		inst_tail->next = new Instance();
		inst_tail = inst_tail->next;
		}

	//	Create the TriangleList Object
	//
	tlist = new TriangleList();
	tlist->setTransform( matrix );

	inst_tail->num_tris = tlist->Read( filename );

#ifdef GEOM_DL
	//-------------------------
	tlist->createDisplayList();
	tlist->deallocRawData();
	//-------------------------
#endif

	inst_tail->next = NULL;
	inst_tail->tobj = tlist;
	inst_tail->bbox = tlist->getObjectSpaceBoundBox();
	strcpy( inst_tail->name, filename );				

	g_num_instances++;

	return( inst_tail->num_tris );
	}

//----------------------------------------------------------
//
//		read_VisualObjects
//
//----------------------------------------------------------
void
Node::read_VisualObjects( FILE* fp )
	{
	int				obj_id,num_constraints,dummy;
	char			obj_filename[256];
	Matrix4f		matrix;
	TriangleList	*tlist;

	m_vobj_list = new VisualObject* [m_num_vobjs];
	m_vobj_inst_tag = new int[m_num_vobjs];

	for( int i=0 ; i<m_num_vobjs ; i++ )
		{
		fscanf( fp,"%d %s", &obj_id, obj_filename );

		matrix.Read( fp );	

		//
		//	Ignore constraints for now.
		//
		fscanf( fp,"%d", &num_constraints );
		for( int j=0 ; j<num_constraints ; j++ )
			fscanf( fp, "%d", &dummy );

		//
		//	TriangleList needs the matrix set before
		//	the vertices are read so that the bounding
		//	box can be transformed properly.
		//

		//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
		//	Search List of instances
		//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
		int found_inst = 0;
		Instance *ip = NULL;

		found_inst = search_inst_list( obj_filename, &ip );
		
		if( found_inst )
			{
			tlist = new TriangleList( *(ip->tobj) );
			tlist->setTransform( matrix );
			tlist->computeWSBB_From_OSBB( ip->bbox );
			//tlist->computeBoundBox();

			//	Share the Geometry Display List !!!
			//
#ifdef GEOM_DL
			tlist->setGeomDispList( ip->tobj->getGeomDispList() );
			tlist->createDisplayList();
#endif
			m_num_tris_local += ip->num_tris;
			m_vobj_inst_tag[i] = 1;
			}

		//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
		//	Check for FLOFF and load in the OFF it references
		//	if needed.
		//
		//	set found_inst to 1 when done.
		//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
		if( !found_inst )
			{
			char	type[256];
			char	ref_filename[256];
			FILE *fp = fopen( obj_filename, "r" );

			if( fp == NULL )
				{
				fprintf(stderr,"Cant read: %s\n", obj_filename );
				exit(1);
				}
			
			fscanf( fp, "%s", type );
			fscanf( fp, "%s", ref_filename );
			fclose(fp);

			if( strcmp( type, "FLOFF") == 0 )
				{
				found_inst = 1;

				//
				//	Scan instance list for the object referred to.
				//
				TriangleList *ref_tlist;
				Instance *ref_ip = NULL;
				int found_ref = search_inst_list( ref_filename, &ref_ip );

				if( ref_ip != NULL )	ref_tlist = ref_ip->tobj;

				if( !found_ref )
					{
					add_instance( ref_filename, g_imat, ref_tlist );
					}

				tlist = new TriangleList();
				tlist->setTransform( matrix );
				tlist->initFLOFF( *ref_tlist, obj_filename );
				tlist->computeBoundBox_FLOFF();

				m_num_tris_local += tlist->getNumTriangles();
				m_vobj_inst_tag[i] = 0;
				}
			}

		//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
		//	Add an Instance to the list
		//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
		if( !found_inst )
			{
			//
			//	Read Tris from a file
			//		
			m_num_tris_local += add_instance( obj_filename, matrix, tlist );
			m_vobj_inst_tag[i] = 0;
			}

		m_vobj_list[i] = tlist;						

		if( i==0 )	m_bbox = tlist->getBoundBox();
		else		m_bbox = m_bbox.superBox( tlist->getBoundBox() );
		}

	m_bloated_bbox = m_bbox;

	return;
	}

//----------------------------------------------------------
//
//
//
//----------------------------------------------------------
int
Node::read_subtree( FILE* fp, double pixel_eps, const Camera &camera )
	{
	double	A,B,C,D;

	fscanf( fp, "%lf %lf %lf %lf", &A,&B,&C,&D );
	fscanf( fp, "%d", &m_num_vobjs );

	m_plane.setValue( A,B,C,D );
	
	switch( m_metric_type )
		{
		case AngleDistance:
			m_metric = new AngleDistanceMetric();
			break;
	
		case Corners:
			m_metric = new CornersMetric( pixel_eps );
			break;
	
		case Brick:
			m_metric = new BrickMetric( pixel_eps, camera );
			break;

		default:
			m_metric = NULL;
			break;
		}

	if( m_num_vobjs != 0 )
		{
		read_VisualObjects( fp );
		m_height = 0;
		}
	else
		{
		m_back  = new Node();
		m_front = new Node();

		m_back->m_parent = this;
		m_back->m_tag	 = NODE_BACK;
		int bh = m_back->read_subtree( fp, pixel_eps, camera );

		m_front->m_parent = this;
		m_front->m_tag    = NODE_FRONT;
		int fh = m_front->read_subtree( fp, pixel_eps, camera );

		m_bbox = m_back->m_bbox;
		m_bbox = m_bbox.superBox( m_front->m_bbox );

		m_bloated_bbox = m_bbox;

		m_num_tris_local = m_back->m_num_tris_local +
					       m_front->m_num_tris_local;

		if( bh > fh )	m_height = bh + 1;
		else			m_height = fh + 1;
		}

	return m_height;

//	m_bbox.Print();
	}

//----------------------------------------------------------
//
//		Read
//
//----------------------------------------------------------

//
//	This assumes that every node has either children
//	or visual objects. No NULL nodes. Full Tree (?)
//
int
Node::Read( char* filename, double pixel_eps, int inflate, const Camera & camera )
	{
	FILE *fp;
	int		Max_ID;	

	g_inflate = inflate;

	fp = fopen( filename, "r" );
	if( fp == NULL )
		{
		fprintf( stderr,"Could not find file %s.\n", filename );
		return 0;
		}

	fscanf( fp, "%d", &Max_ID );

	read_subtree( fp, pixel_eps, camera );

	g_root = this;
	g_radius = (g_root->getBoundBox()).getRadius();

	//
	//	Dellocate the memory holding the raw triangle data
	//
/*
	Instance *ip = inst_list;
	for( int i=0 ; i<g_num_instances ; i++ )
		{
		ip->tobj->deallocRawData();
		ip = ip->next;
		}
*/

	//
	//	Set up the bounding box +10% for periphery drawing
	//
	Point3d min = m_bbox.getMin();
	Point3d max = m_bbox.getMax();
	double xlen = max[0] - min[0];
	double zlen = max[2] - min[2];

	g_per_bbox.setValue( Point3d(min[0] - 0.05*xlen, 0, min[2] - 0.05*zlen),
						 Point3d(max[0] + 0.05*xlen, 0, max[2] + 0.05*zlen)  );

	g_bbox.setValue( min, max );

	search_inst_list( "./island.off", &inst_island );
	
	return 1;
	}

//----------------------------------------------------------
//
//		drawWireFrame
//
//----------------------------------------------------------
void
Node::drawWireFrame()
	{
	if( m_num_vobjs == 0 )
		{
		m_back->drawWireFrame();
		m_front->drawWireFrame();
		return;
		}

	if( m_annotation != NODE_GEOMETRY )
		{
		((TriangleList*)m_vobj_list[m_num_vobjs-1])->draw_wire();
		}
	}
