//============================================================================
//
//		Vector3f
//
//============================================================================

#include <stdio.h>
#include <math.h>
#include "Vector3d.h"
#include "Vector3f.h"
#include "Vector4f.h"

#define	EPSILON		1e-8
#define EPSILON2	1e-16

//----------------------------------------------------------------------------
//
//		Constructors !!
//
//
//----------------------------------------------------------------------------
Vector3f::Vector3f()
	{
	x = y = z = 0.0;
	}

Vector3f::Vector3f( const float v[3] )
	{
	x = v[0]; y = v[1]; z = v[2];
	}

Vector3f::Vector3f( float xx, float yy, float zz )
	{
	x = xx; y = yy; z = zz;
	}

Vector3f::Vector3f( const Vector3f &v )
	{
	x = v[0];
	y = v[1];
	z = v[2];
	}

Vector3f::Vector3f( const Vector4f &v )
	{
	x = v[0];
	y = v[1];
	z = v[2];
	}

//----------------------------------------------------------------------------
//
//	Set Value
//
//----------------------------------------------------------------------------
Vector3f &	Vector3f::setValue( const Vector3f &v )
	{
	x = v.x; y = v.y; z = v.z;

	return *this;
	}

Vector3f &	Vector3f::setValue( const float v[3] )
	{
	x = v[0]; y = v[1]; z = v[2];

	return *this;
	}

Vector3f &	Vector3f::setValue( float xx, float yy, float zz )
	{
	x = xx; y = yy; z = zz;

	return *this;
	}

//----------------------------------------------------------------------------
//
//		getValue
//
//----------------------------------------------------------------------------
const float*	Vector3f::getValue() const
	{
	return( (float *)this );
	}


void	Vector3f::getValue( float &xx, float &yy, float &zz ) const
	{
	xx = x; yy = y; zz = z;
	}

float &	Vector3f::operator[]( int i )
	{
	return( ((float *) this)[i] );
	}

const float &	Vector3f::operator[]( int i ) const
	{
	return( ((float *) this)[i] );
	}


//----------------------------------------------------------------------------
//
//		dot
//
//----------------------------------------------------------------------------
float	Vector3f::dot( const Vector3f &v ) const
	{
	return( x*v.x + y*v.y + z*v.z );
	}

//----------------------------------------------------------------------------
//
//		cross
//
//----------------------------------------------------------------------------
Vector3f	Vector3f::cross( const Vector3f &v ) const
	{
	Vector3f vec (  y*v.z - z*v.y,
					z*v.x - x*v.z,
					x*v.y - y*v.x );

	return vec;
	}



//----------------------------------------------------------------------------
//
//		rotateY
//
//----------------------------------------------------------------------------
void	Vector3f::rotateY( float angle )
	{
	float tx,ty,tz;

	tx =  x * cos( angle ) + z * sin( angle );
	ty =  y;
	tz = -x * sin( angle ) + z * cos( angle );

	x = tx; y = ty; z = tz;
	}


//----------------------------------------------------------------------------
//
//		length
//
//----------------------------------------------------------------------------
float	Vector3f::length() const
	{
	return( sqrt( x*x + y*y + z*z ) );
	}

//----------------------------------------------------------------------------
//
//		negate
//
//----------------------------------------------------------------------------
void	Vector3f::negate()
	{
	x = -x; y = -y; z = -z;
	}

//----------------------------------------------------------------------------
//
//		normalize
//
//----------------------------------------------------------------------------
float	Vector3f::normalize()
	{
	float length = sqrt( x*x + y*y + z*z );

	x /= length;
	y /= length;
	z /= length;

	return( length );
	}

//----------------------------------------------------------------------------
//
//		lesserComponents
//
//----------------------------------------------------------------------------
Vector3f
Vector3f::lesserComponents( const Vector3f &v )
	{
	Vector3f p;

	p.x = ( x < v.x ) ? x : v.x;
	p.y = ( y < v.y ) ? y : v.y;
	p.z = ( z < v.z ) ? z : v.z;

	return p;
	}


//----------------------------------------------------------------------------
//
//		greaterComponents
//
//----------------------------------------------------------------------------
Vector3f
Vector3f::greaterComponents( const Vector3f &v )
	{
	Vector3f p;

	p.x = ( x > v.x ) ? x : v.x;
	p.y = ( y > v.y ) ? y : v.y;
	p.z = ( z > v.z ) ? z : v.z;

	return p;
	}


//----------------------------------------------------------------------------
//
//		+=
//
//----------------------------------------------------------------------------
Vector3f &	Vector3f::operator += ( const Vector3f &v )
	{
	x += v.x;
	y += v.y;
	z += v.z;

	return *this;
	}

//----------------------------------------------------------------------------
//
//		-=
//
//----------------------------------------------------------------------------
Vector3f &	Vector3f::operator -= ( const Vector3f &v )
	{
	x -= v.x;
	y -= v.y;
	z -= v.z;

	return *this;
	}

//----------------------------------------------------------------------------
//
//		*=
//
//----------------------------------------------------------------------------
Vector3f &	Vector3f::operator *= ( const Vector3f &v )
	{
	x *= v.x;
	y *= v.y;
	z *= v.z;

	return *this;
	}

Vector3f &	Vector3f::operator *= ( float d )
	{
	x *= d;
	y *= d;
	z *= d;

	return *this;
	}

//----------------------------------------------------------------------------
//
//		/=
//
//----------------------------------------------------------------------------
Vector3f &	Vector3f::operator /= ( float d )
	{
	x /= d;
	y /= d;
	z /= d;

	return *this;
	}

//----------------------------------------------------------------------------
//
//		Non-destructive unary negation
//
//----------------------------------------------------------------------------
Vector3f	Vector3f::operator - () const
	{
	Vector3f vec( -x, -y, -z );
	
	return vec;
	}

//----------------------------------------------------------------------------
//
//		Print
//
//----------------------------------------------------------------------------
void	Vector3f::Print() const
	{
	fprintf( stderr,"%6.2f %6.2f %6.2f\n", x,y,z );
	}

//----------------------------------------------------------------------------
//
//		Friend operators
//
//----------------------------------------------------------------------------

Vector3f	operator *( const Vector3f &v1, const Vector3f &v2 )
	{
	Vector3f vec( v1.x*v2.x, v1.y*v2.y, v1.z*v2.z );

	return vec;
	}

Vector3f	operator *( const Vector3f &v, float d )
	{
	Vector3f vec( v.x*d, v.y*d, v.z*d );

	return vec;
	}

Vector3f	operator *( float d, const Vector3f &v )
	{
	Vector3f vec( v.x*d, v.y*d, v.z*d );

	return vec;
	}

Vector3f	operator /( const Vector3f &v, float d )
	{
	Vector3f vec( v.x/d, v.y/d, v.z/d );

	return vec;
	}

Vector3f	operator +( const Vector3f &v1, const Vector3f &v2  )
	{
	Vector3f vec( v1.x+v2.x, v1.y+v2.y, v1.z+v2.z );

	return vec;
	}

Vector3f	operator -( const Vector3f &v1, const Vector3f &v2  )
	{
	Vector3f vec( v1.x-v2.x, v1.y-v2.y, v1.z-v2.z );

	return vec;
	}

int	operator ==( const Vector3f &v1, const Vector3f &v2  )
	{
	float	dx = v1.x - v2.x;
	float	dy = v1.y - v2.y;
	float	dz = v1.z - v2.z;

	if( fabs(dx) > EPSILON ||
		fabs(dy) > EPSILON ||
		fabs(dx) > EPSILON   )	return 0;

	return (dx*dx + dy*dy + dz*dz) <= EPSILON2;
	}

