#include <GL/gl.h>
#include <GL/glu.h>
#include <math.h>
#include "projection.h"

#define NUM_TEX_SIZES 8

int power2[NUM_TEX_SIZES] = {16,32,64,128,256,512,1024,2048};

//----------------------------------------------------------------------------
//
//		Find power of 2 widths
//
//----------------------------------------------------------------------------
void	proj_GetPower2Size( int ow, int oh, int & nw, int & nh )
	{
	nw = 2048;
	nh = 2048;

	//	Smallest texture allowed is 64x64
	for( int i=0 ; i<NUM_TEX_SIZES ; i++ )
		if( ow <= power2[i] )
			{
			nw = power2[i];
			break;
			}

	for( i=0 ; i<NUM_TEX_SIZES ; i++ )
		if( oh <= power2[i] )
			{
			nh = power2[i];
			break;
			}
	}


//----------------------------------------------------------------------------
//
//		Get bounding polygon of a bounding box projected onto
//		the front plane of the viewing frustum.
//
//		The current implementation uses the gluProject and
//		gluUnProject functions.
//
//		NO ERROR CHECKING
//
//----------------------------------------------------------------------------
int
proj_GetScreenPolygon(  const BoundBox & bbox, 
						double extra_pixels, int wbox[2][2], double plane[4][3] )
	{
	GLint	viewport[4];
	double	mview_matrix[16];
	double	proj_matrix[16];
	double	x,y,z,tx,ty,bx,by;

	//
	//	Grab some matrices from GL
	//	
	glGetDoublev( GL_MODELVIEW_MATRIX, mview_matrix );
	glGetDoublev( GL_PROJECTION_MATRIX, proj_matrix );
	glGetIntegerv( GL_VIEWPORT, viewport );

	//
	//		Iterate over the 8 corners of the bounding box
	//
	Point3d p;
	for( int i=0 ; i<8 ; i++ )
		{
		p = bbox.getPoint( i );
		int status = gluProject( p[0], p[1], p[2],
					  mview_matrix, proj_matrix, viewport, &x, &y, &z ); 

		if( i==0 )
			{
			bx = tx = x;
			by = ty = y;
			}
		else
			{
			if (x < bx) bx = x;
			if (x > tx) tx = x;
			if (y < by) by = y;
			if (y > ty) ty = y;
			}
		}

	bx -= extra_pixels;
	by -= extra_pixels;
	tx += extra_pixels;
	ty += extra_pixels;

/*
	// Out of frustum - Left
	if (tx < viewport[0] )	return( -1 );
	if (bx < viewport[0] )	return( 0 );

	// Out of frustum - Bottom
	if (ty < viewport[1] )	return( -1 );
	if (by < viewport[1] )	return( 0 );

	// Out of frustum - Right
	if (bx > viewport[2]+viewport[0] )	return( -1 );
	if (tx > viewport[2]+viewport[0] )	return( 0 );

	// Out of frustum - Top
	if (by > viewport[3]+viewport[1] )	return( -1 );
	if (ty > viewport[3]+viewport[1] )	return( 0 );
*/

	//
	//	Counter clockwise polygon: project the centroid and
	//	use the "projected z" as the basis to unproject the
	//	four corners of the polygon.
	//
	Point3d centroid = bbox.getCentroid();
	
	gluProject( centroid[0], centroid[1], centroid[2],
				mview_matrix, proj_matrix, viewport, &x, &y, &z );

	wbox[0][0] = bx = (int) floor(bx+0.5l);
	wbox[0][1] = by = (int) floor(by+0.5l);
	wbox[1][0] = tx = (int) floor(tx+0.5l);
	wbox[1][1] = ty = (int) floor(ty+0.5l);

	//fprintf(stderr,"proj: %f %f %f %f\n",bx,tx,by,ty);

	gluUnProject(bx, by, z,
				mview_matrix, proj_matrix, viewport,
				&(plane[0][0]), &(plane[0][1]), &(plane[0][2]) );

	gluUnProject(tx, by, z,
				mview_matrix, proj_matrix, viewport,
				&(plane[1][0]), &(plane[1][1]), &(plane[1][2]) );

	gluUnProject(tx, ty, z,
				mview_matrix, proj_matrix, viewport,
				&(plane[2][0]), &(plane[2][1]), &(plane[2][2]) );

	gluUnProject(bx, ty, z,
				mview_matrix, proj_matrix, viewport,
				&(plane[3][0]), &(plane[3][1]), &(plane[3][2]) );

	return( 1 );
	}

//----------------------------------------------------------------------------
//
//		Determine if the bounding polygon of a bounding box projected onto
//		the front plane of the viewing frustum lies entirely within the
//		viewport.
//
//		The current implementation uses the gluProject and
//		gluUnProject functions.
//
//----------------------------------------------------------------------------
int
proj_FeasibleViewport(	const BoundBox & bbox,
						double extra_pixels, int vp_delta[2], int max[2] )
	{
	//
	//	Grab some matrices from GL
	//
	GLint	viewport[4];
	double	proj_matrix[16];
	double	mview_matrix[16];
	glGetDoublev( GL_MODELVIEW_MATRIX, mview_matrix );
	glGetDoublev( GL_PROJECTION_MATRIX, proj_matrix );
	glGetIntegerv( GL_VIEWPORT, viewport );

	//
	//		Iterate over the 8 corners of the bounding box
	//
	Point3d	 	p;
	double	x,y,z,tx,ty,bx,by;

	for( int i=0 ; i<8 ; i++ )
		{
		p = bbox.getPoint( i );
		gluProject( p[0], p[1], p[2],
					mview_matrix, proj_matrix, viewport, &x, &y, &z ); 

		if( i==0 )
			{
			bx = tx = x;
			by = ty = y;
			}
		else
			{
			if (x < bx) bx = x;
			if (x > tx) tx = x;
			if (y < by) by = y;
			if (y > ty) ty = y;
			}
		}

	bx -= extra_pixels;
	by -= extra_pixels;
	tx += extra_pixels;
	ty += extra_pixels;

	bx = (int) floor(bx+0.5l);
	by = (int) floor(by+0.5l);
	tx = (int) floor(tx+0.5l);
	ty = (int) floor(ty+0.5l);
	
	vp_delta[0] = (int) bx;
	vp_delta[1] = (int) by;

	int ret_val = 1;

	if( (int)(tx-bx) > max[0] )	ret_val = 0;
	if( (int)(ty-by) > max[1] )	ret_val = 0;

	//fprintf( stderr,"%d: %f %f %f %f: %d %d: %d %d\n",
	//		ret_val,bx,by,tx,ty,vp_delta[0],vp_delta[1],max[0],max[1] );

	return( ret_val );
	}

//----------------------------------------------------------------------------
//
//		Determine if the bounding polygon of a bounding box projected onto
//		the front plane of the viewing frustum lies entirely within the
//		viewport.
//
//		The current implementation uses the gluProject and
//		gluUnProject functions.
//
//----------------------------------------------------------------------------
int
proj_FeasibleScreenPolygon( const BoundBox & bbox,  const Camera & camera,
							const Matrix4d & mview, double extra_pixels )
	{
	GLint	viewport[4];
	double	proj_matrix[16];
	double	x,y,z,tx,ty,bx,by;

	Point3d	 	p;
	Point3d		cam_pos;

	//
	//	Grab some matrices from GL
	//	
	glGetDoublev( GL_PROJECTION_MATRIX, proj_matrix );
	glGetIntegerv( GL_VIEWPORT, viewport );

	//
	//		Iterate over the 8 corners of the bounding box
	//
	cam_pos  = camera.getPosition();

	int nonvalid = 1;

	for( int i=0 ; i<8 ; i++ )
		{
		p = bbox.getPoint( i );
		gluProject( p[0], p[1], p[2],
					mview.getValue(), proj_matrix, viewport, &x, &y, &z ); 

		if( nonvalid )
			if( x>=0.0  &&  x<=viewport[2] &&
			    y>=0.0  &&  y<=viewport[3]    ) nonvalid = 0;

		if( i==0 )
			{
			bx = tx = x;
			by = ty = y;
			}
		else
			{
			if (x < bx) bx = x;
			if (x > tx) tx = x;
			if (y < by) by = y;
			if (y > ty) ty = y;
			}
		}

	if( nonvalid )	return 0;

	bx -= extra_pixels;
	by -= extra_pixels;
	tx += extra_pixels;
	ty += extra_pixels;

	// Out of frustum - Left
	if (tx < 0)				return( 2 );
	if (bx < 0)				return( 2 );

	// Out of frustum - Bottom
	if (ty < 0)				return( 2 );
	if (by < 0)				return( 2 );

	// Out of frustum - Right
	if (bx > viewport[2])	return( 2 );
	if (tx > viewport[2])	return( 2 );

	// Out of frustum - Top
	if (by > viewport[3])	return( 2 );
	if (ty > viewport[3])	return( 2 );

	return( 1 );
	}

//----------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------
void
proj_GetActualScreenCorners( const BoundBox & bbox, const Matrix4d & mview,
							 double sc[8][2] )
	{
	GLint	viewport[4];
	double	proj_matrix[16];
	double	z;
	Point3d	p;

	//
	//	Grab some matrices from GL
	//	
	glGetDoublev( GL_PROJECTION_MATRIX, proj_matrix );
	glGetIntegerv( GL_VIEWPORT, viewport );

	//
	//		Iterate over the 8 corners of the bounding box
	//
	//
	for( int i=0 ; i<8 ; i++ )
		{
		p = bbox.getPoint( i );
		gluProject( p[0], p[1], p[2],
					mview.getValue(), proj_matrix, viewport,
					&sc[i][0], &sc[i][1], &z ); 

//		if( status == GL_FALSE )
//			fprintf(stderr,"Actual::gluProject FAILED !!!\n");
		}

	return;
	}

//----------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------
void
proj_GetApproxScreenCorners( double wc[8][3], const Matrix4d & mview,
							 double sc[8][2] )
	{
	GLint	viewport[4];
	double	proj_matrix[16];
	double	z;

	//
	//	Grab some matrices from GL
	//	
	glGetDoublev( GL_PROJECTION_MATRIX, proj_matrix );
	glGetIntegerv( GL_VIEWPORT, viewport );

	//
	//		Iterate over the 8 corners of the bounding box
	//
	//
	int status;
	for( int i=0 ; i<8 ; i++ )
		{
		status = gluProject( wc[i][0], wc[i][1], wc[i][2],
							mview.getValue(), proj_matrix, viewport,
							&sc[i][0], &sc[i][1], &z ); 

//		if( status == GL_FALSE )
//			fprintf(stderr,"Approx::gluProject FAILED !!!\n");
		}

	return;
	}



//----------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------
void
proj_GetActualWorldCorners( const BoundBox & bbox, const Matrix4d & mview,
							double sc[8][2], double wc[8][3] )
	{
	GLint	viewport[4];
	double	proj_matrix[16];
	double	x,y,z;
	Point3d	centroid = bbox.getCentroid();

	//
	//	Grab some matrices from GL
	//	
	glGetDoublev( GL_PROJECTION_MATRIX, proj_matrix );
	glGetIntegerv( GL_VIEWPORT, viewport );
	
	//
	//	Counter clockwise polygon: project the centroid and
	//	use the "projected z" as the basis to unproject the
	//	eight corners of the bounding box onto the 3D polygon.
	//
	gluProject( centroid[0], centroid[1], centroid[2],
				mview.getValue(), proj_matrix, viewport, &x, &y, &z );

	int status;
	for( int i=0 ; i<8 ; i++ )
		{
		status = gluUnProject( sc[i][0], sc[i][1], z,
								mview.getValue(), proj_matrix, viewport,
								&(wc[i][0]), &(wc[i][1]), &(wc[i][2]) );
//		if( status == GL_FALSE )
//			fprintf(stderr,"gluUnProject FAILED !!!\n");
		}
	return;
	}


//----------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------
int
proj_GetWorldCornersAndScreenPolygon( const BoundBox & bbox, double wc[8][3], int sp[4][2] )
	{
	GLint	viewport[4];
	double	proj_matrix[16];
	double	mview_matrix[16];
	double	x,y,z;
	double	sc[8][2];
	Point3d	p,centroid = bbox.getCentroid();

	//
	//	Grab some matrices from GL
	//	
	glGetDoublev( GL_MODELVIEW_MATRIX, mview_matrix );
	glGetDoublev( GL_PROJECTION_MATRIX, proj_matrix );
	glGetIntegerv( GL_VIEWPORT, viewport );

	//
	//		Iterate over the 8 corners of the bounding box
	//
	//
	for( int i=0 ; i<8 ; i++ )
		{
		p = bbox.getPoint( i );
		gluProject( p[0], p[1], p[2],
					mview_matrix, proj_matrix, viewport,
					&sc[i][0], &sc[i][1], &z ); 
		}
	
	//
	//	Counter clockwise polygon: project the centroid and
	//	use the "projected z" as the basis to unproject the
	//	eight corners of the bounding box onto the 3D polygon.
	//
	gluProject( centroid[0], centroid[1], centroid[2],
				mview_matrix, proj_matrix, viewport, &x, &y, &z );

	int status;
	for( i=0 ; i<8 ; i++ )
		{
		status = gluUnProject( sc[i][0], sc[i][1], z,
								mview_matrix, proj_matrix, viewport,
								&(wc[i][0]), &(wc[i][1]), &(wc[i][2]) );
		}

	//
	//	Sort 8 screen points to find screen polygon
	//
	double bx,by,tx,ty;

	for( i=0 ; i<8 ; i++ )
		{
		x = sc[i][0];
		y = sc[i][1];

		if( i==0 )
			{
			bx = tx = x;
			by = ty = y;
			}
		else
			{
			if (x < bx) bx = x;
			if (x > tx) tx = x;
			if (y < by) by = y;
			if (y > ty) ty = y;
			}
		}

/*
	// Out of frustum - Left
	if (tx < 0) return( 0 );
	if (bx < 0) return( 0 );


	// Out of frustum - Bottom
	if (ty < 0) return( 0 );
	if (by < 0) return( 0 );


	// Out of frustum - Right
	if (bx > viewport[2]) return( 0 );
	if (tx > viewport[2]) return( 0 );


	// Out of frustum - Top
	if (ty > viewport[3]) return( 0 );
	if (by > viewport[3]) return( 0 );
*/

	sp[0][0] = bx = (int) floor(bx+0.5l);
	sp[0][1] = by = (int) floor(by+0.5l);
	sp[1][0] = tx = (int) floor(tx+0.5l);
	sp[1][1] = ty = (int) floor(ty+0.5l);

	if( (int)(tx-bx) > viewport[2] )	return 0;
	if( (int)(ty-by) > viewport[3] )	return 0;
	
	return 1;
	}
