//============================================================================
//
//		readdb.c
//
//
//			Jonathan Shade
//
//============================================================================

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "readdb.h"
#include "Node.h"
#include "Geometry.h"
#include "TriangleList.h"
#include "Billboard.h"


//----------------------------------------------------------------------------
//
//	Read all of the characters until EOL
//
//----------------------------------------------------------------------------
void
FinishLine( FILE *fp )
	{
	char achar;

	do
		{ 
    	fscanf(fp,"%c",&achar);
  		}
	while (achar != '\n');
	}


//----------------------------------------------------------------------------
//
//	Read a bounding box from a LIST file.
//
//----------------------------------------------------------------------------
void
Read_BBox( FILE *dbfile, BoundBox & bbox )
	{
	Point3d	min, max;

	fscanf( dbfile, "{ VECT" );		FinishLine(dbfile);
	fscanf( dbfile, "1 2 0"	);		FinishLine(dbfile);
	fscanf( dbfile, "2" );			FinishLine(dbfile);
	fscanf( dbfile, "0" );			FinishLine(dbfile);

	fscanf( dbfile, "%lf %lf %lf %lf %lf %lf", &min[0],&min[1],&min[2],
											   &max[0],&max[1],&max[2] );
	FinishLine(dbfile);

	fscanf( dbfile, "}" );
	FinishLine(dbfile);

	bbox.setValue( min, max );
	}
	

//----------------------------------------------------------------------------
//
//	Read the transformation matrix from a LIST file.
//
//----------------------------------------------------------------------------
int
Read_xform( FILE * dbfile, Matrix4d & m )
	{
	int status;
	double x[16];

	status = fscanf( dbfile, "{ = INST"	);
	if( status == EOF )
		return( 0 );
	FinishLine(dbfile);
	

	fscanf( dbfile, "transform"	);		FinishLine(dbfile);

	fscanf( dbfile, "%lf %lf %lf %lf", &x[0], &x[4], &x[8],  &x[12] );
	fscanf( dbfile, "%lf %lf %lf %lf", &x[1], &x[5], &x[9],  &x[13] );
	fscanf( dbfile, "%lf %lf %lf %lf", &x[2], &x[6], &x[10], &x[14] );
	fscanf( dbfile, "%lf %lf %lf %lf", &x[3], &x[7], &x[11], &x[15] );

	FinishLine(dbfile);

	m.setValue( x );

	return( 1 );
	}

//----------------------------------------------------------------------------
//
//		Read_Num_Children
//
//		This is a chessy function that searches through
//		the file in a brute force way to find the number
//		of "INST"s. It would be nicier to use grep or
//		something similar.
//
//----------------------------------------------------------------------------
int
Read_Num_Children( char *fileName )
	{
	int		count = 0;
	int		not_done = 1;
	char	str[256];
	BoundBox	bbox;
	Matrix4d	xform;

	FILE	*dbfile = fopen( fileName, "r" );	

	if ( dbfile == NULL )
		return( 0 );

	fscanf( dbfile, "LIST" );	FinishLine(dbfile);

	Read_BBox( dbfile, bbox );

	not_done = Read_xform( dbfile, xform );
	while( not_done )
		{
		count++;
		fscanf( dbfile, "unit { < %s }", str );		FinishLine(dbfile);
		fscanf( dbfile, "}" );						FinishLine(dbfile);
	
		not_done = Read_xform( dbfile, xform );
		}

	fclose( dbfile );

	return( count );
	}
//----------------------------------------------------------------------------
//
//	Assume that "root" has an implicit xform of the identity matrix.
//    
//----------------------------------------------------------------------------
void
buildHierarchy( Node **root,    Matrix4d rootXform,
				char *fileName, int currentLevel )
	{
	//
	//	These are static only to instance a single
	//	tree among many nodes.
	//
	Node		*childNode;
	static VisualObject *vobj;

	int			i, numChildren;
	char		str[256];

	static char     prevfile[4][256];
	static VisualObject *prevvobj[4];
	static int      iPrevObjs = 0;

	Matrix4d	xform;
	Point3d		min,max;
	BoundBox	bbox;
	Point3d		p;
	Normal3d	n;	

	Billboard	*billboard;
	FILE		*dbfile = NULL;

	dbfile = fopen( fileName, "r" );	

	//	If currentLevel > 1, then this is an internal
	//	node. We should allocate numChildren children.
	//
	//	Else, this is a leaf, allocate 0 children.
	//
	if ( dbfile == NULL )
		return;

	fscanf( dbfile, "LIST" );	FinishLine(dbfile);

	// 
	//	Internal node
	//
	if ( currentLevel > 1 )
		{
		//
		//	Scan the file for the number of children
		//
		numChildren = Read_Num_Children( fileName );

		//
		//	Read the Bounding Box
		//
		Read_BBox( dbfile, bbox );

		//
		//	Allocate numChildren children
		//	
		(*root)->Init( numChildren, rootXform, bbox );

		//
		//	Allocate a NULL representation
		//
		billboard = new Billboard();
		(*root)->setBillboard( billboard );
		

		//
		//	Process each of the children in the file
		//
		for ( i=0 ; i<numChildren ; i++ )
			{
			//
			//	Read the transformation matrix
			//
			Read_xform( dbfile, xform );
		
			fscanf( dbfile, "unit { < %s }", str );		FinishLine(dbfile);
			fscanf( dbfile, "}" );						FinishLine(dbfile);

			min.setValue( 0,0,0 );
			max.setValue( 1,1,1 );
			bbox.setValue( min, max );

			(*root)->addChild( 0, xform, bbox );
			childNode = (*root)->getChild( i );

			if ( childNode == NULL )
				fprintf(stderr,"BIG ERROR\n");

			buildHierarchy( &childNode, xform, str, currentLevel-1 );
			}

		}	// if
	else
		{
		//	Leaf Node
		Read_BBox( dbfile, bbox );

		(*root)->Init( 0, rootXform, bbox );

		Read_xform( dbfile, xform );

		fscanf( dbfile, "unit { < %s }", str );		FinishLine(dbfile);
		fscanf( dbfile, "}" );						FinishLine(dbfile);
		fclose( dbfile );

		//
		//	Read in the OFF data
		//
		fprintf(stderr,"Parsing ... %s,\n\t\t%s \n", str, fileName );
		for(i=0; i<iPrevObjs; i++)
		  {
		      if (strcmp(str,prevfile[i]) == 0)
			{
			    vobj = prevvobj[i];
			    break;
			}
		  }
		if (i==iPrevObjs)
		  {
		      Read_OFF( str, &vobj );
		      if (iPrevObjs < 4)
			{
			    strcpy(prevfile[iPrevObjs],str);
			    prevvobj[iPrevObjs] = vobj;
			    iPrevObjs++;
			}
		  }
		(*root)->setVisualObject( vobj );


		//
		//	This does not need to be done here
		//
		billboard = new Billboard();
		(*root)->setBillboard( billboard );	
		}

	fclose( dbfile );
	}

//----------------------------------------------------------------------------
//
//		Read an OFF ( or NOFF ) file from disk.
//    
//----------------------------------------------------------------------------
void	Read_OFF( char *file_name, VisualObject **vobj )
	{
	int		numPoints;
	int		numEdges;
	int		numTris;
	int		i,t1,t2,t3,junk;
	double             red, green, blue;
	char	        off_type_string[80];
	int		off_type;

	Point3d		p;
	Normal3d	n;	

	FILE	*dbfile;
	TriangleList *tlist;

	dbfile = fopen( file_name, "r" );

	if( dbfile == NULL )
		{
		fprintf(stderr," error: cant open %s\n", file_name );
		exit( 0 );
		}

	fscanf( dbfile, "%s", off_type_string );	FinishLine(dbfile);


	if(      strcmp( off_type_string, "OFF"  ) == 0 ) off_type = 0;
	else if( strcmp( off_type_string, "NOFF" ) == 0 ) off_type = 1;

	fscanf( dbfile, "%d %d %d", &numPoints, &numTris, &numEdges );

	tlist = new TriangleList( numTris, numPoints );

	if( off_type == 0 )
		tlist->setCreateNormals( 1 );

	//
	//	Read the Vertices
	//
	n.setValue(0.0,1.0,0.0);

	for( i=0 ; i<numPoints ; i++ )
		{
		switch( off_type )
			{
			//	OFF
			case 0:
				fscanf(dbfile,"%lf %lf %lf", &p[0],&p[1],&p[2] );		
				break;

			//	NOFF
			case 1:
				fscanf(dbfile,"%lf %lf %lf %lf %lf %lf",
								&p[0],&p[1],&p[2],
								&n[0],&n[1],&n[2] );
				break;
			};

		tlist->setVertex( i, p, n );
		}

	//
	//	Read the Triangles
	//
	for( i=0 ; i<numTris ; i++ )
	{
		fscanf(dbfile,"%d %d %d %d",&junk,&t1,&t2,&t3 );
		tlist->setTriangle( i, t1, t2, t3 );
		if ((int)fgetc(dbfile) != 10)
	        {
		    fscanf(dbfile,"%lf %lf %lf",&red,&green,&blue);
		    tlist->setTriangleColor(i, red, green, blue);
	        }
		else
		{
		    tlist->setTriangleColor( i, 0, 128, 0);
	        }
	}

	*vobj = tlist;
}

//----------------------------------------------------------------------------
//
//	Read in the file that describes the terrain.
//
//
//----------------------------------------------------------------------------
void	Read_Terrain( char *file_name, VisualObject **vobj )
	{
	int i;
	char off_file[256];
	double mat[16];
	FILE *dbfile;

	dbfile	= fopen( file_name, "r" );

	if ( dbfile == NULL )
		return;

	for( i=0 ; i<16 ; i++ )
		fscanf( dbfile, "%lf", &mat[i] );

	fscanf( dbfile, "%s", off_file		);

	fclose( dbfile );

	fprintf(stderr,"Reading Terrain File: %s\n",off_file);

	Read_OFF( off_file, vobj );	

	((TriangleList *) *vobj)->setMaterial( mat );

	}

//----------------------------------------------------------------------------
//
//	Read in the file that describes the overall
//		structure of the OOGL files that make
//		up the database.
//
//
//----------------------------------------------------------------------------
int
readDescFile( char *dbfile, char list_file[256],
			  int *num_levels, char terr_file[256] )
	{
	FILE *desc_file = NULL;

	desc_file = fopen( dbfile, "r" );

	if ( desc_file == NULL )
		return 0;

	fscanf( desc_file, "%s", list_file		);
	fscanf( desc_file, "%d", num_levels		);
	fscanf( desc_file, "%s", terr_file		);
	fclose( desc_file );

	return 1;
	}



//----------------------------------------------------------------------------
//
//	Build a hierarchical scene database in root
//		from the data in files described by dbfile.
//
//
//----------------------------------------------------------------------------
void
readDB(char *dbfile, Node **root, VisualObject **background )
	{
	Matrix4d	xform;
	char		LISTFileName[256];
	char		TERRFileName[256];
	int			numLevels		= 0;

	//
	//	Read the high level description
	//
	if ( readDescFile( dbfile, LISTFileName, &numLevels, TERRFileName ) == 0 )
		{
		fprintf( stderr, "Cannot find %s.\n", dbfile );
		return;
		}

	//
	//	Traverse the LIST,OFF files, building 
	//		up the hierarchy, in the Node structure.
	//
	*root = new Node();

	xform.Identity();
	buildHierarchy( root, xform, LISTFileName, numLevels );

	//
	//	Read in the static background object
	//
	Read_Terrain( TERRFileName, background );
	}
