
//============================================================================
//
//	Scene.c
//
//		This file contains all of the "drawing" routines
//		that are called to update the rendering in an
//		OpenGL window of the Walkthru app.
//
//		scene_Init
//		scene_Init_DB
//
//============================================================================

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <iostream.h>
#include <math.h>
#include <Xm/Form.h>
#include <Xm/Frame.h>
#include <Xm/DrawingA.h>
#include <X11/keysym.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>

#include "Geometry.h"
#include "Camera.h"
#include "Node.h"
#include "scene.h"
#include "state.h"
#include "stats.h"
#include "timing.h"
#include "animate.h"
#include "ui.h"

#include "VisualObject.h"
#include "TriangleList.h"

#include "../../include/axes.h"

#include "glutstroke.h"
extern void* glutStrokeRoman;

void	draw_frustum( GraphicsState *gr_state );

//----------------------------------------------------------------------------
//
//	Variables Global to this file
//
//----------------------------------------------------------------------------

GLfloat		color_clear[3]	= { 0.6, 0.6, 0.6 };
//GLfloat		color_sky[3]	= { 0.7, 0.7, 0.0, 1.0 };
GLfloat		color_sky[4]	= { 0.20, 0.5, 0.80, 1.0 };

Matrix4d		Imat;
Node			*object_graph;

GLfloat		light_col0[4]  = { 1, 1, 1, 1 };
GLfloat		light_col1[4]  = { 1, 1, 1, 1 };
GLfloat		light_pos0[4] =	{ 0.0, 1.0, 0.0, 0.0 };
GLfloat		light_pos1[4] =	{ 0.0, 0.5, 1.0, 0.0 };
//GLfloat		light_pos0[4] =	{ -1.0, 1.0, 1.0, 0.0 };

stuffAxes	*axes = NULL;

char	g_db_filename[256];
int		g_db_valid		= 0;
int		g_db_window_id = -1;
double	g_scene_radius = 0.0;

//----------------------------------------------------------------------------
//
//		scene_Init_DB
//
//			Read the object Hierarchy from disk.
//
//----------------------------------------------------------------------------
void
scene_Init_DB( char *db_filename, int window_id )
	{
	//
	//	Read a hierarchy from disk
	//
	sprintf( g_db_filename,"%s",db_filename);
	g_db_window_id = window_id;
	}

void
foo_init_db( AppState *app_state )
	{
	Timer read_timer,neighbor_timer;
	GraphicsState *gr_state = &(app_state->gr_state);

	fprintf(stderr,"Reading the Database\n");
	fprintf(stderr,"-----------------------------\n");
	RESET_TIMER( read_timer );
	START_TIMER( read_timer );
	object_graph = new Node();
	object_graph->setMetricType( gr_state->metric_type );
	object_graph->Read( g_db_filename, gr_state->pixel_eps,
						gr_state->inflate, *(gr_state->camera) );
	STOP_TIMER( read_timer );
	app_state->stats.read_time =  calc_usec_time( read_timer );
	fprintf(stderr,"Done\n\n");

	fprintf(stderr,"Finding Neighboring Triangles\n");
	fprintf(stderr,"-----------------------------\n");
	RESET_TIMER( neighbor_timer );
	START_TIMER( neighbor_timer );
	if( gr_state->inflate )
		object_graph->traversal_findNeighbors( gr_state->inflation );
	else
		object_graph->traversal_initBrickMetricWOInflation();
	STOP_TIMER( neighbor_timer );
	app_state->stats.read_time =  calc_usec_time( neighbor_timer );
	fprintf(stderr,"Done\n\n");

	object_graph->traversal_numBytes( app_state->stats.size, 
									  app_state->stats.size_extra );

	//
	//	Allocate one frame of stats to be used over and over
	//
	app_state->stats.frame = new FrameStat;
	init_FrameStat( app_state->stats.frame[0] );

	for( int i=0; i<app_state->gr_state.num_env_objs; i++ )
		{
		app_state->gr_state.env_objs[i]
			.Read( app_state->gr_state.env_obj_names[i] );
		}
	}

//----------------------------------------------------------------------------
//
//      scene_Init_Camera
//
//----------------------------------------------------------------------------
void
scene_Init_Camera( void *ui_data )
    {
	static AppState	*app_state = (AppState *) ui_data;

    g_scene_radius = (object_graph->getBoundBox()).getRadius();

    // angle,aspect,near,far	640x480 = 1.33 aspect ratio
	//							640x640 = 1.0  aspect ratio
	//
	app_state->gr_state.camera->setViewport( 0,0,640,640 );
    app_state->gr_state.camera->initFrustum( 60.0, 1.0, 0.1, g_scene_radius*4 );
    app_state->gr_state.camera->Reset();
	app_state->gr_state.camera->Move(	app_state->gr_state.ic_x,
										app_state->gr_state.ic_y,
										app_state->gr_state.ic_z );

    app_state->gr_state.camera->Update();

    app_state->gr_state.zoomSpeed = g_scene_radius/150.0;
    app_state->gr_state.dragSpeed = g_scene_radius/50.0;
	app_state->anim_state.speed = app_state->gr_state.zoomSpeed;
    }


 
//----------------------------------------------------------------------------
//
//		scene_Init_GL
//
//			Setup OpenGL state
//
//----------------------------------------------------------------------------
void
scene_Init_GL( uiRenderWindow *ren_win )
	{
	AppState *app_state = (AppState *) ren_win->uiData;

	glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );

	//glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
	glBlendFunc( GL_ONE, GL_ONE_MINUS_SRC_ALPHA );

	glClearDepth(1.0);
	glDepthFunc(GL_LEQUAL);

	glClearAccum(0.0, 0.0, 0.0, 0.0);

	glLightModeli( GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE );
	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHT1);
	glEnable(GL_NORMALIZE);
	glLightfv( GL_LIGHT0, GL_DIFFUSE,  light_col0 );
	glLightfv( GL_LIGHT1, GL_DIFFUSE,  light_col1 );

	//
	//	Initialize the Database
	//

	//
	//	Warm start the Camera
	//
	app_state->gr_state.camera->setViewport( 0,0,640,640 );
    app_state->gr_state.camera->initFrustum( 60.0, 1.0, 0.1, 100000.0 );
	app_state->gr_state.camera->Reset();
	app_state->gr_state.camera->Update();

	if( ren_win->window_id == g_db_window_id )
		{ 
		if( !g_db_valid )
			{
			foo_init_db( app_state );
			g_db_valid = 1;

			scene_Init_Camera( app_state );

			//
			//	Look for extensions
			//
			const GLubyte *extensions = glGetString( GL_EXTENSIONS );
			char *do_multisample = strstr( (const char*)extensions,
											"SGIS_multisample" );
		
			if( do_multisample != NULL )
				object_graph->setMultisample( 1 );
			}
		}
	
	return;
	}
	

//----------------------------------------------------------------------------
//
//		scene_Draw_Preview
//
//			Rendering Routine for the "Preview" window
//
//----------------------------------------------------------------------------
void
scene_Draw_Preview(Widget w, Display *dpy, int *dirtyViewport, void *ui_data)
	{
	static AppState		 *app_state = (AppState *) ui_data;
	static GraphicsState *gr_state = &(app_state->gr_state);

	static Dimension	width, height;

	if( !g_db_valid )	return;

	if (*dirtyViewport)
		{
    	XtVaGetValues(w, XmNwidth, &width, XmNheight, &height, NULL);
    	*dirtyViewport = 0;
		glViewport(0, 0, (GLint) width, (GLint) height);
		}

	glClearColor(color_clear[0], color_clear[1], color_clear[2], 1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glLightfv( GL_LIGHT0, GL_POSITION, light_pos0 );
	glLightfv( GL_LIGHT1, GL_POSITION, light_pos1 );

	gr_state->camera->setupGLProjectionMoveCamera();
	gr_state->camera->setupGLModelViewMoveModel();

	//
	//	Fancy axes
	//
	if( axes == NULL )	axes = new stuffAxes();

	glEnable(GL_DEPTH_TEST);
	glDepthMask(GL_TRUE);
	glEnable(GL_LIGHTING);

	glPushMatrix();
		glScalef( 10.0, 10.0, 10.0 );
		axes->stuffAxesDraw();
	glPopMatrix();

	//
	//	Bounding Boxes
	//
//	object_graph->traversal_drawBoundingBoxes( 0, 0 );
	object_graph->traversal_drawBoundingBoxesSolid();

	glDisable(GL_LIGHTING);

	glFlush();
	glXSwapBuffers(dpy, XtWindow(w));
	}
	

//----------------------------------------------------------------------------
//
//		scene_Draw_Geom
//
//			Rendering Routine for the "Geometry Render" window
//
//----------------------------------------------------------------------------
void
scene_Draw_Geom( Widget w, Display *dpy, int *dirtyViewport, void *ui_data)
	{
	static AppState			*app_state = (AppState *) ui_data;
	static GraphicsState	*gr_state = &(app_state->gr_state);
 	static Dimension		width, height;
	static Timer			frame_timer,cull_timer,draw_timer;

	static double	hz = 0.0;

	int		num_tris = 0;
	char	msg[256];

	if( !g_db_valid )	return;

RESET_TIMER( cull_timer   );
RESET_TIMER( draw_timer   );

	init_FrameStat( app_state->stats.frame[0] );

RESET_TIMER( frame_timer );
START_TIMER( frame_timer );

	if (*dirtyViewport)
		{
    	XtVaGetValues(w, XmNwidth, &width, XmNheight, &height, NULL);
    	*dirtyViewport = 0;
		glViewport(0, 0, (GLint) width, (GLint) height);
		}

	glEnable(GL_DEPTH_TEST);
	glDepthMask(GL_TRUE);

	glClearColor(color_sky[0], color_sky[1], color_sky[2], color_sky[3]);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	gr_state->camera->setupGLProjectionMoveCamera();
	gr_state->camera->setupGLModelViewMoveCamera();

	glLightfv( GL_LIGHT0, GL_POSITION, light_pos0 );
	glLightfv( GL_LIGHT1, GL_POSITION, light_pos1 );

	//
	//	Draw the hierarchy with explicit geometry
	//
START_TIMER( cull_timer );

	if( gr_state->metric_type == Brick )
		object_graph->traversal_cullAndAnnotate_Brick( *gr_state->camera,
													app_state->stats.frame[0] );
	else
		object_graph->traversal_cullAndAnnotate( *gr_state->camera, 0,
													app_state->stats.frame[0] );

STOP_TIMER( cull_timer );


START_TIMER( draw_timer );

	object_graph->traversal_drawGeometry( *gr_state->camera,
										  app_state->stats.frame[0], 0 );

STOP_TIMER( draw_timer );

	if( gr_state->drawBoundingBoxes )
		{
		object_graph->traversal_drawBoundingBoxes( 0,0 );
		}


	//
	//	Environment Objects
	//
	if( gr_state->drawEnvironment )
		{
		glEnable( GL_BLEND );
		glEnable( GL_TEXTURE_2D );
		//
		// tri-list sets the texture env - DECAL,MODULATE,etc.
		//	
		//	This is not optimal if one of these objects is not
		//	texture mapped.
		//
		for( int i=0 ; i<gr_state->num_env_objs; i++ )
			gr_state->env_objs[i].draw();	

		glDisable( GL_TEXTURE_2D );
		glDisable( GL_BLEND );
		}

	//
	//
	//
	if( gr_state->drawHz )
		{ 
		glMatrixMode( GL_PROJECTION );
			glLoadIdentity();
			glOrtho( 0.0,30.0, 0.0,30.0, 0.0,30.0 );

		glMatrixMode( GL_MODELVIEW );
			glLoadIdentity();
			glScalef( .01, .01, .01 );

			sprintf( msg, "Hz: %6.2f  NT: %d", hz, num_tris );
			glColor3f( 0.0, 0.0, 0.0 );
			for( int i=0 ; i<strlen(msg) ; i++ )
				glutStrokeCharacter(&glutStrokeRoman, msg[i]);
		}

//	glFlush();
	glFinish();

	STOP_TIMER( frame_timer );

	glXSwapBuffers(dpy, XtWindow(w));
	
	if( gr_state->drawHz )
		{
		long usec = calc_usec_time( frame_timer );
		hz = 1.0 / ( usec * .000001 );
		}

	if( gr_state->printTimings )
		{
		fprintf( stderr,"~~~~~~~Geometry~~~~~\n");
		fprintf( stderr,"cull:\t%d\n", calc_usec_time( cull_timer ) );
		fprintf( stderr,"draw:\t%d\n", calc_usec_time( draw_timer ) );
		fprintf( stderr,"~~~~~~~~~~~~~~~~~~~~\n");
		}

	app_state->stats.frame[0].time_cull   = calc_usec_time( cull_timer );
	app_state->stats.frame[0].time_update = 0;
	app_state->stats.frame[0].time_draw   = calc_usec_time( draw_timer );
	app_state->stats.frame[0].time        = calc_usec_time( frame_timer );
	}


//----------------------------------------------------------------------------
//
//		scene_Draw_Image
//
//
//----------------------------------------------------------------------------
void
scene_Draw_Image(Widget w, Display *dpy, int *dirtyViewport, void *ui_data )
	{
	static AppState		 *app_state = (AppState *) ui_data;
	static GraphicsState *gr_state = &(app_state->gr_state);
	static Dimension	 width, height;
	static Timer		 frame_timer,cull_timer,update_timer,draw_timer;

	static double	hz = 0.0;

	int		num_tris = 0;
	char	msg[256];

	if( !g_db_valid )	return;

	init_FrameStat( app_state->stats.frame[0] );

RESET_TIMER( frame_timer  );
RESET_TIMER( cull_timer   );
RESET_TIMER( update_timer );
RESET_TIMER( draw_timer   );

START_TIMER( frame_timer );

	//
	//	Set Viewport
	//
	if (*dirtyViewport)
		{
    	XtVaGetValues(w, XmNwidth, &width, XmNheight, &height, NULL);
    	*dirtyViewport = 0;
		glViewport(0, 0, (GLint) width, (GLint) height);
		}

	glEnable( GL_DEPTH_TEST );
	glDepthMask( GL_FALSE );

	gr_state->camera->setupGLProjectionMoveCamera();
	gr_state->camera->setupGLModelViewMoveCamera();

	glLightfv( GL_LIGHT0, GL_POSITION, light_pos0 );
	glLightfv( GL_LIGHT1, GL_POSITION, light_pos1 );


START_TIMER( cull_timer );

	if( gr_state->metric_type == Brick )
		object_graph->traversal_cullAndAnnotate_Brick( *gr_state->camera,
													app_state->stats.frame[0]);
	else
		object_graph->traversal_cullAndAnnotate( *gr_state->camera, 0,
												app_state->stats.frame[0] );

STOP_TIMER( cull_timer );
	
	if( gr_state->drawGeom )
		{
		glClearColor(color_sky[0], color_sky[1], color_sky[2], color_sky[3]);

		glDepthMask(GL_TRUE);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		object_graph->traversal_drawGeometry( *gr_state->camera,
											  app_state->stats.frame[0], 0 );

		//
		//	Environment Objects
		//
		glEnable( GL_BLEND );
		glEnable( GL_TEXTURE_2D );
		if( gr_state->drawEnvironment )
			{
			for( int i=0 ; i<gr_state->num_env_objs; i++ )
				gr_state->env_objs[i].draw();	
			}

		glDisable( GL_TEXTURE_2D );
		glDisable( GL_BLEND );
	
		glDepthMask(GL_FALSE);

		glXSwapBuffers(dpy, XtWindow(w));

		return;
		}


START_TIMER( update_timer );

	glClearColor( 0.0, 0.0, 0.0, 0.0 );
	glEnable( GL_BLEND );
	glEnable( GL_TEXTURE_2D );
	glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
	
	Point3d ray;

	if( app_state->anim_state.animating )
		{
		ray[0] = gr_state->bspline->eval_tan( app_state->anim_state.param, 0 );
		ray[1] = gr_state->bspline->eval_tan( app_state->anim_state.param, 1 );
		ray[2] = gr_state->bspline->eval_tan( app_state->anim_state.param, 2 );
		ray.normalize();
		}
	else
		{
		ray = gr_state->camera->getLook();
		}

	if( gr_state->metric_type == Brick )
		object_graph->traversal_UIC_Brick( *gr_state->camera,
											app_state->anim_state.speed,
											ray,
											gr_state->aa,
											gr_state->printTimings,
											app_state->stats.frame[0] );
	else
		object_graph->traversal_UIC( *gr_state->camera,
									 gr_state->printTimings,
									 gr_state->drawFrames,
									 gr_state->aa,
									 app_state->stats.frame[0] );

STOP_TIMER( update_timer );

START_TIMER( draw_timer );

	if( gr_state->aa )
		{
		glClear(GL_ACCUM_BUFFER_BIT);
		glClearColor(color_sky[0], color_sky[1], color_sky[2], color_sky[3]);

		for( int j=0 ; j<gr_state->numSamples ; j++ )
			{
			gr_state->camera->setupGLProjectionMoveCameraAntiAlias(j);
	
			glDepthMask( GL_TRUE );
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
			glDepthMask( GL_FALSE );

			gr_state->camera->setupGLModelViewMoveCamera();

			object_graph->traversal_drawImageCache( *gr_state->camera,
													gr_state->drawFrames,
													gr_state->drawLeafTextures,
													gr_state->drawPsuedoColor,
													app_state->stats.frame[0] );
			glAccum(GL_ACCUM, 1.0/8.0 );
			}
		glAccum(GL_RETURN, 1.0);
		}
	else
		{
		Camera *cam = gr_state->camera;

		cam->setupGLProjectionMoveCamera();
		cam->setupGLModelViewMoveCamera();

		glLightfv( GL_LIGHT0, GL_POSITION, light_pos0 );
		glLightfv( GL_LIGHT1, GL_POSITION, light_pos1 );

		glClearColor(color_sky[0], color_sky[1], color_sky[2], color_sky[3]);

		glDepthMask( GL_TRUE );
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glDepthMask( GL_FALSE );

		//
		//	Environment Objects
		//
		if( gr_state->drawEnvironment )
			{
			glDepthMask( GL_TRUE );
			for( int i=0 ; i<gr_state->num_env_objs; i++ )
				gr_state->env_objs[i].draw();	
			glDepthMask( GL_FALSE );
			}

		//
		//	In case one of the environment objects screws up the
		//	texture environment.
		//
		glEnable( GL_TEXTURE_2D );
		glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );

		object_graph->traversal_drawImageCache( *cam,
												gr_state->drawFrames,
												gr_state->drawLeafTextures,
												gr_state->drawPsuedoColor,
												app_state->stats.frame[0] );
		}

	glDisable( GL_BLEND );
	glDisable( GL_TEXTURE_2D );
	
STOP_TIMER( draw_timer );

	//
	//
	// 
	if( gr_state->drawHz )
		{
		glMatrixMode( GL_PROJECTION );
			glLoadIdentity();
			glOrtho( 0.0,30.0, 0.0,30.0, 0.0,30.0 );

		glMatrixMode( GL_MODELVIEW );
			glLoadIdentity();
			glScalef( .01, .01, .01 );
			sprintf( msg, "Hz: %6.2f NT:%d", hz, num_tris );
		
			glColor3f( 0.0, 0.0, 0.0 );
			for( int i=0 ; i<strlen(msg) ; i++ )
				glutStrokeCharacter(&glutStrokeRoman, msg[i]);
		}

//	glFlush();
	glFinish();

	STOP_TIMER( frame_timer );

	glXSwapBuffers(dpy, XtWindow(w));

	if( gr_state->drawHz )
		{
		long usec = calc_usec_time( frame_timer );
		hz = 1.0 / ( usec * .000001 );
		}

	if( gr_state->printTimings )
		{
		fprintf( stderr,"~~~~~~~Image~~~~~~~~\n");
		fprintf( stderr,"cull:  \t%d\n", calc_usec_time( cull_timer ) );
		fprintf( stderr,"update:\t%d\n", calc_usec_time( update_timer ) );
		fprintf( stderr,"draw:  \t%d\n", calc_usec_time( draw_timer ) );
		fprintf( stderr,"~~~~~~~~~~~~~~~~~~~~\n");
		}

	object_graph->traversal_finalPassStats( app_state->stats.frame[0] );

	app_state->stats.frame[0].time_cull   = calc_usec_time( cull_timer   );
	app_state->stats.frame[0].time_update = calc_usec_time( update_timer );
	app_state->stats.frame[0].time_draw   = calc_usec_time( draw_timer   );
	app_state->stats.frame[0].time        = calc_usec_time( frame_timer  );
	}


//----------------------------------------------------------------------------
//
//		scene_Draw_SkyCam
//
//			Rendering Routine for the "SkyCam" window
//
//----------------------------------------------------------------------------
void
scene_Draw_SkyCam(	Widget w, Display *dpy, int *dirtyViewport,	void *ui_data)
	{
	static AppState		 *app_state = (AppState *) ui_data;
	static GraphicsState *gr_state = &(app_state->gr_state);
	static Dimension	 width, height;


	Point3d 	t,point;

	static int		got_scene_extents = 0;
	static Point3d	min,max,center;
	static double	radius;

	if( !g_db_valid )	return;

	if( !got_scene_extents )
		{
		object_graph->getSceneExtents( min, max );
		object_graph->traversal_init2DBoundary();
		radius = 0.8 * (max-min).length() / 2.0;
		center = (max+min)/2.0;
		got_scene_extents = 1;
		}

	if (*dirtyViewport)
		{
    	XtVaGetValues(w, XmNwidth, &width, XmNheight, &height, NULL);
    	*dirtyViewport = 0;
		glViewport(0, 0, (GLint) width, (GLint) height);
		}

	glEnable( GL_DEPTH_TEST );
	glDepthMask( GL_TRUE );
/*
	glEnable( GL_LINE_SMOOTH );
	glEnable( GL_BLEND );
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
	glHint( GL_LINE_SMOOTH_HINT, GL_NICEST );
*/
	glClearColor(color_clear[0], color_clear[1], color_clear[2], 1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);


	//
	//	Do the Sky Cam view stuff
	//
	glMatrixMode( GL_PROJECTION );
		glLoadIdentity();

		glOrtho( -radius, radius, -radius, radius, -100000.0, 100000.0 );

	glMatrixMode( GL_MODELVIEW );
		glLoadIdentity();

		gluLookAt(  0.0,  10.0,  0.0,	// position
					0.0,  -1.0,	 0.0,	// direction
					0.0,   0.0,	-1.0 );	// up

		glTranslated( -center[0], -center[1], -center[2] );

	/*
	object_graph->traversal_drawGeometry( *gr_state->camera,
										  app_state->stat_state, 1 );
	glFlush();
	glXSwapBuffers(dpy, XtWindow(w));
	return;
	*/

	//
	//	Draw the splitting planes
	//	Draw the Bounding Boxes   -- w/ ID & Frames
	//	Draw the frustum          -- transformed
	//
	glPushMatrix();
		glLineWidth( 2.0 );
		
		glColor3d( 0.0, 0.0, 0.0 );

		//object_graph->drawPeriphery();
		//object_graph->traversal_draw2DBoundary();

		object_graph->traversal_drawBoundingBoxes( 1,0 );

		/*
		glPopMatrix();
		glFlush();
		glXSwapBuffers(dpy, XtWindow(w));
		return;
		*/

		//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
		//	Draw the Bspline Camera path
		//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
/*	
*/
		glDisable( GL_DEPTH_TEST );

		glColor3d( 1.0, 1.0, 1.0 );
		glLineWidth( 6.0 );
		glBegin( GL_LINE_STRIP );
		for( int i=0 ; i<=100 ; i++ )
			{
			double t = ((double) i )/ 100.0 ; 
			double x =  gr_state->bspline->eval( t, 0 );
			double y =  gr_state->bspline->eval( t, 1 );
			double z =  gr_state->bspline->eval( t, 2 );

			glVertex3d( x,y,z );
			}
		glEnd();
		glLineWidth( 1.0 );

	/*
		int ncp = gr_state->bspline->get_numControlPoints();
		for( i=0 ; i<ncp ; i++ )
			{
			double x = gr_state->bspline->get_cp( i, 0 );
			double y = gr_state->bspline->get_cp( i, 1 );
			double z = gr_state->bspline->get_cp( i, 2 );
			double f = g_scene_radius/40.0;

			//fprintf(stderr,"%f %f\n",x,z );

			glPushMatrix();
			glTranslated( x, y, z );
			glRotated( -90, 1,0,0 );
			glRectd( -f, -f, f, f );
			glPopMatrix();
			}
	*/



		//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
		//	Draw the frustum
		//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
		glLineWidth( 6.0 );
		glColor3d( 0.0, 1.0, 0.3 );
		draw_frustum( gr_state );
		glLineWidth( 1.0 );

	//^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

	glEnable( GL_DEPTH_TEST );
	//glDisable( GL_LINE_SMOOTH );

	glPopMatrix();
	
	glFlush();
	glXSwapBuffers(dpy, XtWindow(w));
	}

//----------------------------------------------------------------------------
//
//		scene_Draw_SideCam
//
//
//----------------------------------------------------------------------------
void
scene_Draw_SideCam(Widget w, Display *dpy, int *dirtyViewport, void *ui_data )
	{
	static AppState		 *app_state = (AppState *) ui_data;
	static GraphicsState *gr_state = &(app_state->gr_state);
	static Dimension	 width, height;


	if( !g_db_valid )	return;

	init_FrameStat( app_state->stats.frame[0] );

	//
	//	Set Viewport
	//
	if (*dirtyViewport)
		{
    	XtVaGetValues(w, XmNwidth, &width, XmNheight, &height, NULL);
    	*dirtyViewport = 0;
		glViewport(0, 0, (GLint) width, (GLint) height);
		}

	glEnable( GL_DEPTH_TEST );
	glDepthMask( GL_FALSE );

	gr_state->camera->setupGLProjectionMoveCamera();
	gr_state->camera->setupGLModelViewMoveCamera();

	glLightfv( GL_LIGHT0, GL_POSITION, light_pos0 );
	glLightfv( GL_LIGHT1, GL_POSITION, light_pos1 );


	if( gr_state->metric_type == Brick )
		object_graph->traversal_cullAndAnnotate_Brick( *gr_state->camera,
													app_state->stats.frame[0]);
	else
		object_graph->traversal_cullAndAnnotate( *gr_state->camera, 0,
												app_state->stats.frame[0] );


	glClearColor( 0.0, 0.0, 0.0, 0.0 );
	glEnable( GL_BLEND );
	glEnable( GL_TEXTURE_2D );
	glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
	
	Point3d ray;

	if( app_state->anim_state.animating )
		{
		ray[0] = gr_state->bspline->eval_tan( app_state->anim_state.param, 0 );
		ray[1] = gr_state->bspline->eval_tan( app_state->anim_state.param, 1 );
		ray[2] = gr_state->bspline->eval_tan( app_state->anim_state.param, 2 );
		ray.normalize();
		}
	else
		{
		ray = gr_state->camera->getLook();
		}

	if( gr_state->metric_type == Brick )
		object_graph->traversal_UIC_Brick( *gr_state->camera,
											app_state->anim_state.speed,
											ray,
											gr_state->aa,
											gr_state->printTimings,
											app_state->stats.frame[0] );
	else
		object_graph->traversal_UIC( *gr_state->camera,
									 gr_state->printTimings,
									 gr_state->drawFrames,
									 gr_state->aa,
									 app_state->stats.frame[0] );

	Camera shoulder_cam;
	shoulder_cam.Reset();
	shoulder_cam.setViewport( 0,0,640,640 );
	shoulder_cam.initFrustum( 50.0, 1.0, 1.0, 9000.0 );
	shoulder_cam.Place( 7500.0, 600.0, 2000.0 );
	shoulder_cam.Look( -0.7, -0.3, 0.7 );

	//Camera *cam = gr_state->camera;
	Camera *cam = &shoulder_cam;

	cam->setupGLProjectionMoveCamera();
	cam->setupGLModelViewMoveCamera();

		glLightfv( GL_LIGHT0, GL_POSITION, light_pos0 );
		glLightfv( GL_LIGHT1, GL_POSITION, light_pos1 );

		glClearColor(color_sky[0], color_sky[1], color_sky[2], color_sky[3]);

		glDepthMask( GL_TRUE );
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glDepthMask( GL_FALSE );

		//
		//	Environment Objects
		//
		if( gr_state->drawEnvironment )
			{
			glDepthMask( GL_TRUE );
			for( int i=0 ; i<gr_state->num_env_objs; i++ )
				gr_state->env_objs[i].draw();	
			glDepthMask( GL_FALSE );
			}

		//GLfloat fog_color[4] = { 0.0, 0.0, 0.0, 1.0 };
		GLfloat fog_color[4] = { 0.20, 0.5, 0.80, 0.0 };

		//glDisable( GL_BLEND );
		glDisable( GL_TEXTURE_2D );

		glEnable( GL_FOG );
		glHint( GL_FOG_HINT, GL_DONT_CARE );
		glFogfv( GL_FOG_COLOR, fog_color );
		//glFogf( GL_FOG_START,  100.0 );
		//glFogf( GL_FOG_END,   7000.0 );
		//glFogi( GL_FOG_MODE, GL_LINEAR );
		glFogi( GL_FOG_MODE, GL_EXP2 );
		glFogf( GL_FOG_DENSITY, 1.0/1000.0  );

		glColor4f( .3, .3, .3, 1.0 );
		object_graph->drawWireFrame();

		glDepthMask( GL_TRUE );
		glLineWidth(3.0);
		glColor4f( 1.0, 1.0, 0.0, 1.0 );
		draw_frustum( gr_state );
		glLineWidth(1.0);
		glDepthMask( GL_FALSE );

		glDisable( GL_FOG );
		glEnable( GL_BLEND );


	//
	//	In case one of the environment objects screws up the
	//	texture environment.
	//
	glEnable( GL_TEXTURE_2D );
	glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );

	object_graph->traversal_drawImageCache( *cam,
											gr_state->drawFrames,
											gr_state->drawLeafTextures,
											gr_state->drawPsuedoColor,
											app_state->stats.frame[0] );

	glDisable( GL_BLEND );
	glDisable( GL_TEXTURE_2D );
	
//	glFlush();
	glFinish();

	glXSwapBuffers(dpy, XtWindow(w));
	}


//----------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------
void
draw_frustum( GraphicsState *gr_state )
	{
	Point3d point;

		//
		//	Near Plane
		//
		glBegin( GL_LINE_LOOP );

			//	left,top,near
			point = (*gr_state->camera)(0,0);
			glVertex3d( point[0], point[1], point[2] );

			// left,bot,near
			point = (*gr_state->camera)(0,1);
			glVertex3d( point[0], point[1], point[2] );

			// right,bot,near
			point = (*gr_state->camera)(0,2);
			glVertex3d( point[0], point[1], point[2] );

			//	right,top,near
			point = (*gr_state->camera)(0,3);
			glVertex3d( point[0], point[1], point[2] );

		glEnd();

		//
		//	Connecting lines
		//
		glBegin( GL_LINES );

			//	left,top,near
			point = (*gr_state->camera)(0,0);
			glVertex3d( point[0], point[1], point[2] );

			//	left,top,far
			point = (*gr_state->camera)(1,0);
			glVertex3d( point[0], point[1], point[2] );

			//	left,bot,near
			point = (*gr_state->camera)(0,1);
			glVertex3d( point[0], point[1], point[2] );

			//	left,bot,far
			point = (*gr_state->camera)(1,1);
			glVertex3d( point[0], point[1], point[2] );

			//	right,bot,near
			point = (*gr_state->camera)(0,2);
			glVertex3d( point[0], point[1], point[2] );

			//	right,bot,far
			point = (*gr_state->camera)(1,2);
			glVertex3d( point[0], point[1], point[2] );

			//	right,top,near
			point = (*gr_state->camera)(0,3);
			glVertex3d( point[0], point[1], point[2] );

			//	right,top,far
			point = (*gr_state->camera)(1,3);
			glVertex3d( point[0], point[1], point[2] );
		glEnd();

		//
		//	Far Plane
		//
		glBegin( GL_LINE_LOOP );

			//	left,top,far
			point = (*gr_state->camera)(1,0);
			glVertex3d( point[0], point[1], point[2] );

			// left,bot,far
			point = (*gr_state->camera)(1,1);
			glVertex3d( point[0], point[1], point[2] );

			// right,bot,far
			point = (*gr_state->camera)(1,2);
			glVertex3d( point[0], point[1], point[2] );

			//	right,top,far
			point = (*gr_state->camera)(1,3);
			glVertex3d( point[0], point[1], point[2] );

		glEnd();
	}
